/************************************************************************
 *
 * ba.c            Version 1.12                                           
 *  Copyright (C) 1997 Rockwell Corporation
 *        Yuval Fisher (fisher@risc.rockwell.com)
 *  
 ************************************************************************/

/* This software module was originally  developed  by  Yuval  Fisher */
/* (Rockwell Science Center) in the course of development of MPEG-4. */
/* This software module is an implementation of a  part  of  one  or */
/* more  MPEG-4  tools  as  specified  by the MPEG-4.  ISO/IEC gives */
/* users of the MPEG-4  free license  to  this  software  module  or */
/* modifications  thereof  for  use in hardware or software products */
/* claiming conformance to the MPEG-4.  Those intending to use  this */
/* software module in hardware or software products are advised that */
/* its use may infringe existing patents.  The original developer of */
/* this  software module and his/her company, the subsequent editors */
/* and their companies, and ISO/IEC have no  liability  for  use  of */
/* this   software   module   or   modifications   thereof   in   an */
/* implementation.   Copyright  is  not  released  for  non   MPEG-4 */
/* conforming  products.  Rockwell Science Center retains full right */
/* to use the code for his/her own purpose,  assign  or  donate  the */
/* code to a  third  party and  to  inhibit third parties from using */
/* the code for non  MPEG-4  conforming  products.   This  copyright */
/* notice  must  be   included   in  all copies or derivative works. */
/* Copyright (c) 1997.                                               */


#define MAKE_SCENE_TREE 0

#include <stdio.h>
#include <stdlib.h>
#include "saol_prec.h"
#include "bifs.h"
#include "bifs_nums.h"
#include "bifs_types.h"

#ifdef _WIN32
#define strcasecmp stricmp
#endif

/*#ifdef macintosh
#define strcasecmp strcmp
#endif*/

/* 
A linked list that associates an internal node type with an internal 
node id. The node id is the 10 bit number stored in the bit stream.          
*/
struct list {
  int id;
  mpeg4_node *node;
  struct list *next;
} *head = NULL, *node = NULL;

struct route_list {
  int id;
  mpeg4_node *out_node, *in_node;
  int out_field, in_field;
  struct route_list *next;
} *route_head = NULL, *route = NULL;

#ifdef macintosh
extern FILE *inf;
#else
FILE *inf;
#endif

FILE *outf;

long unpack();
long bit();
mpeg4_node *new_node();
mpeg4_node *parse_node();

static int nodeIDbits = 10, routeIDbits = 10;

fatal(char *s)
{
  fprintf(stderr,"%s%", s);
  exit(-1);
}

#if 0
main(int argc, char **argv) 
{
  int hasWorldinfo, hasRoutes;
  mpeg4_node *sceneheader = NULL, *sceneroot = NULL;

  /* usage t2b infile outfile */
  if (argc < 3) 
    outf = stdout;
  else 
    outf = fopen(argv[2],"w");

  if (argc < 2) 
    inf = stdin;
  else 
    inf = fopen(argv[1],"r");
	
  unpack(-2, (long)0, inf);


  /* set the positions of each field in the node tables */
  set_field_positions();

  /* Read Nodes */
  hasWorldinfo = bit(1);

  if (hasWorldinfo) {
    sceneheader = parse_node(SFWorldInfoNodeContext);
  }
  sceneroot = parse_node(SFTopNodeContext);


  hasRoutes = bit(1);
  if (hasRoutes) {
    parse_routes();
  }

  if (sceneheader) render(sceneheader);
  render(sceneroot);
  if (hasRoutes) render_routes();

  fclose(inf);
  fclose(outf);

}
#endif

/* ************************************************************ */
/* unpack value using size bits read from inf.                  */
/* this routine can return a value or store into a pointer      */
/* ************************************************************ */
long unpack(size, valueptr, inf)
     int size; char *valueptr;
     FILE *inf;
{
  int i;
  long value = 0;
  static int ptr = 1; /* how many bits are packed in sum so far */
  static int sum;
  static num_of_packed_bytes = 0;

#ifdef _DEBUG
  printf("ba.c:unpack(), size = %d\n", size);
#endif

  /* size == -2 means we initialize things */
  if (size == -2) {
    sum = (unsigned char)fgetc(inf);
    sum <<= 1;
    return((long)0);
  }

  /* size == -1 means we want to peek at the next bit without */
  /* advancing the pointer */
  if (size == -1)
    return((long)((sum&256)>>8));

  if (!valueptr) {
    for (i=0; i<size; ++i, ++ptr,  sum <<= 1, value <<= 1) {
      if (sum & 256) value |= 1;

#ifdef _DEBUG
  printf("i = %d, ptr = %d, sum = %d, value= %d\n", i, ptr, sum, value);
#endif
	
      if (ptr == 8) {
	sum = (unsigned char)getc(inf);
	ptr=0;
      }
    }
#ifdef _DEBUG
  printf("ba.c:unpack(), value = %d, sum = %d\n", value, sum);
#endif

    return((long)value>>1);
  } else { /*put values into a pointed-to value*/
    i = 0;
    while (i < size) {
      if (sum & 256) value |= 1;
      if (ptr == 8) {
	sum = (unsigned char)getc(inf);
	ptr=0;
      }
      ++i;
      ++ptr;
      sum <<= 1;
      if (i%8 == 0) {
	*valueptr = (unsigned char)value;
	value = 0;
	++valueptr;
      }
      value <<= 1;
    }
    if (i%8 != 0) {/* write out last bits if they weren't just written */
      *valueptr = (unsigned char)value;
    }
  }
}


long bit(int n)  {
  static ct = 0;

  ct += n;
  return (unpack(n,0,inf));
}

/* read bits into a pointer location -- kludge for reading floats */
bit_ptr(int n, char *valueptr) {
  unpack(n,valueptr,inf);
}

/* ************************************************************ */

store_id(int nodeID, mpeg4_node *the_node) 
     /* 
Associates the nodeIDbits bit nodeID that is written in the bitstream with
the internal node type.
*/
{
  node = (struct list *)malloc(sizeof (struct list));
  node->next = head;
  node->id = nodeID; 
  node->node = the_node; 
  head = node;
}

mpeg4_node *retrieve_type(int id)
{
  for (node = head;
       node != NULL && node->id != id; node=node->next) ;
 
  if (node == NULL)
    return ((mpeg4_node *)NULL);
  else
    return (node->node);
}
 



/* ************************************************** 
   parsing routines                                  
   ************************************************** */

mpeg4_node *parse_node(int context)
{
  long num_bits, node, isReused, isUpdatable, MaskAccess;	
  int nodeID=-1;
  mpeg4_node *the_node, *temp_node;

#ifdef _DEBUG
  printf("ba.c:parse_node(), context = %d\n", context);
#endif
	
  isReused = bit(1);

// For now, things shouldn't be reused...
  if (isReused) {
#ifdef _DEBUG
  printf("ba.c:parse_node() Reused Node...BAD\n");
#endif
    nodeID = bit(nodeIDbits);
//   the_node = new_node(REUSED_NODE);
    temp_node = retrieve_type(nodeID);
    the_node->id = nodeID;
//    the_node->type = REUSED_NODE;
    ((struct Reused_node *)the_node)->node = temp_node;
  } else {
    num_bits = contextnodes[context].bits;
    node = bit(num_bits);
#ifdef _DEBUG
  printf("ba.c:parse_node(), num_bits = %d, node = %d\n", num_bits, node);
#endif
	
    if (node > contextnodes[context].count) 
      fatal("\nBad node reference in bit stream.\n");
	
    /* convert the context dependent node number to a global number */
    node = contextnodes[context].nodes[node];
#ifdef _DEBUG
  printf("ba.c:parse_node(), node = %d\n", node);
#endif

    the_node = new_node(node);
    isUpdatable = bit(1);
    if (isUpdatable) {
      nodeID = bit(nodeIDbits);
      store_id(nodeID, the_node);
    } 

    the_node->type = node;
    the_node->id = nodeID;

    MaskAccess = bit(1);
    if (MaskAccess) 
      mask_parse_node(the_node);
    else 
      list_parse_node(the_node);

  }
#ifdef _DEBUG
  printf("ba.c:parse_node() Exiting...\n");
#endif
  return(the_node);
}

mask_parse_node(mpeg4_node *the_node) 
{
  long mask, i;

  for (i=0; i<nodes[the_node->type].num_def_fields; ++i) {
    mask = bit(1);
    if (mask) {
      /* read_type here */
      read_field(the_node, i);
    }
  }
}

list_parse_node(mpeg4_node *the_node) 
{
  long endFlag, fieldReference;
	
  endFlag = bit(1);
  while (!endFlag && !feof(inf)) {
    fieldReference = bit(nodes[the_node->type].def_field_bits);
    if (fieldReference > nodes[the_node->type].num_fields) 
      fatal("\nBad field Reference in bit stream.");
    read_field(the_node, fieldReference);
    endFlag = bit(1);
  }
}

read_field(mpeg4_node *the_node, int field) 
{
  long isListDescription, moreFields=1, NumBits, NumFields, i;
  long field_pos;
	
  /* convert to absolute field index */
  field = nodes[the_node->type].def_fields[field];
		
  if (nodes[the_node->type].field_types[field].isList) {
    isListDescription = bit(1);
    if (isListDescription) {

      for (i=0; moreFields && !feof(inf); ++i, moreFields=bit(1)) {
	make_memory(the_node,field, i+1);
	read_native_MFfield(the_node, field, i);
      }

    } else { /* vector description */
      
      NumBits = bit(5);
      NumFields = bit(NumBits); 


      make_memory(the_node,field, NumFields);
      for (i=0; i<NumFields; ++i) {
	read_native_MFfield(the_node, field, i);
      }
    }
		
    /* adjust the num_ value that specifies how many fields we have */

    field_pos = (long)the_node 
      + nodes[the_node->type].field_positions[field] 
      - sizeof (long);
    *((int *)field_pos) = i;
		
  } else 
    read_native_field(the_node, field);
}

/* 
we read field (with index "field") and put it in the node structure 
How do we know where to put it ? We have a fieldposition array that
tells us the position in memory of each field relative to the address
of the node. 
*/

extern void swapint(int *);

read_native_field(mpeg4_node *the_node, int field) 
{
  int type; 
  long field_pos;
  long i;
  float f;
  double d;
  unsigned char c;
  #ifndef macintosh
  unsigned char str[65536];
  #else
  unsigned char *str = (unsigned char *)calloc(65536,sizeof(unsigned char));
  #endif
  
  /* in order to make this code not dependent on the */
  /* parsing order of the template, we don't check the */
  /* index of the type, but actually the string that defines */
  /* the type...  It's less efficient, but nevermind for now. */
	
  field_pos = nodes[the_node->type].field_positions[field];
  field_pos += (long)the_node;
	
  type = nodes[the_node->type].field_types[field].type;
	
  if (!strcasecmp(fieldtypes[type], "URL")) {
    /* check if there is some data there and free it if so */
    if (*(mpeg4_url *)field_pos) {
      free((*(mpeg4_url *)field_pos));
    }

    /* read the string */
    for (i=0, c=bit(8); c; ++i, c=bit(8)) {
      str[i] = c;
    }
    str[i] = 0;

    (*(mpeg4_url *)field_pos) = 
      (mpeg4_url )malloc(sizeof(char)*strlen((char *)str));
    strcpy((*(char **)field_pos),(char *)str);
  }  else if (!strcasecmp(fieldtypes[type], "Bool")) {
    i = bit(1); 
    *((mpeg4_bool *)field_pos) = (mpeg4_bool)i;
  } else if (!strcasecmp(fieldtypes[type], "Vec3f")) {
    bit_ptr(32, (char *)&f);
    swapint((int *)&f); ((mpeg4_vec3f *)field_pos)->x = f;
    bit_ptr(32, (char *)&f); 
    swapint((int *)&f); ((mpeg4_vec3f *)field_pos)->y = f;
    bit_ptr(32, (char *)&f); 
    swapint((int *)&f); ((mpeg4_vec3f *)field_pos)->z = f;
  } else if (!strcasecmp(fieldtypes[type], "Int32")) {
    i = bit(32); /* swapint(&i);  */
    *((mpeg4_int32 *)field_pos) = (mpeg4_int32)i;
  } else if (!strcasecmp(fieldtypes[type], "Vec2f")) {
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_vec2f *)field_pos)->x = f;
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_vec2f *)field_pos)->y = f;
  } else if (!strcasecmp(fieldtypes[type], "Float")) {
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); *((mpeg4_float*)field_pos) = f;
  } else if (!strcasecmp(fieldtypes[type], "String")) {
    /* check if there is some data there and free it if so */
    /*     if (*(mpeg4_string *)field_pos) {
      free((*(mpeg4_string *)field_pos));
    }*/

    /* read the string */
    for (i=0, c=bit(8); c > 0; ++i, c=bit(8)) {
      str[i] = c;
    }
    str[i] = 0;

    (*(mpeg4_string *)field_pos) = 
      (mpeg4_string)malloc(sizeof(char)*strlen((char *)str));
    strcpy((*(char **)field_pos),(char *)str);
  } else if (!strcasecmp(fieldtypes[type], "Time")) {
    bit_ptr(64, (char *)&d);  
    swapdouble(&d); *((mpeg4_time *)field_pos) = d;
  } else if (!strcasecmp(fieldtypes[type], "Color")) {
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_color *)field_pos)->r = f;
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_color *)field_pos)->g = f;
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_color *)field_pos)->b = f;
  } else if (!strcasecmp(fieldtypes[type], "Rotation")) {
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_rotation *)field_pos)->x = f;
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_rotation *)field_pos)->y = f;
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_rotation *)field_pos)->z = f;
    bit_ptr(32, (char *)&f);  
    swapint((int *)&f); ((mpeg4_rotation *)field_pos)->theta = f;
  } else /* we assume it's an object */
    /* the context comes after the regular types */
    *((mpeg4_node **)field_pos) = parse_node(type-NON_OBJECT_TYPES); 

}

/* 
we read field (with index "field") and put it in the node structure 
How do we know where to put it ? We have a fieldposition array that
tells us the position in memory of each field relative to the address
of the node. We read an MF field and put it into position "index".
*/
read_native_MFfield(mpeg4_node *the_node, int field, int index) 
{
  int type; 
  long field_pos;
  long i;
  float f;
  double d;
  char c;
  #ifndef macintosh
  char str[65536];
  #else
  char *str = (char *)calloc(65536,sizeof(char));
  #endif
  
  /* in order to make this code not dependent on the */
  /* parsing order of the template, we don't check the */
  /* index of the type, but actually the string that defines */
  /* the type...  It's less efficient, but nevermind for now. */
	
  field_pos = nodes[the_node->type].field_positions[field];
  field_pos += (long)the_node;
	
  type = nodes[the_node->type].field_types[field].type;
	
  if (!strcasecmp(fieldtypes[type], "URL")) {
    /* check if there is some data there and free it if so 
    if ((*(mpeg4_url **)field_pos)[index]) {
      free(((*(mpeg4_url **)field_pos))[index]));
    } */

    /* read the string */
    for (i=0, c=bit(8); c; ++i, c=bit(8)) {
      str[i] = c;
    }
    str[i] = 0;

    (*(mpeg4_url **)field_pos)[index]= 
      (mpeg4_url )malloc(sizeof(char)*strlen(str));
    strcpy((*(char ***)field_pos)[index],str);
  }  else if (!strcasecmp(fieldtypes[type], "Bool")) {
    i = bit(1); 
    (*((mpeg4_bool **)field_pos))[index] = (mpeg4_bool)i;
  } else if (!strcasecmp(fieldtypes[type], "Vec3f")) {
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_vec3f **)field_pos))[index].x = f;
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_vec3f **)field_pos))[index].y = f;
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_vec3f **)field_pos))[index].z = f;
  } else if (!strcasecmp(fieldtypes[type], "Int32")) {
    i = bit(32); /* swapint(&i);  */
    (*((mpeg4_int32 **)field_pos))[index] = (mpeg4_int32)i;
  } else if (!strcasecmp(fieldtypes[type], "Vec2f")) {
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_vec2f **)field_pos))[index].x = f;
    bit_ptr(32, (char *)&f);swapint((int *)&f); 
    (*((mpeg4_vec2f **)field_pos))[index].y = f;
  } else if (!strcasecmp(fieldtypes[type], "Float")) {
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_float **)field_pos))[index] = f;
  } else if (!strcasecmp(fieldtypes[type], "String")) {
    /* check if there is some data there and free it if so */
    if ((*(mpeg4_string **)field_pos)[index]) {
      free(((*(mpeg4_string **)field_pos))[index]);
    }

    /* read the string */
    for (i=0, c=bit(8); c; ++i, c=bit(8)) {
      str[i] = c;
    }
    str[i] = 0;

    (*(mpeg4_string **)field_pos)[index]= 
      (mpeg4_string)malloc(sizeof(char)*strlen(str));
    strcpy((*(char ***)field_pos)[index],str);
  } else if (!strcasecmp(fieldtypes[type], "Time")) {
    bit_ptr(64, (char *)&d); swapdouble((double *)&d); 
    (*((mpeg4_time **)field_pos))[index] = d;
  } else if (!strcasecmp(fieldtypes[type], "Color")) {
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_color **)field_pos))[index].r = f;
    bit_ptr(32, (char *)&f);swapint((int *)&f); 
    (*((mpeg4_color **)field_pos))[index].g = f;
    bit_ptr(32, (char *)&f);swapint((int *)&f); 
    (*((mpeg4_color **)field_pos))[index].b = f;
  } else if (!strcasecmp(fieldtypes[type], "Rotation")) {
    bit_ptr(32, (char *)&f);swapint((int *)&f); 
    (*((mpeg4_rotation **)field_pos))[index].x = f;
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_rotation **)field_pos))[index].y = f;
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_rotation **)field_pos))[index].z = f;
    bit_ptr(32, (char *)&f); swapint((int *)&f); 
    (*((mpeg4_rotation **)field_pos))[index].theta = f;
  } else /* we assume it's an object */
    /* the context comes after the regular types */
    (*((mpeg4_node ***)field_pos))[index] = parse_node(type-NON_OBJECT_TYPES); 

}


/* 
This subroutine creates memory at the proper location in
the structure to hold a certain number of MF fields.
*/
make_memory(mpeg4_node *the_node, int field, int size) 
{
  long field_pos, native_size;
  int type; 

  field_pos = nodes[the_node->type].field_positions[field];
  field_pos += (long)the_node;
  type = nodes[the_node->type].field_types[field].type;
	
  /* 
     we could avoid this with a sizeof array, but we don't do
     this for now because we don't know how many types there will
     be eventually and we don't want to hard encode the ordering of 
     the types right now
  */

  if (!strcasecmp(fieldtypes[type], "URL")) {
    native_size = sizeof(mpeg4_url);
  } else if (!strcasecmp(fieldtypes[type], "Bool")) {
    native_size = sizeof(mpeg4_bool);
  } else if (!strcasecmp(fieldtypes[type], "Vec3f")) {
    native_size = sizeof(mpeg4_vec3f);
  } else if (!strcasecmp(fieldtypes[type], "Int32")) {
    native_size = sizeof(mpeg4_int32);
  } else if (!strcasecmp(fieldtypes[type], "Vec2f")) {
    native_size = sizeof(mpeg4_vec2f);
  } else if (!strcasecmp(fieldtypes[type], "Float")) {
    native_size = sizeof(mpeg4_float);
  } else if (!strcasecmp(fieldtypes[type], "String")) {
    native_size = sizeof(mpeg4_string *);
  } else if (!strcasecmp(fieldtypes[type], "Time")) {
    native_size = sizeof(mpeg4_time);
  } else if (!strcasecmp(fieldtypes[type], "Color")) {
    native_size = sizeof(mpeg4_color);
  } else if (!strcasecmp(fieldtypes[type], "Rotation")) {
    native_size = sizeof(mpeg4_rotation);
  } else /* we assume it's an object */
    native_size = sizeof(mpeg4_node *);
	
  /* oh lord ... */
  *((void **)field_pos) = (void *)
    realloc(*((void **)field_pos), size*native_size);
}

parse_routes() 
{
  long isListDescription, moreRoutes, numberofRoutes, i;

  isListDescription =  bit(1);

  if (isListDescription) {
    do {
      parse_one_route();
      moreRoutes = bit(1);
    } while (moreRoutes);
  } else {
    numberofRoutes = bit(routeIDbits);
    for (i=0; i< numberofRoutes; ++i) 
      parse_one_route();
  }
}
		
parse_one_route()
{
  long isUpdatable, id, the_field, routeID = -1;
  mpeg4_node *the_node;
  struct route_list *con;
	
  con = (struct route_list *)malloc(sizeof(struct route_list));
  con->next = route_head;
  route_head = con;
	

  isUpdatable = bit(1);
  if (isUpdatable) {
    routeID = bit(routeIDbits);
    fprintf(outf, "ID RouteID%d ", routeID);
  }


  id = bit(routeIDbits);
  the_node = retrieve_type(id);
  the_field = bit(nodes[the_node->type].out_field_bits);
  /* convert the_field to absolution field reference */
  the_field = nodes[the_node->type].out_fields[the_field];

  /* can check for duplicate routeID here */
  con->id = routeID;
  con->out_node = the_node;
  con->out_field = the_field;

  id = bit(routeIDbits);
  the_node = retrieve_type(id);
  the_field = bit(nodes[the_node->type].in_field_bits);
  /* convert the_field to absolution field reference */
  the_field = nodes[the_node->type].in_fields[the_field];

  con->in_node = the_node;
  con->in_field = the_field;

}

#ifdef UPDATE_CODE_CURRENTLY_NOT_FINISHED
updateFrame()
{
  int continue, command;
  do {
    /* the update command is here */
    command = bit(2);
    switch (command) {
    case 0: 
      update_insertion_command();
      break;
    case 1: 
      update_deletion_command();
      break;
    case 2: 
      update_replace_command();
      break;
    case 3:  /* replace the whole scene */
      delete_node(sceneroot);
      sceneroot = parse_node(SFTopNodeContext);
      break;
    }
  }
}


update_insertion_command()
{
  int parameterType1,parameterType2;
  parameterType1 = bit(1);
  switch (parameterType1) {
  case 0: nodeInsert();
    break;
  case 1:
    parameterType2 = bit(1);
    switch (parameterType2) {
    case 0: idxInsert();
      break;
    case 1:
      routeInsert();
      break;
    }
    break;


  }
}

nodeInsert()
{
  int nodeID,insertionPosition1,insertionPosition2, position;

  nodeID = bit(nodeIDbits);

  insertionPosition1 = bit(1);
  switch (insertionPosition1) {
  case 0: /* insert at specified position */
    position = bit(8); /* should we cut this down to the number of */
    /* maximal positions */
    /* figure out our context and get the node */
    parse_node(); FIx
		    break;
  case 1:
    insertionPosition1 = bit(1);
    switch (insertionPosition1) {
    case 0: /* insert at beginning */
      /* figure out our context and get the node */
      parse_node(); FIX
		      break;
    case 1: /* insert at end*/
      /* figure out our context and get the node */
      parse_node(); FIX
		      break;
    }
    break;	
  }
}


/* insert value in a node's field */
idxInsert() 
{
  int nodeID, fieldID, insertionPosition1, insertionPosition2, fieldbits,
    position;
	
  nodeID = bit(nodeIDbits);

  /* figure out how many bits are needed to specify a field */
  fieldbits = FIX;	
  fieldID = bit(fieldbits);

  insertionPosition1 = bit(1);
  switch (insertionPosition1) {
  case 0: /* insert at a certain position */
    position = bit(16);
				/* figure out value type and read it in */
    FIX;
    break;
  case 1: /* insert at beginning or end */
    insertionPosition2 = bit(1);
    switch (insertionPosition2) {
    case 0: /* inseart at beginning */
      /* figure out value type and read it in */
      FIX;
      break;
    case 1: /* insert at end */
      /* figure out value type and read it in */
      FIX;
      break;
    }
    break;
  }
}

routeInsert() {

  /*
    probably just parse_one_route();
    1.5.1.3 Connect Insertion 
    class ConnectInsertion {
    bit(10) departureNodeID ;		// the ID of the departure node
    // it still needs to be clarified
    // how routes are designated
    bit(x) departureFieldID;		// the index of the departure field
    bit(10) arrivalNodeID ;		// the ID of the arrival node
    bit(y) arrivalFieldID;		// the index of the arrival field
    }
  */
}


update_deletion_command()
{
  int parameterType1,parameterType2;
  parameterType1 = bit(1);
  switch (parameterType1) {
  case 0: nodeDelete();
    break;
  case 1:
    parameterType2 = bit(1);
    switch (parameterType2) {
    case 0: idxDelete();
      break;
    case 1:
      routeDelete();
      break;
    }
    break;


  }
}
	
nodeDelete()
{
  int nodeID;

  nodeID = bit(nodeIDbits);
  / * delete the node from the graph */
	}


idxDelete()
{
  int nodeID, fieldID, deletePosition1, deletePosition2, fieldbits,
    position;
	
  nodeID = bit(nodeIDbits);

  /* figure out how many bits are needed to specify a field */
  fieldbits = FIX;	
  fieldID = bit(fieldbits);

  deletePosition1 = bit(1);
  switch (deletePosition1) {
  case 0: /* delete at a certain position */
    position = bit(16);
				/* figure out value type and read it in */
    FIX;
    break;
  case 1: /* delete at beginning or end */
    deletePosition2 = bit(1);
    switch (deletePosition2) {
    case 0: /* delete at beginning */
      FIX;
      break;
    case 1: /* delete at end */
      FIX;
      break;
    }
    break;
  }
}

routeDelete()
{
  int routeID;

  routeID = bit(routeIDbits)
    /* Delete it */
    }

update_replace_command()
{
  int command;
  do {
    /* the update command is here */
    command = bit(2);
    switch (command) {
    case 0: 
      nodeReplace();
      break;
    case 1: 
      fieldReplace();
      break;
    case 2: 
      idxReplace();
      break;
    case 3: 
      routeReplace();
      break;
    }
  }
}

nodeReplace()
{
  int nodeID;

  nodeID = bit(nodeIDbits);

  /* figure out context */
  FIX
    /* get new node */
    parse_node();
}

fieldReplace()
{
  int nodeID,fieldID, fieldbits;

  nodeID = bit(nodeIDbits);

  /* figure out number of bits for field*/
  fieldbits = FIX

    fieldID = bit(fieldbits);
  /* figure out field type */
  /* get new value*/
  FIX
    }

idxReplace()
{
  int nodeID,fieldID, fieldbits, replacePosition1, replacePosition2;

  nodeID = bit(nodeIDbits);

  /* figure out number of bits for field */
  fieldbits = FIX

    fieldID = bit(fieldbits);
  /* figure out field type */
  /* get new value*/

  replacePosition1 = bit(1);
  switch (replacePosition1) {
  case 0: /* replace at a certain position */
    position = bit(16);
				/* figure out value type and read it in */
    FIX;
    break;
  case 1: /* replace at beginning or end */
    replacePosition2 = bit(1);
    switch (replacePosition2) {
    case 0: /* replace at beginning */
      FIX;
      break;
    case 1: /* replace at end */
      FIX;
      break;
    }
    break;
  }
}

routeReplace()
{
  /*
1.5.4.4 Connect Replacement 
class ConnectReplacement {
	bit(10) ConnectID ;			// the ID of the route to be replaced
	bit(10) departureNodeID ;		// the ID of the departure node
	bit(x) departureFieldID;		// the index of thedeparture field
	bit(10) arrivalNodeID ;		// the ID of the arrival node
	bit(y) arrivalFieldID;		// the index of the arrival field
}
*/

  update_scene_replace_command()
    {
      delete_node(sceneroot);
      sceneroot = parse_node(SFTopNodeContext);
    }

  /* 
delete a node and all its children from the scene graph, freeing
all the memory and deleting references to deleted nodes 
*/

  delete_node(mpeg4_node *node)
    {

    }
#endif
