#ifndef __BIFS_NODES_H
#define __BIFS_NODES_H

#include "bifs_types.h"

#define TRUE 1
#define FALSE 0

typedef struct mpeg4_node_struct {
  long id, type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
} mpeg4_node;

struct Reused_node {
  long id, node_type;
  mpeg4_node *node;
};

mpeg4_node *parse_node();


struct AudioBuffer {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  mpeg4_bool loop;
  mpeg4_float pitch;
  mpeg4_time startTime;
  mpeg4_time stopTime;
  long num_children;
  mpeg4_node **children;
  mpeg4_int32 numChan;
  long num_phaseGroup;
  mpeg4_int32 *phaseGroup;
  mpeg4_float length;
  mpeg4_time duration_changed;
  mpeg4_bool isActive;
  sa_real ***inbuf_last, **delbuf;
};  


struct AudioClip {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  mpeg4_string description;
  mpeg4_bool loop;
  mpeg4_float pitch;
  mpeg4_time startTime;
  mpeg4_time stopTime;
  long num_url;
  mpeg4_url **url;
  mpeg4_time duration_changed;
  mpeg4_bool isActive;
  int offset;	// Private: keeps position within clip buffer
  struct decoder_struct *decoder;
  sa_real ***inbuf_last, **clipbuf;
};


struct AudioDelay {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  mpeg4_time delay;
  mpeg4_int32 numChan;
  long num_phaseGroup;
  mpeg4_int32 *phaseGroup;
  sa_real ***inbuf_last, **delbuf;
}; 


struct AudioFX {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  mpeg4_string orch;
  mpeg4_string score;
  long num_params;
  mpeg4_float *params;
  mpeg4_int32 numChan;
  long num_phaseGroup;
  mpeg4_int32 *phaseGroup;
  struct decoder_struct *decoder;
  sa_real ***inbuf_last, **fxinbuf, **fxoutbuf;
  int fxinpt,fxoutpt;
}; 


struct AudioMix {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  mpeg4_int32 numInputs;
  long num_matrix;
  mpeg4_float *matrix;
  mpeg4_int32 numChan;
  long num_phaseGroup;
  mpeg4_int32 *phaseGroup;
  sa_real ***inbuf_last;
}; 

struct AudioSource {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  long num_url;
  mpeg4_url *url;
  mpeg4_float pitch;
  mpeg4_time startTime;
  mpeg4_time stopTime;
  mpeg4_int32 numChan;
  long num_phaseGroup;
  mpeg4_int32 *phaseGroup;
  struct decoder_struct *decoder;
}; 


struct AudioSwitch {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  long num_whichChoice;
  mpeg4_int32 *whichChoice;
  mpeg4_int32 numChan;
  long num_phaseGroup;  
  mpeg4_int32 *phaseGroup;
  sa_real ***inbuf_last;
}; 


struct Group {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  sa_real ***inbuf_last;
}; 


struct Layer2D {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  mpeg4_vec2f size;
  mpeg4_node *background;
  mpeg4_node *viewport;
  sa_real ***inbuf_last;
};


struct Layer3D {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  long num_addChildren;
  mpeg4_node **addChildren;
  long num_removeChildren;
  mpeg4_node **removeChildren;
  long num_children;
  mpeg4_node **children;
  mpeg4_vec2f size;
  mpeg4_node *background;
  mpeg4_node *fog;
  mpeg4_node *navigationInfo;
  mpeg4_node *viewport;
  sa_real ***inbuf_last;
};


struct ListeningPoint {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  mpeg4_bool set_bind;
  mpeg4_bool jump;
  mpeg4_rotation orientation;
  mpeg4_vec3f position;
  mpeg4_string description;
  mpeg4_time bindTime;
  mpeg4_bool isBound;
  sa_real ***inbuf_last;
}; 


struct Sound {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  mpeg4_vec3f direction;
  mpeg4_float intensity;
  mpeg4_vec3f location;
  mpeg4_float maxBack;
  mpeg4_float maxFront;
  mpeg4_float minBack;
  mpeg4_float minFront;
  mpeg4_float priority;
  mpeg4_node *source;
  mpeg4_bool spatialize;
  sa_real ***inbuf_last;
}; 


struct Sound2D {
  long id, node_type;
  long inchan, outchan, srate;
  sa_real **inbuf, **outbuf;
  mpeg4_float intensity;
  mpeg4_vec2f location;
  mpeg4_node *source;
  mpeg4_bool spatialize;
  sa_real ***inbuf_last;
}; 


struct WorldInfo {
  long id, node_type;
  mpeg4_string *info;
  mpeg4_string title;
};

#endif