#include <stdio.h>
#include "saol.h"
#include "bifs_nodes.h"
#include "ac.h"
#include "saol_sched.h"

#ifdef _WIN32
#define strcasecmp stricmp
#endif

#ifdef macintosh
#define strcasecmp strcmp
#endif

/*extern void swapdouble(double *b);
extern void swapint(int *b);
extern void swapshort(short *b);
*/

extern decoder *decoders;
void sa_startup(decoder *d, mpeg4_url url);
comp_buffer *sa_run_frame(decoder *d,au_list *aul);
void vm_startup(decoder *d, mpeg4_url url);
comp_buffer *vm_run_frame(decoder *d,au_list *aul);
struct cu_file_struct *open_cu_file(char *fn);
extern comp_buffer *cb;
extern double ac_time;


char *url_map[] = {"","","","","cu/celp-english.cu","cu/aac.cu","bass.mp4",
		   "piano.mp4","PCd.mp4","sbfd.mp4","claps.mp4","vowels.mp4"};
int url_decoded[] = {0,0,0,0,1,1,0,0,0,0,0,0};
int url_type[] = {0,0,0,0,CELP,AAC,SA,SA,SA,SA,SA,SA};
int decoder_frame[] = {0,-1,-1,-1,240,1024,-1,-1,-1,-1,-1,-1};

/*
char *url_map[] = {"claps.mp4"};
int url_decoded[] = {0};
int url_type[] = {SA};
int decoder_frame[] = {-1};
*/

typedef struct cu_file_struct {
  int srate;
  int outchan;
  FILE *fp;
} cu_file;


int get_decoder_type(mpeg4_url url) {
  int i;
  
/* Test change */
  int j=0;
  char *ext;
  i = strlen(url);
  
  ext = strchr(url, '.');
  
  if (!strcasecmp(ext, ".mp4")) {
    return SA;
  }
  else if (!strcasecmp(ext, ".aac")) {
    return AAC;
  }
  else if (!strcasecmp(ext, ".celp")) {
  	return CELP;
  }
  
  return 0;
  	
/*  i = atoi(url);
  return(url_type[i]); */
}

decoder *new_decoder(int decoder_type) {
  decoder *d;

  d = (decoder *)malloc(sizeof(decoder));
  d->type = decoder_type;
  d->srate = d->frame = d->time = d->done = 0;
  d->priv = NULL;
  d->time = 0;
  
  /* link into main list */
  d->next = decoders;
  decoders = d;

  /* figure out function pointers */
  switch (decoder_type) {
  case SA:
    d->startup = sa_startup;
    d->run_frame = sa_run_frame;
    break;
  default:
    d->startup = vm_startup;
    d->run_frame = vm_run_frame;
  }
  
  return d;
}

void sa_startup(decoder *d,mpeg4_url url) {
  struct cmdinfo cmd;
  sa_decoder *sa;
  
  cmd.sco = cmd.orc = cmd.out = cmd.midi = NULL;
  cmd.sfbank0 = 0;
  cmd.sfbank1 = 0;
  cmd.decoder = d;
  cmd.inct = 0;
  cmd.text = 0;
  cmd.verbose = 0;
  cmd.has_midi = cmd.has_sco = cmd.has_sbf = 0;
  cmd.interp = 5;
/* Test change */
/*  cmd.bit = url_map[atoi((char *)url)]; */
  cmd.bit = url;

  cmd.temp = ".";
	cmd.midi = 0;
	cmd.southpark = 0;
	cmd.segfault = 0;

  d->priv = (void *)malloc(sizeof(sa_decoder));
  sa = (sa_decoder *)d->priv;

  saol_startup(sa,&cmd);

  d->srate = sa->all->g->srate;
  d->outchan = sa->all->g->outchan;
  d->frame = 1/(double)sa->all->g->krate;
  d->time = 0;
  d->done = 0;
}

void vm_startup(decoder *d,mpeg4_url url) {
  cu_file *cu;

/* Test change */
/*  cu = open_cu_file(url_map[atoi((char *)url)]); */
  cu = open_cu_file(url);
  d->srate = cu->srate;
  d->outchan = cu->outchan;
  d->frame = (double)decoder_frame[atoi((char *)url)] / (double)cu->srate;
  d->time = 0;
  d->done = 0;
  d->priv = (void *)cu;
}

comp_buffer *sa_run_frame(decoder *d,au_list *aul) {
  sa_decoder *sa = (sa_decoder *)d->priv;
  comp_buffer *cb;

  cb = sched_run_kcycle(sa,d);
  if (sa->ended) d->done = 1;
  else return cb;
}

FILE *cu_fp(cu_file *cu) {
  return(cu->fp);
}


comp_buffer *vm_run_frame(decoder *d, au_list *aul) {
  cu_file *cu = (cu_file *)d->priv;
  comp_buffer *this_cu = NULL;
  signed char cubuf[4096];
  static sa_real **buf = NULL;
  double time;
  int culen,i,ret;
  FILE *in = cu_fp(cu);

  if (!buf) buf = buffer(1,1024);

  ret = fread(&time,sizeof(double),1,in);
  swapdouble(&time);

#ifdef _DEBUG
  printf("time = %f\n", time);
#endif

  if (ret <= 0) {
    if (ferror(in)) 
		printf("Read error.\n");
    d->done = 1;
    return NULL;
  }
  
  if (time - d->time < 1./d->srate) {
    ret = fread(&culen,sizeof(int),1,in);
    swapint(&culen);
    
    if (ret <= 0) {
      d->done = 1;
      return NULL;
    }
    if ((ret = fread(cubuf, 1, culen, in)) < culen) {
		if (ferror(in))
			printf("Read error.\n");
		else if (feof(in))
			printf("End of file.\n");
      d->done = 1;
	  culen = ret;
    }
    for (i=0;i<culen;i+=2) /* convert big-endian uint16 to double */ {
      swapshort((short *)&cubuf[i]);
      buf[0][i/2] = *((short *)&cubuf[i])/32768.;
    }
    this_cu = add_samples_cb(buf,d,time,culen/2);
    if ((int)(time+d->frame) > (int)time) {
      printf("*"); fflush(stdout);
    } else if ((int)((time+d->frame) * 10) > (int)(time * 10)) {
      printf("."); fflush(stdout);
    }
  }
  else fseek(in,-sizeof(double),SEEK_CUR); /* rewind past timesteamp */
  return(this_cu);
}

     
cu_file *open_cu_file(char *fn) {
  FILE *fp;
  cu_file *cu;
  char buf[1024],token[128];
  int len,ct=0;
  
  cu = (cu_file *)malloc(sizeof(cu_file));
  if (!(fp = fopen(fn,"rb"))) {
    printf("Couldn't open CU file '%s' for read.\n",fn);
    exit(1);
  }
  cu->fp = fp;
  fread(buf,1,4,fp);
  if (strcmp(buf,".cu")) {
    printf("Bad magic number: '%s'\n",buf);
    exit(1);
  }
  len = 0;
  while (buf[len++] = fgetc(fp)); /* get keys and values */
  
  sscanf(buf + ct,"%s",token);
  while (ct < len-1) {
    ct += strlen(token) + 1;
    if (!strcasecmp("streamtype",token)) {
      /* eat and ignore */
      sscanf(buf + ct,"%s",token);
      ct += strlen(token) + 1;
      sscanf(buf + ct,"%s",token);
      ct += strlen(token) + 1;
    } else if (!strcasecmp("channels",token)) {
      sscanf(buf + ct,"%s", token); /* = */
      ct += strlen(token) + 1;
      sscanf(buf + ct,"%s", token);
      ct += strlen(token) + 1;
      cu->outchan = atoi(token);
    } else if (!strcasecmp("format",token)) {
      /* eat and ignore */
      sscanf(buf + ct,"%s",token);
      ct += strlen(token) + 1;
      sscanf(buf + ct,"%s",token);
      ct += strlen(token) + 1;
    } else if (!strcasecmp("samplerate",token)) {
      sscanf(buf + ct,"%s", token); /* = */
      ct += strlen(token) + 1;
      sscanf(buf + ct,"%s", token);
      ct += strlen(token) + 1;
      cu->srate = atoi(token);
    }
    sscanf(buf + ct,"%s",token);
  }
  return(cu);
}
    
  
void get_decoder_samples(decoder *d, sa_real **buf,double time) {
  /* return all the samples in CB from decoder D
     corresponding to time <= TIME.  Put them in BUF */

  int ct, bufpt,j,i,buflen;
  comp_buffer *p;
  sa_real **newd;
  
  ct = 0; bufpt = 0; buflen = framesize(d->node);
  
  for (p=cb;p;p=p->next) {
    ct = 0;
    if (p->d == d && bufpt < buflen) { 
      if (!bufpt && p->time - (time - buflen/(double)d->srate) > 1./d->srate)
	/* this means that the time stamp of the first CU in the CB
	 is later than the beginning of the buffer we're trying to
	fill.  So we only fill the end of the buffer */
	bufpt = buflen - (int)((time - p->time) * d->srate);
      
      while (bufpt < buflen && ct < p->length) {
	for (j=0;j!=d->outchan;j++)
	  buf[j][bufpt] = p->buf[j][ct];
	bufpt++; ct++;
      }
      
      if (ct && ct != p->length) { /* save the end */
	newd = buffer(d->outchan,p->length - ct);
	for (j=0;j!=d->outchan;j++)
	  for (i=0;i!=p->length-ct;i++)
	    newd[j][i] = p->buf[j][ct+i];
	add_samples_cb(newd,d,p->time+(double)ct/d->srate,p->length-ct);
	free_buf(newd,d->outchan);
      }
    }
  }
}

extern double maxframetime,ac_frame_time;

void cleanup_cb() {
  comp_buffer *old;

  while (cb && cb->time - ac_time < -1./cb->d->srate) {
    old = cb;
    cb = cb->next;
    free_buf(old->buf,old->d->outchan);
    free(old);
    
  }
}

comp_buffer *add_samples_cb(sa_real **buf,decoder *d,double time,int length) {
  comp_buffer *p,*p2,*last=NULL;
  int i,j;

#ifdef _DEBUG
  printf("decoders.c:add_samples_cb()\n");
#endif
  p = (comp_buffer *)malloc(sizeof(comp_buffer));
  p->buf = buffer(d->outchan,length);
  for (i=0;i!=d->outchan;i++)
    for (j=0;j!=length;j++)
      p->buf[i][j] = buf[i][j];

  p->time = time;
  p->length = length;
  p->d = d;
  
  for (p2 = cb; p2 && p2->time <= p->time;last=p2,p2=p2->next) ;
  if (!last) {
    p->next = cb;
    cb = p;
  } else {
    p->next = p2;
    last->next = p;
  }
  return p;
}

sa_real **buffer(int numchan, int length) {
  sa_real **buf;
  int j;

  buf = (sa_real **)calloc(numchan,sizeof(sa_real *));
  for (j=0;j!=numchan;j++)
    buf[j] = (sa_real *)calloc(length,sizeof(sa_real));
  return buf;
}

void clear_buf(sa_real **buf,int numchan, int length) {
  int i,j;

  for (i=0;i!=numchan;i++)
    for (j=0;j!=length;j++)
      buf[i][j] = 0;
}

void free_buf(sa_real **buf, int numchan) {
  int j;

  for (j=0;j!=numchan;j++)
    free(buf[j]);

  free(buf);
}
	
void add_buf(sa_real **x, sa_real **target, int length, int *start, int numchan) {
  /* add samples from buffer X onto the end of buffer TARGET;
     START tells where the current end of target is */
  int i, j, off;

  off = start ? *start : 0;
  for (i=0;i!=numchan;i++)
    for (j=0;j!=length;j++)
      target[i][off+j] = x[i][j];

  if (start) *start += length;
}

void shift_buf(sa_real **x, int *len, int ct, int numchan) {
  /* shift the buffer X down CT samples;
     LEN is the current length of X */
  int i,j;

  for (i=0;i!=numchan;i++)
    for (j=0;j!=*len-ct;j++)
      x[i][j] = x[i][j+ct];

  *len -= ct;
}


    
