/* Fix the context for insert etc. */
/*************************************************************************/
/* b2t.c           Version 1.12                                          */
/* BIFS binary to text converter.                                        */
/* Yuval Fisher Mar 10, 1998                                             */
/*************************************************************************/

/* This software module was originally  developed  by  Yuval  Fisher */
/* (Rockwell Science Center) in the course of development of MPEG-4. */
/* This software module is an implementation of a  part  of  one  or */
/* more  MPEG-4  tools  as  specified  by the MPEG-4.  ISO/IEC gives */
/* users of the MPEG-4  free license  to  this  software  module  or */
/* modifications  thereof  for  use in hardware or software products */
/* claiming conformance to the MPEG-4.  Those intending to use  this */
/* software module in hardware or software products are advised that */
/* its use may infringe existing patents.  The original developer of */
/* this  software module and his/her company, the subsequent editors */
/* and their companies, and ISO/IEC have no  liability  for  use  of */
/* this   software   module   or   modifications   thereof   in   an */
/* implementation.   Copyright  is  not  released  for  non   MPEG-4 */
/* conforming  products.  Rockwell Science Center retains full right */
/* to use the code for his/her own purpose,  assign  or  donate  the */
/* code to a  third  party and  to  inhibit third parties from using */
/* the code for non  MPEG-4  conforming  products.   This  copyright */
/* notice  must  be   included   in  all copies or derivative works. */
/* Copyright (c) 1997.                                               */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef macintosh
#include <console.h>
#endif

#include "bifs.h"

/* 
A linked list that associates an internal node type with an internal 
node id. The node id is the 10 bit number stored in the bit stream.          
*/
struct list {
    int id, type;
    struct list *next;
} *head = NULL, *node = NULL;

static char indent[100]=""; /* used for indenting the output */
static int nodeIDbits = 10, routeIDbits = 10;
FILE *inf, *outf;

long unpack(int size, char *valueptr, FILE *inf);
long bit(int n);
void parse_node(int context);
void parse_routes();
void updateFrame();
void add_indentation(char *indent);
void remove_indentation(char *indent);
void mask_parse_node(int node);
void list_parse_node(int node);
void read_field(int node, int field);
void read_native_field(int type);
void parse_one_route();
void update_insertion_command();
void nodeInsert();
void idxInsert();
void routeInsert();
void update_deletion_command();
void nodeDelete();
void idxDelete();
void routeDelete();
void update_replace_command();
void nodeReplace();
void fieldReplace();
void idxReplace();
void routeReplace();

int strcasecmp(const char *str1, const char *str2);

fatal(char *s)
{
	fprintf(stderr,"%s%", s);
	exit(-1);
}

void main(int argc, char **argv) 
{
	int hasWorldinfo, hasRoutes;
	int temp;

/* MacOS Change */
#ifdef macintosh
	argc = ccommand( &argv);
#endif

	/* usage t2b infile outfile */
	if (argc < 3) 
		outf = stdout;
	else 
		outf = fopen(argv[2],"w");

	if (argc < 2) 
		inf = stdin;
	else 
		inf = fopen(argv[1],"r");
	
	unpack(-2, (long)0, inf);


	/* Read Nodes */
	hasWorldinfo = bit(1);

//	if (hasWorldinfo) {
//		parse_node(SFWorldInfoNodeContext);
//	}
	parse_node(SFTopNodeContext);


	hasRoutes= bit(1);
	if (hasRoutes) {
		parse_routes();

	}

	// Change 1/20/2000: Need to eat an extra character
	temp = getc(inf);

	if ( (!feof(inf)) && (temp != -1) )
	{
		ungetc(temp, inf);
		updateFrame();
	}
	
	fclose(inf);
	fclose(outf);

}

/* ************************************************************ */
/* unpack value using size bits read from inf.                  */
/* this routine can return a value or store into a pointer      */
/* ************************************************************ */
long unpack(int size, char *valueptr, FILE *inf)
// Change 1/20/2000: Don't know why it was originally declared this way
//int size; char *valueptr;
//FILE *inf;
{
    int i;
    long value = 0;
    static int ptr = 1; /* how many bits are packed in sum so far */
    static int sum; 


    /* size == -2 means we initialize things */
    if (size == -2) {
        sum = (unsigned char)fgetc(inf);
        sum <<= 1;
        return((long)0);
    }

    /* size == -1 means we want to peek at the next bit without */
    /* advancing the pointer */
    if (size == -1)
        return((long)((sum&256)>>8));

	if (!valueptr) {
     	for (i=0; i<size; ++i, ++ptr,  sum <<= 1, value <<= 1) {
        	if (sum & 256) value |= 1;
	
         	if (ptr == 8) {
            	sum = (unsigned char)getc(inf);
            	ptr=0;
        	}
     	}
    	return((long)value>>1);
	} else { /*put values into a pointed-to value*/
		i = 0;
		while (i < size) {
       		if (sum & 256) value |= 1;
        		if (ptr == 8) {
           		sum = (unsigned char)getc(inf);
           		ptr=0;
       		}
			++i;
			++ptr;
			sum <<= 1;
			if (i%8 == 0) {
				*valueptr = (unsigned char)value;
				value = 0;
				++valueptr;
			}
			value <<= 1;
		}
		if (i%8 != 0) {/* write out last bits if they weren't just written */
				*valueptr = (unsigned char)value;
		}
    }
}


long bit(int n)  {
	if (!feof(inf))
		return (unpack(n,0,inf));
	else
		return 0;
}

/* read bits into a pointer location -- kludge for reading floats */
bit_ptr(int n, char *valueptr) {
	unpack(n,valueptr,inf);
}

/* ************************************************************ */

store_id(int nodeID, int node_type) 
/* 
Associates the N-bit nodeID that is written in the bitstream with
the internal node type.
*/
{
    node = (struct list *)malloc(sizeof (struct list));
    node->next = head;
    node->id = nodeID; 
    node->type= node_type; 
    head = node;
}

retrieve_type(int id)
{
    for (node = head;
        node != NULL && node->id != id; node=node->next) ;
 
    if (node == NULL)
        return -1;
    else
        return (node->type);
}
 

void add_indentation(char *indent) 
//char *indent;
{
	int i;
	i = strlen(indent);
	indent[i] = '\t';
	indent[i+1] = 0;
}

void remove_indentation(char *indent) 
//char *indent;
{
	int i;
	i = strlen(indent);
	if (i == 0) return;
	indent[i-1] = 0;
}


/* ************************************************** */
/* parsing routines                                   */
/* ************************************************** */

void parse_node(int context)
{
	long num_bits, node, isReused, isUpdatable, MaskAccess;	
	int nodeID;
	
	isReused = bit(1);
	if (isReused) {
		nodeID = bit(nodeIDbits);
		fprintf(outf, "\n%sUSE ID%d ", indent, nodeID);
	} else {
		num_bits = contextnodes[context].bits;
		node = bit(num_bits);
	
		if (node > contextnodes[context].count) 
			fatal("\nBad node reference in bit stream.\n");
	
		/* convert the context dependent node number to a global number */
		node = contextnodes[context].nodes[node];

		fprintf(outf, "\n%s", indent);
		isUpdatable = bit(1);
		if (isUpdatable) {
			nodeID = bit(nodeIDbits);
			store_id(nodeID, node);
			fprintf(outf, "DEF ID%d ", nodeID);
		} 

		fprintf(outf, "%s {", nodes[node].nodename);
		add_indentation(indent);

		MaskAccess = bit(1);
		if (MaskAccess) 
			mask_parse_node(node);
		 else 
			list_parse_node(node);

		remove_indentation(indent);
		fprintf(outf, "\n%s}",indent);
	}
}

void mask_parse_node(int node) 
{
	long mask, i;

	for (i=0; i<nodes[node].num_def_fields; ++i) {
		mask = bit(1);
		if (mask) {
			/* read_type here */
			read_field(node, i);
		}
	}
}

void list_parse_node(int node) 
{
	long endFlag, fieldReference;
	
	endFlag = bit(1);
	while (!endFlag && !feof(inf)) {
		fieldReference = bit(nodes[node].def_field_bits);
		if (fieldReference > nodes[node].num_fields) 
			fatal("\nBad field Reference in bit stream.");
		read_field(node, fieldReference);
		endFlag = bit(1);
	}
}

void read_field(int node, int field) 
{
	long isListDescription, moreFields, NumBits, NumFields, i;
	
	/* convert to absolute field index */
	field = nodes[node].def_fields[field];
		
	fprintf(outf, "\n%s%s ",indent, nodes[node].field_names[field]);
	if (nodes[node].field_types[field].isList) {
		fprintf(outf, "[ ");
		add_indentation(indent);

		isListDescription = bit(1);
		if (isListDescription) {
			do {
				read_native_field(nodes[node].field_types[field].type);
				moreFields = bit(1);
				if (moreFields) fprintf(outf, " , ");
			} while (moreFields && !feof(inf));
		} else { /* vector description */
			NumBits = bit(5);
			NumFields = bit(NumBits);
			/* don't want a trailing comma... hence next 6 lines */
			if (NumFields > 0) 
				read_native_field(nodes[node].field_types[field].type);

			for (i=0; i<NumFields-1; ++i) {
				fprintf(outf, " , ");
				read_native_field(nodes[node].field_types[field].type);
			}
		}
		fprintf(outf, "] ");
		remove_indentation(indent);
		
	} else 
		read_native_field(nodes[node].field_types[field].type);

}

void read_native_field(int type) 
{
	long i;
	float f;
	double d;
	char c;
	/* in order to make this code not dependent on the */
	/* parsing order of the template, we don't check the */
	/* index of the type, but actually the string that defines */
	/* the type...  It's less efficient, but nevermind for now. */
	
	if (!strcasecmp(fieldtypes[type], "url")) {
		c = bit(8);
		do {
			fprintf(outf, "%c", c);
			c = bit(8);
		} while (c && !feof(inf));
	} else if (!strcasecmp(fieldtypes[type], "Bool")) {
		i = bit(1); 
		if (i) 
			fprintf(outf, " TRUE ");
		else 
			fprintf(outf, " FALSE ");
		
	} else if (!strcasecmp(fieldtypes[type], "Vec3f")) {
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
	} else if (!strcasecmp(fieldtypes[type], "Int32")) {
		i = bit(32); fprintf(outf, " %d ", i);
	} else if (!strcasecmp(fieldtypes[type], "Vec2f")) {
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
	} else if (!strcasecmp(fieldtypes[type], "Float")) {
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
	} else if (!strcasecmp(fieldtypes[type], "String")) {
		c = bit(8);
		fprintf(outf, "\"");
		do {
			fprintf(outf, "%c", c);
			c = bit(8);
		} while (c && !feof(inf));
		fprintf(outf, "\"");
	} else if (!strcasecmp(fieldtypes[type], "Time")) {
		bit_ptr(64, (char *)&d); fprintf(outf, " %lf ", d);
	} else if (!strcasecmp(fieldtypes[type], "Color")) {
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
	} else if (!strcasecmp(fieldtypes[type], "Rotation")) {
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
		bit_ptr(32, (char *)&f); fprintf(outf, " %f ", f);
	} else /* we assume it's an object */
		/* the context comes after the regular types */
		parse_node(type-NON_OBJECT_TYPES); 

}

void parse_routes() 
{
	long isListDescription, moreRoutes, numberofRoutes, i;

	isListDescription =  bit(1);

	if (isListDescription) {
		do {
			parse_one_route();
			moreRoutes = bit(1);
		} while (moreRoutes);
	} else {
		numberofRoutes = bit(nodeIDbits);
		for (i=0; i< numberofRoutes; ++i) 
			parse_one_route();
	}
}
		
void parse_one_route()
{
	long isUpdatable, id, the_field, routeID;
	int the_node;
	
	fprintf(outf, "\n");

	isUpdatable = bit(1);
	if (isUpdatable) {
		routeID = bit(routeIDbits);
		fprintf(outf, "DEF RouteID%d ", routeID);
	}

	id = bit(routeIDbits);
	the_node = retrieve_type(id);
	the_field = bit(nodes[the_node].out_field_bits);
	/* convert the_field to absolution field reference */
	the_field = nodes[the_node].out_fields[the_field];

	fprintf(outf, "route ID%d.%s to ", id,
			nodes[the_node].field_names[the_field]);

	id = bit(routeIDbits);
	the_node = retrieve_type(id);
	the_field = bit(nodes[the_node].in_field_bits);
	/* convert the_field to absolution field reference */
	the_field = nodes[the_node].in_fields[the_field];

	fprintf(outf, "ID%d.%s", id,
			nodes[the_node].field_names[the_field]);

}


void updateFrame()
{
	int command, moreUpdates;
	do {
			/* the update command is here */
			fprintf(outf, "\n%sUpdate {", indent);
			add_indentation(indent);
			
			command = bit(2);
			switch (command) {
				case 0: 
					fprintf(outf, "\n%saction \"INSERT\"", indent);
					update_insertion_command();
					break;
				case 1: 
					fprintf(outf, "\n%saction \"DELETE\"", indent);
					update_deletion_command();
					break;
				case 2: 
					fprintf(outf, "\n%saction \"REPLACE\"", indent);
					update_replace_command();
					break;
				case 3:  /* replace the whole scene */
					fprintf(outf, "\n%saction \"REPLACESCENE\"", indent);
					parse_node(SFTopNodeContext);
					break;
			}

		remove_indentation(indent);
		fprintf(outf, "\n%s}", indent);

		moreUpdates = bit(1);
	} while (moreUpdates);
}


void update_insertion_command()
{
	int parameterType1,parameterType2;
	parameterType1 = bit(1);
	switch (parameterType1) {
		case 0: 
				fprintf(outf, "\n%swhat \"NODE\"", indent);
				nodeInsert();
				break;
		case 1:
				parameterType2 = bit(1);
				switch (parameterType2) {
					case 0: 
							fprintf(outf, "\n%swhat \"INDEXEDVALUE\"", indent);
							idxInsert();
							break;
					case 1:
							fprintf(outf, "\n%swhat \"ROUTE\"", indent);
							routeInsert();
							break;
				}
				break;


	}
}

void nodeInsert()
{
	int nodeID,insertionPosition1,insertionPosition2, position, 
		node_type, context;

	nodeID = bit(nodeIDbits);
	node_type = retrieve_type(nodeID);
	fprintf(outf, "\n%sID \"%d\"", indent, nodeID);

	insertionPosition1 = bit(1);
	switch (insertionPosition1) {
		case 0: /* insert at specified position */
			position = bit(8); /* should we cut this down to the number of */
							   /* maximal positions */
			fprintf(outf, "\n%sposition %d", indent, position);
			/* figure out our context and get the node */
			break;
		case 1:
				insertionPosition1 = bit(1);
				switch (insertionPosition1) {
					case 0: /* insert at beginning */
						fprintf(outf, "\n%sposition 0", indent);
						break;
					case 1: /* insert at end*/
						fprintf(outf, "\n%sposition -1", indent);
						break;
				}
				break;	
	}
	fprintf(outf, "\n%snode ", indent);
	parse_node(SFWorldNodeContext); /* FIX the context */
}


/* insert value in a node's field */
void idxInsert() 
{
	int nodeID, fieldID, insertionPosition1, insertionPosition2, fieldbits,
		position, node_type, context, the_field, field;

	nodeID = bit(nodeIDbits);
	fprintf(outf, "\n%sID \"%d\"", indent, nodeID);
	node_type = retrieve_type(nodeID);
	
	/* figure out how many bits are needed to specify a field */
	fieldbits= bit(nodes[node_type].in_field_bits);
	fieldID = bit(fieldbits);
	field = nodes[node_type].in_fields[fieldID];

	fprintf(outf, "\n%sfield \"%s\"", indent, 
	 			nodes[node_type].field_names[field]);
	
	insertionPosition1 = bit(1);
	switch (insertionPosition1) {
		case 0: /* insert at a certain position */
				position = bit(16);
				fprintf(outf, "\n%sposition %d", indent, position);
				break;
		case 1: /* insert at beginning or end */
				insertionPosition2 = bit(1);
				fprintf(outf, "\n%sposition 0", indent);
				switch (insertionPosition2) {
					case 0: /* inseart at beginning */
						fprintf(outf, "\n%sposition 0", indent);
						break;
					case 1: /* insert at end */
						fprintf(outf, "\n%sposition -1", indent);
						break;
				}
				break;
	}
	fprintf(outf, "\n%svalue ", indent);
	read_field(node_type, field);
}

void routeInsert() {
	fprintf(outf, "\n%sROUTE ", indent);
	parse_one_route();
}


void update_deletion_command()
{
	int parameterType1,parameterType2;
	parameterType1 = bit(1);
	switch (parameterType1) {
		case 0: 
				fprintf(outf, "\n%swhat \"NODE\"", indent);
				nodeDelete();
				break;
		case 1:
				parameterType2 = bit(1);
				switch (parameterType2) {
					case 0: 
							fprintf(outf, "\n%swhat \"INDEXEDVALUE\"", indent);
							idxDelete();
							break;
					case 1:
							fprintf(outf, "\n%swhat \"ROUTE\"", indent);
							routeDelete();
							break;
				}
				break;


	}
}
	
void nodeDelete()
{
	int nodeID;

	nodeID = bit(nodeIDbits);
	fprintf(outf, "\n%sID \"%d\"", indent, nodeID);
}


void idxDelete()
{
	int nodeID, fieldID, deletePosition1, deletePosition2, fieldbits,
		position, node_type, field;
	
	nodeID = bit(nodeIDbits);
	node_type = retrieve_type(nodeID);
	fprintf(outf, "\n%sID \"%d\"", indent, nodeID);

	/* figure out how many bits are needed to specify a field */
	fieldbits= bit(nodes[node_type].def_field_bits);
	fieldID = bit(fieldbits);
	field = nodes[node_type].def_fields[fieldID];

	fprintf(outf, "\n%sfield \"%s\"", indent, 
	 			nodes[node_type].field_names[field]);
	
	deletePosition1 = bit(1);
	switch (deletePosition1) {
		case 0: /* delete at a certain position */
				position = bit(16);
				fprintf(outf, "\n%sposition %d", indent, position);
				break;
		case 1: /* delete at beginning or end */
				deletePosition2 = bit(1);
				switch (deletePosition2) {
					case 0: /* delete at beginning */
						fprintf(outf, "\n%sposition 0", indent);
						break;
					case 1: /* delete at end */
						fprintf(outf, "\n%sposition -1", indent);
						break;
				}
				break;
	}
}

void routeDelete()
{
	int routeID;

	routeID = bit(routeIDbits);
	fprintf(outf, "\n%sID \"%d\"", indent, routeID);
}

void update_replace_command()
{
	int command;

	/* the update command is here */
	command = bit(2);
	switch (command) {
		case 0: 
			fprintf(outf, "\n%swhat \"NODE\"", indent);
			nodeReplace();
			break;
		case 1: 
			fprintf(outf, "\n%swhat \"FIELD\"", indent);
			fieldReplace();
			break;
		case 2: 
			fprintf(outf, "\n%swhat \"INDEXEDVALUE\"", indent);
			idxReplace();
			break;
		case 3: 
			fprintf(outf, "\n%swhat \"ROUTE\"", indent);
			routeReplace();
			break;
	}
}

void nodeReplace()
{
	int nodeID;

	nodeID = bit(nodeIDbits);
	fprintf(outf, "\n%sID \"%d\"", indent, nodeID);

	/* get new node */
	fprintf(outf, "\n%snode ", indent);
	parse_node(SFWorldNodeContext); /* FIX the context */
}

void fieldReplace()
{
	int nodeID,fieldID, fieldbits, field, node_type;

	nodeID = bit(nodeIDbits);
	fprintf(outf, "\n%sID \"%d\"", indent, nodeID);
	node_type = retrieve_type(nodeID);

	/* figure out number of bits for field*/
	fieldbits= bit(nodes[node_type].def_field_bits);
	fieldID = bit(fieldbits);
	field = nodes[node_type].def_fields[fieldID];

	fprintf(outf, "\n%sfield \"%s\"", indent, 
	 			nodes[node_type].field_names[field]);

	/* get new value*/
	fprintf(outf, "\n%svalue ", indent);
	read_field(node_type, field);
}

void idxReplace()
{
	int nodeID,fieldID, fieldbits, replacePosition1, replacePosition2, 
		node_type, field, position;


	nodeID = bit(nodeIDbits);
	fprintf(outf, "\n%sID \"%d\"", indent, nodeID);
	node_type = retrieve_type(nodeID);

	/* figure out number of bits for field*/
	fieldbits= bit(nodes[node_type].def_field_bits);
	fieldID = bit(fieldbits);
	field = nodes[node_type].def_fields[fieldID];

	fprintf(outf, "\n%sfield \"%s\"", indent, 
	 			nodes[node_type].field_names[field]);


	replacePosition1 = bit(1);
	switch (replacePosition1) {
		case 0: /* replace at a certain position */
				position = bit(16);
				fprintf(outf, "\n%sposition %d", indent, position);
				break;
		case 1: /* replace at beginning or end */
				replacePosition2 = bit(1);
				switch (replacePosition2) {
					case 0: /* replace at beginning */
						fprintf(outf, "\n%sposition 0", indent);
						break;
					case 1: /* replace at end */
						fprintf(outf, "\n%sposition -1", indent);
						break;
				}
				break;
	}

	fprintf(outf, "\n%svalue ", indent);
	read_field(node_type, field);
}

void routeReplace()
{
	fprintf(stderr, "\n Route replacement not 100% complient.");
	fprintf(outf, "\n%sROUTE ", indent);
	parse_one_route();
}

