/*************************************************************************/
/* b2t.c           Version 1.12                                         */
/* BIFS text to binary converter.                                        */
/* Yuval Fisher Mar 10, 1998                                             */
/*************************************************************************/

/* This software module was originally  developed  by  Yuval  Fisher */
/* (Rockwell Science Center) in the course of development of MPEG-4. */
/* This software module is an implementation of a  part  of  one  or */
/* more  MPEG-4  tools  as  specified  by the MPEG-4.  ISO/IEC gives */
/* users of the MPEG-4  free license  to  this  software  module  or */
/* modifications  thereof  for  use in hardware or software products */
/* claiming conformance to the MPEG-4.  Those intending to use  this */
/* software module in hardware or software products are advised that */
/* its use may infringe existing patents.  The original developer of */
/* this  software module and his/her company, the subsequent editors */
/* and their companies, and ISO/IEC have no  liability  for  use  of */
/* this   software   module   or   modifications   thereof   in   an */
/* implementation.   Copyright  is  not  released  for  non   MPEG-4 */
/* conforming  products.  Rockwell Science Center retains full right */
/* to use the code for his/her own purpose,  assign  or  donate  the */
/* code to a  third  party and  to  inhibit third parties from using */
/* the code for non  MPEG-4  conforming  products.   This  copyright */
/* notice  must  be   included   in  all copies or derivative works. */
/* Copyright (c) 1997.                                               */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#ifdef macintosh
#include <console.h>
#endif

#include <string.h>
#include <math.h>

#include "bifs.h"


/* 
A linked list that associates an internal node number with an internal 
node type. The node number is the 10 bit number stored in the bit stream. 
The id is a text string holding the node name.
*/
struct list {
	int type, num;
	char *id; 
	struct list *next;
} *head = NULL, *node = NULL;


#define READ_WHITE_SPACE for (c = getc(inf); \
							isspace(c) && !feof(inf); \
								c = getc(inf)) if (c == '\n') ++line; \
							ungetc(c, inf);

char token[2000];
static int line=0;
FILE *inf, *outf;
static int nodeID=0, routeID=0;
static int nodeIDbits = 10, routeIDbits = 10;

struct update_command {
		char action[200], what[200], ID[200], field[200], 
				position[200], value[200];
} update_com;


// Forward declarations
void put_bit(int bits, long val);
void put_bit_ptr(int bits, char *val);
int pack(int size, long value, char *valueptr, FILE *foutf);
int parse_node(int context);
void read_value(struct FieldType type);
void read_native_value(int type);
void read_route();
int read_update();
void output_update();

int isspecial(char c);
void get_token(char *token);
void peek_token(char *token);
void get_string(char *token);
void store_id(char *id,int type,int nodeID);
int retrieve_id(char *token, int *type);
void separate(char *token, char *part1, char *part2);

int strcasecmp(const char *str1, const char *str2);


void main(int argc, char **argv) {
	
	char c;
	long hasWorldInfo, hasRoutes, hasUpdates;

/* MacOS Change */
#ifdef macintosh
	argc = ccommand( &argv);
#endif

	/* usage t2b infile outfile */
	if (argc < 3) 
		outf = stdout;
	else 
		outf = fopen(argv[2],"w");

	if (argc < 2) 
		inf = stdin;
	else 
		inf = fopen(argv[1],"r");
	
	/* We have to see (by hand) if there is a WorldInfo */
/*	get_token(token);
	if (!strcasecmp(token, "DEF")) { 
		get_token(token);  // get the ID
		get_token(token); 
		if (!strcasecmp(token, "WorldInfo")) { 
			hasWorldInfo = 1;
		} else  {
			hasWorldInfo = 0;
		}
	} else {
		if (!strcasecmp(token, "WorldInfo")) { 
			hasWorldInfo = 1;
		} else  {
			hasWorldInfo = 0;
		}
	}

	rewind(inf);
*/

	hasWorldInfo = 0;	
	put_bit(1,hasWorldInfo);
/*	if (hasWorldInfo) parse_node(SFWorldInfoNodeContext); */

	parse_node(SFTopNodeContext); /* parse the top group node */

	READ_WHITE_SPACE
	peek_token(token);

	if (!strcasecmp(token, "route") || !strcasecmp(token, "DEF")) {
		hasRoutes = 1;
	} else {
		hasRoutes = 0;
	}
	
	put_bit(1,hasRoutes);
	if (hasRoutes) {
		/* we just do a list description for now */
		put_bit(1,1);
		do {
			read_route();
			/* see if there are more routes */	
		
			peek_token(token);
			if (!strcasecmp(token, "route") || !strcasecmp(token, "DEF")) 
				put_bit(1,1); /* more routes coming */
		 	else 
				put_bit(1,0); /* no more routes coming */
		} while (!strcasecmp(token, "route") || !strcasecmp(token, "DEF"));
	}

	READ_WHITE_SPACE
	peek_token(token);

	if (!strcasecmp(token, "update")) {
		hasUpdates= 1;
	} else {
		hasUpdates = 0;
	}
	
	do {
		if (hasUpdates) 
			read_update();

		READ_WHITE_SPACE
		peek_token(token);

		if (!strcasecmp(token, "update")) {
			hasUpdates= 1;
		} else {
			hasUpdates = 0;
		}

		put_bit(1,hasUpdates);
	} while (hasUpdates);
	

	pack(-1,0,0,outf);
	printf("\n %d bytes written.\n", pack(-2,0,0,outf));
	fclose(inf);
	fclose(outf);
}


/* ************************************************************ */
/* pack value using size bits and output into foutf */
/* We can put things both by giving a ptr to a value or a value */
/* ************************************************************ */

int pack(int size, long value, char *valueptr, FILE *foutf)

// Changed 1/19/2000: Why this is declared this way, I don't know
//int pack(size, value, valueptr, foutf)
//int size; char *valueptr; long value;
//FILE *foutf;
{
     int i; 
	 unsigned long bit_head;
     static int ptr = 1, /* how many bits are packed in sum so far */
                sum = 0, /* packed bits */ 
                num_of_packed_bytes = 0; /* total bytes written out */

    /* size == -1 means we are at the end, so write out what is left */
    if (size == -1 && ptr != 1) {
        fputc(sum<<(8-ptr), foutf);
        ++num_of_packed_bytes;
        return(0);
    }
    
    /* size == -2 means we want to know how many bytes we have written */
    if (size == -2) 
            return(num_of_packed_bytes);
    
	if (valueptr == 0) { /* write by value */
    	for (i=0, bit_head = 1<<(size-1);  
				i<size; ++i, ++ptr, bit_head >>= 1, sum <<= 1) {
        	if (value & bit_head) sum |= 1;
	
         	if (ptr == 8) {
            	fputc(sum,foutf);
            	++num_of_packed_bytes;
            	sum=0;
            	ptr=0;
        	}
     	}
	} else { /* write by ptr to value */
		i = 0;
		if (size > 8) {
			bit_head = 1<<7; 
		} else 
			bit_head = 1<<(size-1);
		value = *valueptr;
		while (i<size) {
        	if (value & bit_head) sum |= 1;
	
         	if (ptr == 8) {
            	fputc(sum,foutf);
            	++num_of_packed_bytes;
            	sum=0;
            	ptr=0;
        	}
	
			++i;
			++ptr;
			sum <<= 1;
			bit_head >>= 1;
	
			if (i%8 == 0) {
				++valueptr;
				value = *valueptr;
				if (size-i > 8) {
					bit_head = 1<<7; 
				} else 
					bit_head = 1<<(size-i-1);
			}
     	}
	}
}

void put_bit(int bits, long val)  {
	if ( pow(2,bits) <= val)
	{
		fprintf(stderr, "\nError: Trying to put too large a value, %d, in only %d bits.\n",val, bits);
		exit(-1);
	}
	pack(bits, val, 0, outf);
}

/* write bits from a pointer location -- kludge for writing floats */
void put_bit_ptr(int bits, char *val)  {
	pack(bits, 0, val, outf);
}


/* ************************************************************ */

int parse_node(int context) 
{
	char token[200], token1[200], identifier[200];
	int type;
	int	isUpdatable = 0, id, bits, i, 
		the_node, context_index, the_field;

	get_token(token);
	if (!*token) return 0; /* end of file */

	if (!strcasecmp(token, "USE")) {
		put_bit(1,1); 		/* isReused */
		get_token(token); 	/* read identifier */
		if ((id = retrieve_id(token, &type)) == -1) {
			fprintf(stderr, 
			"\nError: No previous ID for node %s near line %d\n",token,line);
			exit(-1);
		}; 
		put_bit(nodeIDbits, id);
		return 1;
	} else {
		put_bit(1,0); 		/* isReused */
	}

	if (!strcasecmp(token, "DEF")) {
		/* the bits for this come after the node type */
		isUpdatable = 1;
		get_token(identifier); 
		get_token(token); /* the node type */
	}

	/* compare the token to the list of node names */
	for (the_node=0; the_node<contextnodes[context].count && 
				strcasecmp(token, /* horrible indirection here */
					nodes[ contextnodes[context].nodes[the_node] ].nodename); 
				++the_node); 

	if (the_node == contextnodes[context].count) {
		fprintf(stderr, "\nIn context: %s. ",
				 nodetypes[context]);
		fprintf(stderr,"Unexpected '%s' near line %d.\n",token, line);
		exit(-1);
	}

	/* convert the_node to an absolute internal number */
	the_node = contextnodes[context].nodes[the_node]; 

	/* find current context */
	for (context_index=0; context_index<nodes[the_node].num_context && 
		nodes[the_node].context[context_index] != context; ++context_index) ;

	if (context_index == nodes[the_node].num_context) {
		fprintf(stderr,"Error: inconsistant context '%s' in table for %s \
node. Context = %d, the node=%d\n",
				 nodetypes[context], token, context, the_node);
		exit(-1);
	}

	put_bit(nodes[the_node].context_bits[context_index], 
				nodes[the_node].context_code[context_index]);


	if (isUpdatable) {
		put_bit(1,1);
		store_id(identifier, the_node, nodeID); /* store into list of IDs */	
		put_bit(nodeIDbits, nodeID++);
	} else {
		put_bit(1,0);
	}


	get_token(token1); 
	if (strcmp(token1,"{")) {
		fprintf(stderr, "\nError: '{' expected near line %d", line);
		exit(-1);
	}
	get_token(token1); 

	if (!strcmp(token1,"}")) { /* empty node:  */
			 put_bit(1,0);  /* use a list description */
			 put_bit(1,1);  /* endFlag = 1 -- nothing in list */
			 return(1);
	}

	/* Here we can figure out if mask or list descriptions are */
	/* more bit efficient. For now, everything is a list...   */

	put_bit(1,0);  /* use a list description */
	put_bit(1,0); /* Not the endFlag  */


	do {
		for (the_field = 0; 
				the_field < nodes[the_node].num_def_fields && 
				/* compare the field (token1) with the list of field names */
				strcasecmp(token1, 
					nodes[the_node].field_names[
						nodes[the_node].def_fields[the_field] 
					]);
				++the_field);
	
		if (the_field == nodes[the_node].num_def_fields)		
		{
			fprintf(stderr,"\nUnrecognized DEF field '%s' near line %d.\n",
					token1, line);
			exit(-1);
		}

//		Changed 1/29/00, want to store the DEF field index, not absolute field index
//		put_bit(nodes[the_node].def_field_bits, nodes[the_node].def_fields[the_field]);
		put_bit(nodes[the_node].def_field_bits, the_field);

		read_value(nodes[the_node].field_types[
						nodes[the_node].def_fields[the_field] 
					]);

		get_token(token1); 
		if (!strcmp(token1,"}")) { /* last field */
				 put_bit(1,1); /* endFlag = 1: no more fields */
				 return (1);
		} else {
 			put_bit(1,0); /* endFlag = 0: more fields coming */
		}
	} while(strcmp(token1, "}"));

	return(1);
}

/* ************************************************************ */
void read_value(struct FieldType type)
{
	char token[200];
	if (type.isList) {
		put_bit(1,1); /* always use a ListDescription for now */
		get_token(token); 
		if (strcmp(token,"[")) {
			fprintf(stderr, "\nError: '[' expected near line %d", line);
			exit(-1);
		}
		type.isList=0;
		do { 
			read_native_value(type.type);
			peek_token(token); 
			if (strcmp(token,"]")) {
				put_bit(1,1); /* more fields comming */
			}
			if (!strcmp(token,",") || !strcmp(token, "]")) 
				get_token(token);  /* eat commas  and closing brackets */
		} while (strcmp(token, "]"));
		put_bit(1,0); /* no more fields comming */
		if (strcmp(token,"]")) {
			fprintf(stderr, "\nError: ']' expected near line %d", line);
			exit(-1);
		}
	} else 
		read_native_value(type.type);

}
/* ************************************************************ */
void read_native_value(int type)
{
	
	char token[200];
	int i;
	float f;
	double d;
	char c;
	/* in order to make this code not dependent on the */
	/* parsing order of the template, we don't check the */
	/* index of the type, but actually the string that defines */
	/* the type...  It's less efficient, but nevermind for now. */
	
	if (!strcasecmp(fieldtypes[type], "url")) {
		get_token(token); 
		for (i=0; token[i]; ++i) put_bit(8,token[i]);
		put_bit(8,0);
	} else if (!strcasecmp(fieldtypes[type], "Bool")) {
		get_token(token);
		if (!strcasecmp(token, "true")) 
				put_bit(1, 1);
		else 
				put_bit(1, 0);
	} else if (!strcasecmp(fieldtypes[type], "Vec3f")) {
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
	} else if (!strcasecmp(fieldtypes[type], "Int32")) {
		get_token(token); i = atoi(token); 
		put_bit(32, i);
	} else if (!strcasecmp(fieldtypes[type], "Vec2f")) {
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
	} else if (!strcasecmp(fieldtypes[type], "Float")) {
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
	} else if (!strcasecmp(fieldtypes[type], "String")) {
		get_string(token); 
		for (i=0; token[i]; ++i) put_bit(8,token[i]);
		put_bit(8,0);
	} else if (!strcasecmp(fieldtypes[type], "Time")) {
		get_token(token); d = atof(token); put_bit_ptr(64, (char *)&d);/*FIX*/
	} else if (!strcasecmp(fieldtypes[type], "Color")) {
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
	} else if (!strcasecmp(fieldtypes[type], "Rotation")) {
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
		get_token(token); f = atof(token); put_bit_ptr(32, (char *)&f);
	} else /* we assume it's an object */
		/* the context comes after the regular types */
		parse_node(type-NON_OBJECT_TYPES); 


}

/* ************************************************************ */
void read_route()
{
	char token[200], token1[200], token2[200], identifier[200];
	long isUpdatable, id, the_field;
	int the_node;


	get_token(token); 
	if (!strcasecmp(token, "DEF")) {
		isUpdatable = 1;
		get_token(identifier); 
		store_id(identifier, 0, routeID); /* store into symbol table*/	
		get_token(token);
	} else 
		isUpdatable = 0;

	if (strcasecmp(token,"route")) {
		fprintf(stderr, "\nError: 'route' expected near line %d", line);
		exit(-1);
	}

	get_token(token);
	separate(token, token1, token2); /* separate "token1.token2" */

	put_bit(1,isUpdatable);

	if (isUpdatable) {
		put_bit(routeIDbits, routeID++);
	}
	id = retrieve_id(token1, &the_node);
	if (id==-1) {
		fprintf(stderr, "\nError: unrecognized node ID: %s near line %d", 
					token1, line);
		exit(-1);
	}

	put_bit(routeIDbits,id);

	/* put the bits for the out field */
	for (the_field = 0; 
			the_field < nodes[the_node].num_out_fields && 
			/* nasty indirection */
			strcasecmp(token2, 
				nodes[the_node].field_names[
					nodes[the_node].out_fields[the_field] 
				]);
			++the_field);

	if (the_field == nodes[the_node].num_out_fields) {
		fprintf(stderr,
			"\nUnrecognized field '%s' in route near line %d.\n",
				token2,line);
		exit(-1);
	}

	put_bit(nodes[the_node].out_field_bits, the_field);
	
	
	/* get the "TO" and the in node and field */
	get_token(token);
	if (strcasecmp(token,"to")) {
		fprintf(stderr, "\nError: 'to' expected near line %d", line);
		exit(-1);
	}

	get_token(token);
	separate(token, token1, token2);

	/* get the id for the target of the route */
	id = retrieve_id(token1, &the_node);
	if (id==-1) {
		fprintf(stderr, "\nError: unrecognized node ID: %s near line %d", 
					token1, line);
		exit(-1);
	}

	put_bit(routeIDbits,id);

	/* put the bits for the in field */
	for (the_field = 0; 
			the_field < nodes[the_node].num_in_fields && 
			/* find the field in the list of all input fields */
			strcasecmp(token2, 
				nodes[the_node].field_names[
					nodes[the_node].in_fields[the_field] 
				]);
			++the_field);

	if (the_field == nodes[the_node].num_in_fields) {
		fprintf(stderr,
			"\nUnrecognized field '%s' in route to %s near line %d.\n",
				token2,nodes[the_node].nodename, line);
		exit(-1);
	}

	put_bit(nodes[the_node].in_field_bits, the_field);
		

}


/* ************************************************************ */
int read_update()
{
	char token[200], token1[200], outputted = 0;

	get_token(token); /* The Update command */
	if (!*token) {
			fprintf(stderr, "\nUnexpected end of file, line %d, line");
			exit(-1);
	}


	get_token(token1); 
	if (strcmp(token1,"{")) {
		fprintf(stderr, "\nError: '{' expected near line %d", line);
		exit(-1);
	}
	get_token(token1); 

	if (!strcmp(token1,"}")) { /* empty node:  */
		fprintf(stderr, "\nError: Empty update node: line %d.", line);
		exit(-1);
	}

	/* we read the node into a structure first and then output */
	/* what we need using the structure */
	do {
		if (!strcasecmp(token1, "action")) {
			get_string(token);
			strcpy(update_com.action, token);
		} else
		if (!strcasecmp(token1, "what")) {
			get_string(token);
			strcpy(update_com.what, token);
		} else
		if (!strcasecmp(token1, "ID")) {
			get_string(token);
			strcpy(update_com.ID, token);
		} else
		if (!strcasecmp(token1, "position")) {
			get_token(token);
			strcpy(update_com.position, token);
		} else
		if (!strcasecmp(token1, "field")) {
			get_string(token);
			strcpy(update_com.field, token);
		} else
		if (!strcasecmp(token1, "value")) {
			get_string(token);
			strcpy(update_com.value, token);
		} else
		if (!strcasecmp(token1, "ROUTE")) {
			/* A nasty kluge here for now:                       */
			/* We insist that the ROUTE come last and when we hit */
			/* it, we output the update and then parse the node  */
			printf("\nIncluding ROUTE in update stream. This should be ");
			printf("\nthe last field in the update command...");
			output_update();
			outputted = 1;
		} else
		if (!strcasecmp(token1, "NODE")) {
			/* A nasty kluge here for now:                       */
			/* We insist that the node come last and when we hit */
			/* it, we output the update and then parse the node  */
			printf("\nIncluding NODE in update stream. This should be ");
			printf("\nthe last field in the update command...");
			output_update();
			outputted = 1;

		}
			
		get_token(token1); 
	} while(strcmp(token1, "}"));

	if (!outputted) output_update();

	return(1);
}

/* ************************************************** 
Actually output the contents of update_com into a 
BIFS update command 
**************************************************  */
void output_update()
{
	int id, position, the_field, the_node;
	char token[200], token1[200], token2[200], identifier[200];

	/* *****************************************************************  */
	if (!strcasecmp(update_com.action, "insert")) {
		put_bit(2,0); 
		/* *******************************************************  */
		if (!strcasecmp(update_com.what, "node")) {
			put_bit(1,0); 
			if ((id = retrieve_id(update_com.ID, &the_node)) == -1) {
				fprintf(stderr, 
				"\nError: No previous ID for node %s in update\n",
						update_com.ID);
				exit(-1);
			} 
			put_bit(nodeIDbits, id);
			position = atoi(update_com.position);
			if (position > 0) {
					if (position > 255) {
						fprintf(stderr,"\nError: Only 8 update position bits.");
							exit(-1);
					}
					put_bit(1,0); 
					put_bit(8,position); 
					parse_node(SFWorldNodeContext);
			} else if (position == 0) {
					put_bit(2,2); 
					parse_node(SFWorldNodeContext);
			} else if (position == -1) {
					put_bit(2,3); 
					parse_node(SFWorldNodeContext);
			} else {
					fprintf(stderr,"\nError: Bad update position near line %d.",
							line);
					exit(-1);
			}
		/* *******************************************************  */
		} else if (!strcasecmp(update_com.what, "indexedvalue")) {
			put_bit(2,2); 
			if ((id = retrieve_id(update_com.ID, &the_node)) == -1) {
				fprintf(stderr, 
				"\nError: No previous ID for node %s in indexed value update\n",
						update_com.ID);
				exit(-1);
			} 
			put_bit(nodeIDbits, id);

			/* figure out which binary reference for this field */
			for (the_field = 0; 
				the_field < nodes[the_node].num_in_fields && 
				/* find the field in the list of all input fields */
				strcasecmp(update_com.field, 
					nodes[the_node].field_names[
						nodes[the_node].in_fields[the_field] 
					]);
				++the_field);
	
			if (the_field == nodes[the_node].num_in_fields) {
				fprintf(stderr,
				"\nUnrecognized field '%s' in update near line %d.\n",
					update_com.field, line);
				exit(-1);
			}
	
			put_bit(nodes[the_node].in_field_bits, the_field);
	
			/* Now specify where to put the new element */
			position = atoi(update_com.position);
			if (position > 0) {
					if (position > 65535) {
						fprintf(stderr,"\nError:Only 16 update position bits.");
							exit(-1);
					}
					put_bit(1,0); 
					put_bit(16,position); 
			} else if (position == 0) {
					put_bit(2,2); 
			} else if (position == -1) {
					put_bit(2,3); 
			} else {
					fprintf(stderr,"\nError: Bad update position near line %d.",
							line);
					exit(-1);
			}
			read_value(nodes[the_node].field_types[
				nodes[the_node].def_fields[the_field] 
				]);
			
		/* *******************************************************  */
		} else if (!strcasecmp(update_com.what, "route")) {
			put_bit(2,3); 
			read_route();
		} else {
			fprintf(stderr, "\nUnknown update-insert element %s.",
				 update_com.what);
			exit(-1);
		}
	/* *****************************************************************  */
	} else if (!strcasecmp(update_com.action, "delete")) {
		put_bit(2,1); 
		/* *******************************************************  */
		if (!strcasecmp(update_com.what, "node")) {
			put_bit(1,0); 
			if ((id = retrieve_id(update_com.ID, &the_node)) == -1) {
				fprintf(stderr, 
				"\nError: No previous ID for node %s in update-delete\n",
						update_com.ID);
				exit(-1);
			} 
			put_bit(nodeIDbits, id);

		/* *******************************************************  */
		} else if (!strcasecmp(update_com.what, "indexedvalue")) {
			put_bit(2,2); 
			if ((id = retrieve_id(update_com.ID, &the_node)) == -1) {
				fprintf(stderr, 
				"\nError:No previous ID for node %s in idxvalue update-delete.",
						update_com.ID);
				exit(-1);
			} 
			put_bit(nodeIDbits, id);

			/* figure out which binary reference for this field */
			for (the_field = 0; 
				the_field < nodes[the_node].num_def_fields && 
				/* find the field in the list of all input fields */
				strcasecmp(update_com.field, 
					nodes[the_node].field_names[
						nodes[the_node].def_fields[the_field] 
					]);
				++the_field);
	
			if (the_field == nodes[the_node].num_def_fields) {
				fprintf(stderr,
				"\nUnrecognized field '%s' in update near line %d.\n",
					update_com.field, line);
				exit(-1);
			}
	
			put_bit(nodes[the_node].def_field_bits, the_field);
	
			/* Now specify where to put the new element */
			position = atoi(update_com.position);
			if (position > 0) {
					if (position > 65535) {
						fprintf(stderr,
							"\nError: Only 16 replace position bits.");
							exit(-1);
					}
					put_bit(1,0); 
					put_bit(16,position); 
			} else if (position == 0) {
					put_bit(2,2); 
			} else if (position == -1) {
					put_bit(2,3); 
			} else {
					fprintf(stderr,"\nError: Bad update position near line %d.",
							line);
					exit(-1);
			}

		/* *******************************************************  */
		} else if (!strcasecmp(update_com.what, "route")) {
			put_bit(2,3); 
			id = retrieve_id(update_com.ID, &the_node);
			put_bit(routeIDbits, id);
		} else {
			fprintf(stderr, "\nUnknown update-delete element %s.", 
					update_com.what);
			exit(-1);
		}
	/* *****************************************************************  */
	} else if (!strcasecmp(update_com.action, "replace")) {
		put_bit(2,2); 
		/* *******************************************************  */
		if (!strcasecmp(update_com.what, "node")) {
			put_bit(2,0); 
			if ((id = retrieve_id(update_com.ID, &the_node)) == -1) {
				fprintf(stderr, 
				"\nError: No previous ID for node %s in update-replace\n",
						update_com.ID);
				exit(-1);
			} 
			put_bit(nodeIDbits, id);
			parse_node(SFWorldNodeContext);
		/* *******************************************************  */
		} else if (!strcasecmp(update_com.what, "field")) {
			put_bit(2,1); 
			if ((id = retrieve_id(update_com.ID, &the_node)) == -1) {
				fprintf(stderr, 
				"\nError: No previous ID for node %s in update-replace\n",
						update_com.ID);
				exit(-1);
			} 
			put_bit(nodeIDbits, id);

			/* figure out which binary reference for this field */
			for (the_field = 0; 
				the_field < nodes[the_node].num_def_fields && 
				/* find the field in the list of all input fields */
				strcasecmp(update_com.field, 
					nodes[the_node].field_names[
						nodes[the_node].def_fields[the_field] 
					]);
				++the_field);
	
			if (the_field == nodes[the_node].num_def_fields) {
				fprintf(stderr,
				"\nUnrecognized field '%s' in update near line %d.\n",
					update_com.field, line);
				exit(-1);
			}
	
			put_bit(nodes[the_node].def_field_bits, the_field);

			read_value(nodes[the_node].field_types[
				nodes[the_node].def_fields[the_field] 
				]);

		/* *******************************************************  */
		} else if (!strcasecmp(update_com.what, "indexedvalue")) {
			put_bit(2,2); 
			if ((id = retrieve_id(update_com.ID, &the_node)) == -1) {
				fprintf(stderr, 
				"\nError: No previous ID for node %s in update-replace\n",
						update_com.ID);
				exit(-1);
			} 
			put_bit(nodeIDbits, id);

			/* figure out which binary reference for this field */
			for (the_field = 0; 
				the_field < nodes[the_node].num_def_fields && 
				/* find the field in the list of all input fields */
				strcasecmp(update_com.field, 
					nodes[the_node].field_names[
						nodes[the_node].def_fields[the_field] 
					]);
				++the_field);
	
			if (the_field == nodes[the_node].num_def_fields) {
				fprintf(stderr,
				"\nUnrecognized field '%s' in update near line %d.\n",
					update_com.field, line);
				exit(-1);
			}
	
			put_bit(nodes[the_node].def_field_bits, the_field);

			/* Now specify where to put the new element */
			position = atoi(update_com.position);
			if (position > 0) {
					if (position > 65535) {
						fprintf(stderr,"\nError:Only 16 update position bits.");
							exit(-1);
					}
					put_bit(1,0); 
					put_bit(16,position); 
			} else if (position == 0) {
					put_bit(2,2); 
			} else if (position == -1) {
					put_bit(2,3); 
			} else {
					fprintf(stderr,"\nError: Bad update position near line %d.",
							line);
					exit(-1);
			}
		
			read_value(nodes[the_node].field_types[
				nodes[the_node].def_fields[the_field] 
				]);
		/* *******************************************************  */
		} else if (!strcasecmp(update_com.what, "route")) {
			put_bit(2,2); 
			fprintf(stderr, "\n Route replacement not 100% complient.");
			read_route();
		} else {
			fprintf(stderr, "\nUnknown update-replace element %s.", 
					update_com.what);
			exit(-1);
		}
	/* *****************************************************************  */
	} else if (!strcasecmp(update_com.action, "replacescene")) {
		put_bit(2,3); 
		parse_node(SFTopNodeContext);
	} else if (!strcasecmp(update_com.action, "repeatscene")) {
		printf("\n Repeat Scene not supported by MPEG4.");
	} else {
		fprintf(stderr, "\n Unknown update field %s.", update_com.action);
		exit(-1);
	}


}

/* ************************************************** */
/* Other routines */
/* ************************************************** */

int isspecial(char c)
{
	if (c==',' || c=='{' || c=='}' || c=='[' || c==']') 
		return(1); 
	else 
		return(0);
}

/* Universal tokenizer - works well enough for now */
void get_token(char *token) 
{
	char c, *ptr;

	for (c = getc(inf); isspace(c) && !feof(inf); c = getc(inf)) 
		if (c == '\n') ++line;

	if (feof(inf)) {
		token[0] = 0;
		return;
	}

	if (isspecial(c)) {
		token[0] = c;
		token[1] = 0;
		return;
	}

	ptr = token;
	do {
		*ptr = c;
		++ptr;
		c=getc(inf); 
	} while (!isspace(c) && !isspecial(c)  && !feof(inf)); 
	
	if (c == '\n') ++line;
	if (isspecial(c)) ungetc(c, inf);
	*ptr = 0;
	/* printf("\n In get_token : %s", token); */
}

void peek_token(char *token) /* read token but don't advance stream */
{
	char c, *ptr;
	long pos;

	pos = ftell(inf);
	for (c = getc(inf); isspace(c) && !feof(inf); c = getc(inf)) ;


	if (isspecial(c)) {
		token[0] = c;
		token[1] = 0;
		fseek(inf, pos, SEEK_SET);
		return;
	}

	ptr = token;
	do {
		*ptr = c;
		++ptr;
		c=getc(inf); 
	} while (!isspace(c) && !isspecial(c)  && !feof(inf)); 
	if (isspecial(c)) ungetc(c, inf);
	*ptr = 0;
	fseek(inf, pos, SEEK_SET);
}

void get_string(char *token)  /* reads between "" */
{
	char c, *ptr;

	for (c = getc(inf); isspace(c) && !feof(inf); c = getc(inf)) 
		if (c == '\n') ++line;

	if (feof(inf)) {
		token[0] = 0;
		return;
	}

	if (c != '"') {
		fprintf(stderr, "\nError: String data must begin with a quote.\n");
		exit(-1);
	} else 
		c = getc(inf);

	ptr = token;
	do {
		*ptr = c;
		++ptr;
		c=getc(inf); 
	} while (c != '"' && !feof(inf));
	*ptr = 0;
	/* printf("\n In get_token : %s", token); */
}

void store_id(char *id,int type,int nodeID) /* store id into list of ID numbers */
{
	node = (struct list *)malloc(sizeof (struct list));
	node->next = head;
	node->id = (char *) malloc(1+sizeof(char)*strlen(id));
	strcpy(node->id, id);
	node->type = type;
	node->num = nodeID;
	head = node;
}

int retrieve_id(char *token, int *type)
{
	for (node = head; 
		node != NULL && strcasecmp(token,node->id); node=node->next) ;


	if (node == NULL) 
		return -1;
	else {
		*type =  node->type;
		return (node->num);
	}
}


/* separete a node.field token into the node and field parts. */

void separate(char *token, char *part1, char *part2)

// Changed 1/19/2000: Why this is declared this way, I don't know
//separate(token, part1, part2)
//char *token, *part1, *part2;
{
	int i, j;
	for (i=0; i<strlen(token) && token[i] != '.'; ++i)
		part1[i] = token[i];
	part1[i] = 0;

	/* check for "set_" prefix  */
	/* if (strstr(token, "set_")) i+= 4; */

	for (j=0, ++i; i<strlen(token); ++i)
		part2[j++] = token[i];
	part2[j] = 0;

	/* check for "_changed" postfix  */
	/* if (strstr(part2, "_changed")) part2[j-8] = 0; */
}
