
/* $Id: sa_encoder.lex,v 1.1.1.1 2003/10/23 13:59:15 epfl Exp $ */
/* $Log: sa_encoder.lex,v $
/* Revision 1.1.1.1  2003/10/23 13:59:15  epfl
/* tag_initial-import
/*
# Revision 1.2  1998/05/06  21:44:26  eds
# FCD version.
#
# Revision 1.1  1997/11/10  22:58:10  eds
# Initial revision
# */
/*********************************************************************

ISO_HEADER_START

This software module was originally developed by

  Eric Scheirer (MIT Media Laboratory)

in the course of development of the MPEG-4 standard.
This software module is an implementation of a part of one or more
MPEG-4 tools as specified by the MPEG-4 standard.  ISO/IEC gives users
of the MPEG-4 standard free license to this software module or
modifications thereof for use in hardware or software products
claiming conformance to MPEG-4.  Those intending to use this software
module in hardware or software products are advised that its use may
infringe existing patents.  The original developer of this software
module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software
module or modifications thereof in an implementation.  Copyright is
not released for non MPEG-4 conforming products. The MIT Media
Laboratory retains full right to use the code for its own purpose,
assign or donate the code to a third party and to inhibit third
parties from using the code for non MPEG-4 conforming products.  This
copyright notice must be included in all copies or derivative
works. Copyright (c) 1998.

ISO_HEADER_END

***********************************************************************/

STRCONST	\"(\\.|[^\\"])*\"
IDENT		[a-zA-Z_][a-zA-Z0-9_]*
INTGR		[0-9]+
NUMBER		[0-9]+(\.[0-9]*)?(e[-+]?[0-9]+)?|-?\.[0-9]*(e-+?[0-9]+)?

%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "saol.tab.h"
  
void count(void),comment(void),dumpline(int i);
int yyline=1,yycol=0;
char thisline[1000],**all_lines = NULL;
int yywrap(void) { return(1); }

%}


%%

"//"		{ comment(); }
"aopcode"	{ count(); return(AOPCODE) ; }
"asig"		{ count(); return(ASIG) ; }
"else"          { count(); return(ELSE) ; }
"exports"       { count(); return(EXPORTS) ; }
"extend"        { count(); return(EXTEND) ; }
"global"        { count(); return(GLOBAL) ; }
"if"            { count(); return(IF) ; }
"imports"       { count(); return(IMPORTS); }
"inchannels"    { count(); return(INCHANNELS) ; }
"instr"         { count(); return(INSTR); }
"iopcode"	{ count(); return(IOPCODE); }
"ivar"		{ count(); return(IVAR) ; }
"kopcode"	{ count(); return(KOPCODE); }
"krate"         { count(); return(KRATE) ; }
"ksig"		{ count(); return(KSIG) ; }
"map"	        { count(); return(MAP) ; }
"oparray"       { count(); return(OPARRAY) ; }
"opcode"	{ count(); return(OPCODE) ; }
"outbus"	{ count(); return(OUTBUS) ; }
"outchannels"   { count(); return(OUTCHANNELS) ; }
"output"	{ count(); return(OUTPUT) ; }
"return"	{ count(); return(RETURN) ; }
"route"         { count(); return(ROUTE) ; }
"send"          { count(); return(SEND) ; }
"sequence"      { count(); return(SEQUENCE) ; }
"sasbf"         { count(); return(SASBF) ; }
"spatialize"    { count(); return(SPATIALIZE) ; }
"srate"		{ count(); return(SRATE); }
"table"		{ count(); return(TABLE); }
"tablemap"      { count(); return(TABLEMAP); }
"template"	{ count(); return(TEMPLATE); }
"turnoff"       { count(); return(TURNOFF); }
"while"		{ count(); return(WHILE); }
"with"          { count(); return(WITH); }
"xsig"		{ count(); return(XSIG) ; }
"&&"		{ count(); return(AND); }
"||"		{ count(); return(OR); }
">="		{ count(); return(GEQ); }
"<="		{ count(); return(LEQ); }
"!="		{ count(); return(NEQ); }
"=="		{ count(); return(EQEQ); }
"-"		{ count(); return(MINUS); }
"*"		{ count(); return(STAR); }
"/"		{ count(); return(SLASH); }
"+"		{ count(); return(PLUS); }
">"		{ count(); return(GT); }
"<"		{ count(); return(LT); }
"?"		{ count(); return(Q); }
":"		{ count(); return(COL); }
"("		{ count(); return(LP); }
")"		{ count(); return(RP); }
"{"		{ count(); return(LC); }
"}"		{ count(); return(RC); }
"["		{ count(); return(LB); }
"]"		{ count(); return(RB); }
";"		{ count(); return(SEM); }
","		{ count(); return(COM); }
"="		{ count(); return(EQ); }
"!"             { count(); return(NOT); }

{STRCONST}	{ count(); yytext[strlen(yytext)-1] = 0; yytext; /* strip quotes */
                           return(STRCONST); }
{IDENT}		{ count(); return(IDENT) ; }
{INTGR}		{ count(); return(INTGR) ; }
{NUMBER}	{ count(); return(NUMBER) ; }
[ \t\n]		{ count(); }
.		{ count();
                  printf("Line %d: Unknown character: '%s'\n",yyline,yytext); }

%%

void comment() {
  char c;

  while ((c = input()) != '\n'); /* skip */
  yyline++;
  thisline[0] = 0;
  yycol = 0;
  }
              

void count() {
  char *nl;
  
  if (strcmp(yytext,"\n")) {
    yycol += strlen(yytext);
    strcat(thisline,yytext);
    }
  else {
    int i=0;

    while (thisline[i] == ' ' || thisline[i] == '\t') i++;
    nl = strdup(&thisline[i]);
   
    yycol = 0;
    if (all_lines)
      all_lines = (char **)realloc(all_lines,(yyline +2) * sizeof(char *));
    else
      all_lines = (char **)malloc(2 * sizeof(char *));
    all_lines[yyline-1] = nl; /* since yyline starts from 1 */
    yyline++;
    thisline[0] = 0;
    all_lines[yyline-1] = thisline;
    }
  }

void dumpline(int i) {
 
  printf("%s\n",all_lines[i]);
}
