/* $Id: saol.yacc,v 1.1.1.1 2003/10/23 13:59:15 epfl Exp $ */
/* $Log: saol.yacc,v $
/* Revision 1.1.1.1  2003/10/23 13:59:15  epfl
/* tag_initial-import
/*
# Revision 1.8  1998/05/06  21:43:51  eds
# FCD version.
#
# Revision 1.7  1997/11/20  18:35:05  eds
# WIN32 compatibility -- removed <unistd.h> reference.
#
# Revision 1.6  1997/11/05  20:11:01  eds
# Added tablemaps.
#
# Revision 1.5  1997/11/05  14:58:29  eds
# Added vector parameters/return values/operators.
#
# Revision 1.4  1997/10/03  15:10:48  eds
# Added sfsynth statement.
#
# Revision 1.3  1997/10/01  14:53:24  eds
# Can have table declarations in opcodes.
#
# Revision 1.2  1997/09/17  14:10:21  eds
# Added spatialize statement.
# */
/*********************************************************************

ISO_HEADER_START

This software module was originally developed by

  Eric Scheirer (MIT Media Laboratory)

in the course of development of the MPEG-4 standard.
This software module is an implementation of a part of one or more
MPEG-4 tools as specified by the MPEG-4 standard.  ISO/IEC gives users
of the MPEG-4 standard free license to this software module or
modifications thereof for use in hardware or software products
claiming conformance to MPEG-4.  Those intending to use this software
module in hardware or software products are advised that its use may
infringe existing patents.  The original developer of this software
module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software
module or modifications thereof in an implementation.  Copyright is
not released for non MPEG-4 conforming products. The MIT Media
Laboratory retains full right to use the code for its own purpose,
assign or donate the code to a third party and to inhibit third
parties from using the code for non MPEG-4 conforming products.  This
copyright notice must be included in all copies or derivative
works. Copyright (c) 1998.

ISO_HEADER_END

***********************************************************************/

/* This is the grammar for SAOL, written in 'yacc'.

   It has one shift/reduce conflict which arises when
   looking at table definitions.  Within the grammar,
   'table t1;' is allowed as a definition even though
   it is  a prohibited construction (it is flagged in the syntax-
   check).  So there's an ambiguity between

     table t1(...)
     table t1;

  and you don't know with one lookahead where you are when you
  get 'table' if 't1' is the next token.
  */
   
%token IDENT INTGR NUMBER STRCONST AOPCODE ELSE EXPORTS EXTEND GLOBAL
%token IF IMPORTS INCHANNELS INTERP 
%token INSTR IOPCODE IVAR TABLE KOPCODE KRATE KSIG ASIG MAP
%token OPARRAY OPCODE OUTBUS OUTCHANNELS OUTPUT ROUTE SEND SEQUENCE
%token SRATE TEMPLATE TURNOFF WHILE WITH XSIG AND OR GEQ LEQ
%token NEQ EQEQ MINUS STAR SPATIALIZE SASBF TABLEMAP
%token SLASH PLUS GT LT Q COL LP RP LC RC LB RB SEM COM EQ RETURN NOT
%token ARRAYREF OPCALL IMPEXP VARDECL NOTAG SPECIALOP PRESET
%{
#include "saol.h"
#define YYDEBUG 1
extern sa_decoder *cur_sa;
%}

%start orcfile
%right Q
%left OR
%left AND 
%left EQEQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left STAR SLASH
%right UNOT UMINUS
%token HIGHEST

%%

orcfile           : proclist 
		  /* | error { parse_error("Error in orcfile"); }  */
                  ;
proclist          : proclist instrdecl {
		      if (!cur_sa->all->il)
			 cur_sa->all->il = new_instr_list();
		      add_instr_list(cur_sa->all->il,ptr_index($2));
		      printf("Read instr %s.\n",((instr_decl *)ptr_index($2))->name);
		      }
		  | proclist opcodedecl {
		      if (!cur_sa->all->op)
		        cur_sa->all->op = new_opcode_list();
		      add_opcode_list(cur_sa->all->op,ptr_index($2));
		      printf("Read opcode %s.\n",((opcode_decl *)ptr_index($2))->name);
		      }
		  | proclist globaldecl {
		      printf("Read global block.\n");
		      }
		  | proclist templatedecl {
		      deal_with_template(cur_sa,ptr_index($2));
		      }
		  | /* null */
		  | error { parse_error("Expected 'global', 'opcode, 'template', or 'instr'."); }
		  ;
instrdecl	  : INSTR IDENT LP identlist RP miditag LC vardecls block RC {
		      instr_decl *i;

		      i = new_instr_decl(ptr_index($2),ptr_index($4),
		                         ptr_index($8),ptr_index($9),
					 ptr_index($6));
		      $$ = add_ptr_index(i);
		      }
		  | INSTR IDENT LP identlist RP miditag error
		      { parse_error("'{' expected."); }
		  | INSTR IDENT LP identlist error
		      { parse_error("')' expected."); }
		  | INSTR IDENT error
		      { parse_error("'(' expected."); }
		  | INSTR error
		      { parse_error("Instrument name expected."); }
		  ;
miditag           : PRESET int_list {
		     $$ = $2;
		     }
		  | /* null */ {
		     $$ = add_ptr_index(NULL);
		    }
		  | PRESET error { parse_error("Bad preset tag."); }
		  ;
int_list	  : int_list INTGR {
		      add_terminal_list(ptr_index($1),new_terminal(NUMBER,ptr_index($2)));
		      $$ = $1;
		    } 
		  | INTGR {
		      terminal_list *tl;
		      tl = new_terminal_list(new_terminal(NUMBER,ptr_index($1)));
		      $$ = add_ptr_index(tl);
		    }
		   ;

opcodedecl	  : optype IDENT LP paramlist RP LC opvardecls block RC {
		      opcode_decl *o;

		      o = new_opcode_decl(ptr_index($2),ptr_index($4),
		                          ptr_index($7),ptr_index($8),
					  $1);
		      $$ = add_ptr_index(o);
		      }
		  | optype IDENT LP paramlist RP error
		     { parse_error("'{' expected."); }
		  | optype IDENT error
		     { parse_error("'(' expected."); }
		  | optype error
		     { parse_error("Opcode name expected."); } 
		  ;
globaldecl	  : GLOBAL LC globalblock RC {
		      if (!cur_sa->all->g) 
		         cur_sa->all->g = new_orc_global();
		      else
		         warn_line("Using multiple global blocks.",yyline);
		      sort_global_decls(cur_sa,cur_sa->all->g,ptr_index($3));
		      }
		  | GLOBAL error { parse_error("'{' expected."); }
		  ;
templatedecl	  : TEMPLATE LT identlist GT  /* with preset */
                     PRESET mapblock
		     LP identlist RP
		     MAP LC identlist RC
		     WITH LC mapblock RC LC
		     vardecls block RC {
		       template_decl *t;

		       t = new_template_decl(ptr_index($3),ptr_index($8),
		                             ptr_index($12),ptr_index($16),
					     ptr_index($19),ptr_index($20),
					     ptr_index($6));
		       $$ = add_ptr_index(t);
		       }
		  | TEMPLATE LT identlist GT  /* no preset */
		     LP identlist RP
		     MAP LC identlist RC
		     WITH LC mapblock RC LC
		     vardecls block RC {
		       template_decl *t;

		       t = new_template_decl(ptr_index($3),ptr_index($6),
		                             ptr_index($10),ptr_index($14),
					     ptr_index($17),ptr_index($18),
					     NULL);
		       $$ = add_ptr_index(t);
		       }
		    
		   |  TEMPLATE LT identlist GT LP identlist RP
		     MAP LC identlist RC
		     WITH LC mapblock RC error
		      { parse_error("'{' expected."); }
		   |  TEMPLATE LT identlist GT LP identlist RP
		     MAP LC identlist RC
		     WITH LC mapblock error
		      { parse_error("'}' expected."); }
		   |  TEMPLATE LT identlist GT LP identlist RP
		     MAP LC identlist RC
		     WITH error
		      { parse_error("'{' expected."); }
		   |  TEMPLATE LT identlist GT LP identlist RP
		     MAP LC identlist RC
		     error
		      { parse_error("'with' expected."); }
		   |  TEMPLATE LT identlist GT LP identlist RP
		     MAP LC identlist error
		      { parse_error("'}' expected."); }
		   |  TEMPLATE LT identlist GT LP identlist RP
		     MAP error
		      { parse_error("'{' expected."); }
		   |  TEMPLATE LT identlist GT LP identlist RP error
		      { parse_error("'map' expected."); }
		   |  TEMPLATE LT identlist GT LP identlist error
		      { parse_error("')' expected."); }
		   |  TEMPLATE LT identlist GT error
		      { parse_error("'(' expected."); }
		   |  TEMPLATE LT identlist error
		      { parse_error("'>' expected."); }
		   |  TEMPLATE error
		      { parse_error("'<' expected."); }
		  ;
mapblock	  : mapblock COM LT terminal_list GT {
		       add_mapblock(ptr_index($1),ptr_index($4));
		       $$ = $1;
		       }
		  | LT terminal_list GT {
		       mapblock *map;

		       map = new_mapblock(ptr_index($2));
		       $$ = add_ptr_index(map);
		       }
		  | { parse_error("Empty map block not allowed."); }
		  | mapblock COM LT error
		  		     { parse_error("Bad list in mapblock."); }	
		  | mapblock COM LT terminal_list error
		     { parse_error("Bad list in mapblock."); }
		  | mapblock error
		     { parse_error("',' expected."); }
		  ;
terminal_list	  : terminal_list COM terminal {
		     add_terminal_list(ptr_index($1),ptr_index($3));
		     $$ = $1;
		     }
		  | terminal {
		     terminal_list *tl;
		     tl = new_terminal_list(ptr_index($1));
		     $$ = add_ptr_index(tl);
		     }
		  | terminal_list COM error
		     { parse_error("Expected another value."); }
		  ;
terminal	  : IDENT {
		     terminal *t;
		     t = new_terminal(IDENT,ptr_index($1));
		     $$ = add_ptr_index(t);
		     }
		  | const {
		     terminal *t;
		     t = new_terminal(NUMBER,ptr_index($1));
		     $$ = add_ptr_index(t);
		     }
		  | STRCONST {
		     terminal *t;
		     t = new_terminal(STRCONST, ptr_index($1));
		     $$ = add_ptr_index(t);
		     }
		  ;
globalblock	  : globalblock globaldef {
		     add_gblock(ptr_index($1),ptr_index($2));
		     }
		  | /* null */ {
		     global_block *gb;
		     
		     gb = new_gblock();
		     $$ = add_ptr_index(gb);
		     }
		  /* | error { parse_error("Bad global block."); } */
		  ;
globaldef	  : rtparam { $$ = $1; }
		  | vardecl {
		      global_decl *g;
		      g = new_global(VARDECL);
		      set_global_vardecl(g,ptr_index($1));
		      $$ = add_ptr_index(g);
		      }
		  | routedef { $$ = $1; }
		  | senddef { $$ = $1; }
		  | seqdef { $$ = $1; }
		  | error { parse_error("Bad global declaration"); } 
		  ;
rtparam		  : SRATE INTGR SEM {
		     global_decl *g;

		     g = new_global(SRATE);
		     set_global_rtparam(g,atoi(ptr_index($2)));
		     $$ = add_ptr_index(g);
		     }
		  | KRATE INTGR SEM {
		     global_decl *g;

		     g = new_global(KRATE);
		     set_global_rtparam(g,atoi(ptr_index($2)));
		     $$ = add_ptr_index(g);
		     }
		  | INCHANNELS INTGR SEM {
		     global_decl *g;

		     g = new_global(INCHANNELS);
		     set_global_rtparam(g,atoi(ptr_index($2)));
		     $$ = add_ptr_index(g);
		     }
		  | OUTCHANNELS INTGR SEM {
		     global_decl *g;

		     g = new_global(OUTCHANNELS);
		     set_global_rtparam(g,atoi(ptr_index($2)));
		     $$ = add_ptr_index(g);
		     }
		  | INTERP INTGR SEM {
		     global_decl *g;

		     g = new_global(INTERP);
		     set_global_rtparam(g,atoi(ptr_index($2)));
		     $$ = add_ptr_index(g);
		     }
		     
		  | SRATE INTGR error { parse_error("';' expected."); }
		  | KRATE INTGR error { parse_error("';' expected."); }
		  | INCHANNELS INTGR error { parse_error("';' expected."); }
		  | OUTCHANNELS INTGR error { parse_error("';' expected."); }
		  | INTERP INTGR error { parse_error("';' expected."); }
		  | SRATE error { parse_error("Sampling rate expected."); }
		  | KRATE error { parse_error("Control rate expected."); }
		  | INCHANNELS error { parse_error("# input channels expected."); }
		  | OUTCHANNELS error { parse_error("# output channels expected."); }
		  | INTERP error { parse_error("Interpolation parameter expected."); }
		  ;
routedef	  : ROUTE LP IDENT COM identlist RP SEM {
                     global_decl *g;
		     routedef *r;

		     r = new_routedef(ptr_index($3),ptr_index($5));
		     g = new_global(ROUTE);
		     set_global_routedef(g,r);
		     $$ = add_ptr_index(g);
		     }
		   | ROUTE LP IDENT COM identlist RP error
		      { parse_error("';' expected."); }
		   | ROUTE LP IDENT COM identlist error
		      { parse_error("')' expected."); }
		   | ROUTE LP IDENT error
		      { parse_error("',' expected (instrument required)"); }
		   | ROUTE error
		      { parse_error("'(' expected."); }
		   
senddef		  : SEND LP IDENT SEM exprlist SEM namelist RP SEM {
		     global_decl *g;
		     senddef *s;

		     s = new_senddef(ptr_index($3),ptr_index($5),
		                     ptr_index($7));
		     g = new_global(SEND);
		     set_global_senddef(g,s);
		     $$ = add_ptr_index(g);
		     }
		   | SEND LP IDENT SEM exprlist SEM namelist error 
		     { parse_error("')' expected."); }
		   | SEND LP IDENT SEM exprlist error
		     { parse_error("';' expected."); }
		   | SEND LP IDENT error
		     { parse_error("';' expected."); }
		   | SEND LP error 
		     { parse_error("Instrument expected."); }
		   | SEND error
		     { parse_error("'(' expected."); }
		  ;
seqdef		  : SEQUENCE LP identlist RP SEM {
		     global_decl *g;
		     seqdef *s;
		     
		     s = new_seqdef(ptr_index($3));
		     g = new_global(SEQUENCE);
		     set_global_seqdef(g,s);
		     $$ = add_ptr_index(g);
		     }
		   | SEQUENCE LP identlist RP error
		      { parse_error("';' expected."); }
		   | SEQUENCE LP error
		      { parse_error("List of instruments expected."); }
		   | SEQUENCE error
		      { parse_error("'(' expected."); }
		  ;
block		  : block statement {
		      add_block(ptr_index($1),ptr_index($2));
		      $$ = $1;
		      }
		  | /* null */ {
		      block *b;
		      long i;

		      b = new_block();
		      i = add_ptr_index(b);
		      $$ = i;
		      }
		  ;
statement	  : lvalue EQ expr SEM {
		      statement *st;
		      long i;
		    
 		      st = new_statement(EQ,ptr_index($3),NULL);
		      set_statement_lvalue(st,ptr_index($1));
		      i = add_ptr_index(st);
		      $$ = i;
		      }
		  | lvalue EQ expr error { parse_error("';' expected."); }

		  | expr SEM { /* null assign */
		      statement *st;
		      st = new_statement(EQ,ptr_index($1),NULL);
		      $$ = add_ptr_index(st);
		      }
		  | expr error { parse_error("';' expected."); }
		  | IF LP expr RP LC block RC {
		      statement *st;
		      long i;

		      st = new_statement(IF,ptr_index($3),ptr_index($6));
		      i = add_ptr_index(st);
		      $$ = i;
		      }
		  | IF LP expr RP error { parse_error("'{}' required for 'if'."); }
		  | IF LP expr error { parse_error("')' expected."); }
		  | IF error { parse_error("'(' expected."); }
		  | IF LP expr RP LC block RC ELSE LC block RC {
		      statement *st;
		      long i;

		      st = new_statement(ELSE,ptr_index($3),ptr_index($6));
		      set_statement_else(st,ptr_index($10));
		      i = add_ptr_index(st);
		      $$ = i;
		      }
		  | IF LP expr RP LC block RC ELSE error
		     { parse_error("'{}' required for 'else'."); }
		  | WHILE LP expr RP LC block RC {
		      statement *st;
		      long i;

		      st = new_statement(WHILE,ptr_index($3),ptr_index($6));
		      i = add_ptr_index(st);
		      $$ = i;
		      }
		  | WHILE LP expr RP error { parse_error("'{}' required for 'while'."); }
		  | WHILE LP expr error { parse_error("')' expected."); }
		  | WHILE error {  parse_error("'(' expected."); }
		  | INSTR IDENT LP exprlist RP SEM {
		      statement *st;
		      
		      st = new_statement(INSTR,NULL,NULL);
		      set_statement_iname(st,ptr_index($2));
		      set_statement_params(st,ptr_index($4));
		      $$ = add_ptr_index(st);
		      }
		  | INSTR IDENT LP exprlist RP error { parse_error("';' expected."); }
		  | INSTR IDENT LP exprlist error { parse_error("')' expected."); }
		  | INSTR IDENT error { parse_error("'(' expected."); }
		  | INSTR error { parse_error("Instrument name expected."); }
		  | OUTPUT LP exprlist RP SEM {
		      statement *st;
		      long i;

		      st = new_statement(OUTPUT,NULL,NULL);
		      set_statement_params(st,ptr_index($3));
		      i = add_ptr_index(st);
		      $$ = i;
		      }
		  | OUTPUT LP exprlist RP error { parse_error("';' expected."); }
		  | OUTPUT LP exprlist error { parse_error("')' expected."); }
		  | OUTPUT error { parse_error("'(' expected."); }
  		  | SPATIALIZE LP exprlist RP SEM {
		      statement *st;
		      st = new_statement(SPATIALIZE, NULL, NULL);
		      set_statement_params(st,ptr_index($3));
		      $$ = add_ptr_index(st);
		      }
		  | SPATIALIZE LP exprlist RP error { parse_error("';' expected."); }
		  | SPATIALIZE LP exprlist error { parse_error("')' expected."); }
		  | SPATIALIZE error { parse_error("'(' expected."); }
		  | OUTBUS LP IDENT COM exprlist RP SEM {
		      statement *st;
		      long i;

      		      st = new_statement(OUTBUS,NULL,NULL);
		      set_statement_bus(st,ptr_index($3));
		      set_statement_params(st,ptr_index($5));
		      i = add_ptr_index(st);
		      $$ = i;
		      }
		  | OUTBUS LP IDENT COM exprlist RP error { parse_error("';' expected."); }
		  | OUTBUS LP IDENT COM exprlist error { parse_error("')' expected."); }
		  | OUTBUS LP IDENT error { parse_error("',' expected."); }
		  | OUTBUS LP error { parse_error("Bus name expected."); }
		  | OUTBUS error { parse_error("'(' expected."); }
		  | EXTEND LP expr RP SEM {
		      statement *st;
		      st = new_statement(EXTEND,ptr_index($3),NULL);
		      $$ = add_ptr_index(st);
		      }
		  | EXTEND LP expr RP error { parse_error("';' expected."); }
		  | EXTEND LP expr error { parse_error("')' expected."); }
		  | EXTEND error { parse_error("'(' expected."); }
		  | TURNOFF SEM {
		      statement *st;
		      st = new_statement(TURNOFF,NULL,NULL);
		      $$ = add_ptr_index(st);
		      }
  		  | TURNOFF error { parse_error("';' expected."); } 
		  | RETURN LP exprlist RP SEM { 
		      statement *st;
		      long i;

		      st = new_statement(RETURN,NULL,NULL);
		      st->params = ptr_index($3);
		      i = add_ptr_index(st);
		      $$ = i;
		      }
		  | RETURN LP exprlist RP error { parse_error("';' expected."); }
		  | RETURN LP exprlist error { parse_error("')' expected."); }
		  | RETURN error { parse_error("'(' expected."); }
		  | AOPCODE error { parse_error("No matching '}' for previous block"); }
		  | IOPCODE error { parse_error("No matching '}' for previous block"); }
		  | OPCODE error { parse_error("No matching '}' for previous block"); }
		  | KOPCODE error { parse_error("No matching '}' for previous block"); }
		  | GLOBAL error { parse_error("No matching '}' for previous block"); }
		  ;
lvalue		  : IDENT {
		      terminal *t;
		      expr *p;

		      t = new_terminal(IDENT,ptr_index($1));
		      p = new_expr(NULL,NULL,NULL,IDENT);
		      set_expr_data(p,t);
		      $$ = add_ptr_index(p);
		      }
		  | IDENT LB expr RB {
		      terminal *t;
		      expr *p;

		      t = new_terminal(IDENT,ptr_index($1));
		      p = new_expr(ptr_index($3),NULL,NULL,ARRAYREF);
		      set_expr_data(p,t);
		      $$ = add_ptr_index(p);
		      }
		   | IDENT LB expr error { parse_error("']' expected."); }
		   | error { parse_error("Bad lvalue in expression"); }
		  ;
identlist	  : identlist COM IDENT {
		      name *v;
		      v = new_name(ptr_index($3),1);
		      add_namelist(ptr_index($1),v);
		      $$ = $1;
		      }
		  | IDENT {
		      name *v;
		      namelist *nl;

		      v = new_name(ptr_index($1),1);
		      nl = new_namelist(v);
		      $$ = add_ptr_index(nl);
		      }
		  | /* null */ {
		      namelist *nl;

		      nl = new_namelist(NULL);
		      $$ = add_ptr_index(nl);
		      }
		  | identlist COM error { parse_error("Expected another identifier."); }
		  ;
paramlist	  : paramlist COM paramdecl {
		      add_formalparam_list(ptr_index($1),ptr_index($3));
		      $$ = $1;
		      }
		  | paramdecl {
		      formalparam_list *fpl;

		      fpl = new_formalparam_list();
		      add_formalparam_list(fpl,ptr_index($1));
		      $$ = add_ptr_index(fpl);
		      }
		  |   /* null */ {
		      formalparam_list *fpl;

		      fpl = new_formalparam_list();
		      $$ = add_ptr_index(fpl);
		      }
		  | paramlist COM error { parse_error("Bad parameter in list."); }
		  ;
vardecls	  : vardecls vardecl {
		     add_vardecl_list(ptr_index($1),ptr_index($2));
		     $$ = $1;
		     }
		  | /* null */ {
		     vardecl_list *vdl;

		     vdl = new_vardecl_list();
		     $$ = add_ptr_index(vdl);
		     }
		  ;
vardecl		  : taglist stype namelist SEM { 
		     vardecl *v;

		     v = new_vardecl($2,$1,ptr_index($3));
		     $$ = add_ptr_index(v);
		     }
		  | taglist stype namelist error { parse_error("';' expected."); }
		  | taglist stype error { parse_error("Variable name(s) expected."); }
		  | taglist error { parse_error("Type name expected."); }
		  | stype namelist SEM {
		     vardecl *v;

		     v = new_vardecl($1,NOTAG,ptr_index($2));
		     $$ = add_ptr_index(v);
		     }
		  | stype namelist error { parse_error("';' expected."); }
		  | stype error { parse_error("Variable name(s) expected."); }
		  | tabledecl SEM {
		     vardecl *v;
		     v = new_vardecl(TABLE,NOTAG,NULL);
		     set_vardecl_tabledef(v,ptr_index($1));
		     $$ = add_ptr_index(v);
		     }
		  | tabledecl error { parse_error("';' expected."); }
		  | TABLEMAP IDENT LP identlist RP SEM {
		     vardecl *v;
		     v = new_vardecl(TABLEMAP,NOTAG,ptr_index($4));
		     v->tablemap_name =  ptr_index($2);
		     $$ = add_ptr_index(v);
		     }
		  | TABLEMAP IDENT LP identlist RP error { parse_error("';' expected."); }
		  | TABLEMAP IDENT LP identlist error { parse_error("')' expected."); }
		  | TABLEMAP IDENT error { parse_error("'(' expected."); }
		  | TABLEMAP error { parse_error("Table map name expected."); }
		  ;
opvardecls	  : opvardecls opvardecl {
		     add_vardecl_list(ptr_index($1),ptr_index($2));
		     $$ = $1;
		     }
		  | /* null */ {
		     vardecl_list *vdl;

		     vdl = new_vardecl_list();
		     $$ = add_ptr_index(vdl);
		     }
		  /* | error { parse_error("Bad declaration list."); } */
		  ;
opvardecl	  : taglist otype namelist SEM {
                     vardecl *v;

		     v = new_vardecl($2,$1,ptr_index($3));
		     $$ = add_ptr_index(v);
		     }
		  | taglist otype namelist error { parse_error("';' expected."); }
		  | taglist otype error { parse_error("Variable name(s) expected."); }
		  | taglist error { parse_error("Type name expected."); }
		  | otype namelist SEM {
		     vardecl *v;
		     v = new_vardecl($1,NOTAG,ptr_index($2));
		     $$ = add_ptr_index(v);
		     }
		  | otype namelist error { parse_error("';' expected."); }
		  | otype error { parse_error("Variable name(s) expected."); }
		  | tabledecl SEM {
		     vardecl *v;
		     v = new_vardecl(TABLE,NOTAG,NULL);
		     set_vardecl_tabledef(v,ptr_index($1));
		     $$ = add_ptr_index(v);
		     }
		  | tabledecl error { parse_error("';' expected."); }
		  ;
paramdecl	  : otype name {
		     formalparam *fp;

		     fp = new_formalparam($1,ptr_index($2));
		     $$ = add_ptr_index(fp);
		     }
		  | otype error { parse_error("Parameter name expected."); }
		  ;
namelist	  : namelist COM name {
		      add_namelist(ptr_index($1),ptr_index($3));
		      $$ = $1;
		      }
		  | name {
		      namelist *v;
		      v = new_namelist(ptr_index($1));
		      $$ = add_ptr_index(v);
		      }
		  | namelist COM error { parse_error("Variable name expected."); }
		  ;
name		  : IDENT {
		      name *v;
		      v = new_name(ptr_index($1),1);
		      $$ = add_ptr_index(v);
		      }
		  | IDENT LB INTGR RB {
					/* add expressions here */
		      name *v;
		      v = new_name(ptr_index($1),atoi(ptr_index($3)));
		      v->det = 1;
		      $$ = add_ptr_index(v);
		      }
		  | IDENT LB INCHANNELS RB {
		      name *v;
		      v = new_name(ptr_index($1),-1);
		      $$ = add_ptr_index(v);
		      }
		  | IDENT LB OUTCHANNELS RB {
		      name *v;
		      v = new_name(ptr_index($1),-2);
		      $$ = add_ptr_index(v);
		      }
		  | IDENT LB expr error { parse_error("']' expected."); }
		  ;
stype		  : IVAR { $$ = IVAR; }
		  | KSIG { $$ = KSIG; }
		  | ASIG { $$ = ASIG; }
		  | TABLE { $$ = TABLE; }
		  | OPARRAY { $$ = OPARRAY; }
		  ;
otype		  : XSIG { $$ = XSIG; }
		  | stype { $$ = $1; }
		  ;
tabledecl	  : TABLE IDENT LP IDENT COM exprstrlist RP {
		      tabledef *t;

		      t = new_tabledef(ptr_index($2),ptr_index($4),
		                        ptr_index($6));
		      $$ = add_ptr_index(t);
		      }
		  | TABLE IDENT LP IDENT COM exprstrlist error
		     { parse_error("')' expected."); }
		  | TABLE IDENT LP IDENT error { parse_error("',' expected."); }
		  | TABLE IDENT LP error { parse_error("Table generator expected."); }
		  | TABLE IDENT error { parse_error("'(' expected."); }
		  ;
taglist		  : IMPORTS { $$ = IMPORTS; } 
		  | EXPORTS { $$ = EXPORTS; } 
		  | IMPORTS EXPORTS { $$ = IMPEXP; }
		  | EXPORTS IMPORTS { $$ = IMPEXP; }
		  ;
optype		  : AOPCODE { $$ = AOPCODE; } 
		  | KOPCODE { $$ = KOPCODE; } 
		  | IOPCODE { $$ = IOPCODE; } 
		  | OPCODE { $$ = OPCODE; } 
		  ;
expr		  : IDENT {
		      expr *p;
		      terminal *t;
		      long i;

		      p = new_expr(NULL,NULL,NULL,IDENT);
		      t = new_terminal(IDENT,ptr_index($1));
		      set_expr_data(p,t);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		        
		  | const {
		      long i;
		      expr *p;
		      terminal *t;

		      p = new_expr(NULL,NULL,NULL,NUMBER);
		      t = new_terminal(NUMBER,ptr_index($1));
		      set_expr_data(p,t);

		      i = add_ptr_index(p);
		      $$ = i;
		      }
		      		  
		  | IDENT LB expr RB { /* array reference */
		      long i;
		      expr *p;
		      terminal *t;

		      t = new_terminal(IDENT,ptr_index($1));
		      p = new_expr(ptr_index($3),NULL,NULL,ARRAYREF);
		      set_expr_data(p,t);

		      i = add_ptr_index(p);
		      $$ = i;
		      }
		      
		  | IDENT LB expr error { parse_error("']' expected."); }
		  | IDENT LB error { parse_error("Expression expected."); }
		  | SASBF LP exprlist RP { /* SASBF synthesis */
		      long i;
		      expr *p;
		      terminal *t;

		      p = new_expr(NULL,NULL,NULL,SASBF);
		      set_expr_params(p,ptr_index($3));
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | SASBF LB error { parse_error("'sasbf' is not an opcode, and may not be used in an opcode array."); }
		  | IDENT LP exprlist RP { /* opcode call */
		      long i;
		      expr *p;
		      terminal *t;

		      t = new_terminal(IDENT,ptr_index($1));
		      p = new_expr(NULL,NULL,NULL,OPCALL);
		      set_expr_params(p,ptr_index($3));
		      set_expr_data(p,t);

		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | IDENT LP error { parse_error("Expression expected."); }
		  | IDENT LB expr RB LP exprlist RP { /* oparray call */
		      expr *p;
                      terminal *t;

                      t = new_terminal(IDENT,ptr_index($1));
		      p = new_expr(ptr_index($3),NULL,NULL,OPARRAY);
		      set_expr_params(p,ptr_index($6));
		      set_expr_data(p,t);
		      $$ = add_ptr_index(p);
		      }
		      
		  | IDENT LB expr RB LP error { parse_error("Expression expected."); }
		  
		  | expr Q expr COL expr %prec Q { /* a ? b : c */
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),ptr_index($5),Q);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr Q expr COL error { parse_error("Expression expected."); }
		  | expr Q expr error { parse_error("':' expected."); }
		  | expr Q error { parse_error("Expression expected."); }
		  
		  | expr LEQ expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,LEQ);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr LEQ error { parse_error("Expression expected."); }
		  
		  | expr GEQ expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,GEQ);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr GEQ error { parse_error("Expression expected."); }  

		  | expr NEQ expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,NEQ);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr NEQ error { parse_error("Expression expected."); }  

		  | expr EQEQ expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,EQEQ);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr EQEQ error { parse_error("Expression expected."); }  

		  | expr GT expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,GT);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr GT error { parse_error("Expression expected."); }  

		  | expr LT expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,LT);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr LT error { parse_error("Expression expected."); }  

		  | expr AND expr    {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,AND);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr AND error { parse_error("Expression expected."); }  

		  | expr OR expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,OR);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr OR error { parse_error("Expression expected."); }  

		  | expr PLUS expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,PLUS);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr PLUS error { parse_error("Expression expected."); }  

		  | expr MINUS expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,MINUS);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr MINUS error { parse_error("Expression expected."); }  

		  | expr STAR expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,STAR);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr STAR error { parse_error("Expression expected."); }  

		  | expr SLASH expr {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($1),ptr_index($3),NULL,SLASH);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | expr SLASH error { parse_error("Expression expected."); }
		  
		  | NOT expr     %prec UNOT    {
		      expr *p;

		      p = new_expr(ptr_index($2),NULL,NULL,NOT);
		      $$ = add_ptr_index(p);
		      }
		  | NOT error { parse_error("Expression expected."); }

		  | MINUS expr  {
		      long i;
		      expr *p;

		      p = new_expr(ptr_index($2),NULL,NULL,UMINUS);
		      i = add_ptr_index(p);
		      $$ = i;
		      }
		  | MINUS error { parse_error("';' expected."); }

		  | LP expr RP {
		     long i;
		     expr *p;
		     p = new_expr(ptr_index($2),NULL,NULL,LP);
		     i = add_ptr_index(p);
		     $$ = i;
		     }
		  | LP expr error { parse_error("')' expected."); }
		  | LP error { parse_error("Expression expected."); }
		  ;
exprlist          : exprlist COM expr {
		      add_exprlist(ptr_index($1),ptr_index($3));
		      $$ = $1;
		      }
		  | exprlist COM error { parse_error("Expression expected."); }
		  | expr {
		      exprlist *e;
		      long i;
		      e = new_exprlist(ptr_index($1));
		      i = add_ptr_index(e);
		      $$ = i;
		      }
		  | /* null */ {
		      exprlist *e;
		      long i;
		      e = new_exprlist(NULL);
		      i = add_ptr_index(e);
		      $$ = i;
		      }
		  ;
exprstrlist	  : exprstrlist COM expr {
		     add_exprlist(ptr_index($1),ptr_index($3));
		     $$ = $1;
		     }
		  | exprstrlist COM STRCONST {
		     terminal *t;
		     expr *p;
		     t = new_terminal(STRCONST,ptr_index($3));
		     p = new_expr(NULL,NULL,NULL,STRCONST);
		     set_expr_data(p,t);
		     add_ptr_index(p);
		     add_exprlist(ptr_index($1),p);
		     $$ = $1;
		     }
		  | exprstrlist COM error { parse_error("Expression or string constant expected."); }
		  | STRCONST {
		     exprlist *e;
		     terminal *t;
		     expr *p;
		     long i;
		     t = new_terminal(STRCONST,ptr_index($1));
		     p = new_expr(NULL,NULL,NULL,STRCONST);
		     set_expr_data(p,t);
		     add_ptr_index(p);
		     e = new_exprlist(p);
		     i = add_ptr_index(e);
		     $$ = i;
		     }		    
		  | expr {
		     exprlist *e;
		     long i;
		     e = new_exprlist(ptr_index($1));
		     i = add_ptr_index(e);
		     $$ = i;
		     }
		  ;		 
const   	  : INTGR { $$ = $1; }
		  | NUMBER { $$ = $1; }
		  ;
%%

#include "saol.yy.c"

void parse_error(char *status) {
  int i;
  char c;
  
  printf("Line %d: %s\n", yyline,status);
  printf("%s\n",thisline);
  for (i=0;i<yycol-strlen(yytext) && i < 100;i++) printf(" ");
  printf("^\n");
  cur_sa->synerror = 1;
  }

void parse_error_line(char *status,int line) {
  int i;
  char c;
  
  printf("Line %d: %s\n", line,status);
  printf("   %s\n\n",all_lines[line-1]);
  cur_sa->synerror = 1;
  }

void warn_line(char *status,int line) {
  int i;
  char c;
 
  if (line) { 
    printf("Warning: Line %d: %s\n", line,status);
    printf("   %s\n\n",all_lines[line-1]);
    }
  else printf("Warning: %s\n",status);
  }

  

void yyerror(char *foo) {}
