/*
        This software module was originally developed by Jung-Chul Lee
        (ETRI: Electronics and Telecommunications Research Institute)
        and edited by Youngjik Lee (ETRI),  in
        the course of development of the MPEG-4 Audio (ISO/IEC 14496-3).
        This software module is an implementation of a part of one or more
        MPEG-4 Audio (ISO/IEC 14496-3) tools as
        specified by the MPEG-4 Audio (ISO/IEC 14496-3). ISO/IEC gives
        users of the MPEG-4 Audio (ISO/IEC 14496-3) free license to this
        software module or modifications thereof for use in hardware or
        software products claiming conformance to the MPEG-4 Audio
        (ISO/IEC 14496-3). Those intending to use this software module in
        hardware or software products are advised that its use may infringe
        existing patents. The original developer of this software module
        and his/her company, the subsequent
        editors and their companies, and ISO/IEC have no liability for use
        of this software module or modifications thereof in an implementation.
        Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
        conforming products. ETRI retains full right to use the code for
        his/her own purpose, assign or donate the code to a
        third party and to inhibit third parties from using the code for
        non MPEG-4 Audio (ISO/IEC 14496-3) conforming products. This
        copyright notice must be included in all
        copies or derivative works. Copyright 1997, 1998, 1999.
*/ 

#include "MP4Movies.h"
#include "ttsi.h"

// ISO/IEC 14496-3(MPEG-4) Audio Object & TTS Object Test BitStream
static u8 ttsiDecoderSpecificInfo[][6] = {
	{0x64, 0x0B, 0x2C, 0x40, 0xCE, 0x1C},
	{0x64, 0x0B, 0x2C, 0x40, 0xCA, 0xDC}
};
/*
{
	{0x65, 0x88, 0x19, 0xc3, 0x80, 0x00}, // TTS1
	{0x65, 0x88, 0x18, 0x20, 0x80, 0x00}, // TTS2
	{0x65, 0x88, 0x18, 0x10, 0x80, 0x00}, // TTS3
	{0x65, 0x88, 0x18, 0x08, 0x80, 0x00}, // TTS4
	{0x65, 0x88, 0x18, 0x30, 0x80, 0x00}, // TTS5
	{0x65, 0x88, 0x18, 0x28, 0x80, 0x00}, // TTS6
	{0x65, 0x88, 0x19, 0x03, 0x80, 0x00}, // TTS7
	{0x65, 0x88, 0x19, 0x5B, 0x80, 0x00}  // TTS8
};
*/
FILE *fp;
int TTS;

MP4Err createMyMovie( char *infile, char *outfile );
MP4Err addMySamples( char *infile, MP4Track theTrack, MP4Media theMedia );
MP4Err getDecoderSpecificInfo( MP4Handle sampleH );
MP4Err getNextAudioFrame( FILE *fp, MP4Handle sampleH );

int main( int argc, char **argv )
{
	MP4Err err;
	

	if (argc < 4) {
		fprintf( stderr, 
			"usage: %s <sentence_filename> <ttsi_bitstream_filename> <TTS_Object>\n", 
			argv[0]);
		exit(1);
	}

	TTS=atoi(argv[3]);
	err = createMyMovie( argv[1], argv[2] );
	fprintf( stderr, "%s returns %d\n", argv[0], err );
	return err;
}

MP4Err createMyMovie( char *infile, char *outfile )
{
	MP4Err err;
	MP4Movie moov;
	MP4Track trak;
	MP4Media media;
	u8 OD_profileAndLevel;
	u8 scene_profileAndLevel;
	u8 audio_profileAndLevel;
	u8 visual_profileAndLevel;
	u8 graphics_profileAndLevel;
	u32 initialObjectDescriptorID;
	u32 timeScale;
	u64 mediaDuration;
	
	err = MP4NoErr;
	initialObjectDescriptorID = 1;
	OD_profileAndLevel = 0xff; /* none required */
	scene_profileAndLevel = 0xff; /* none required */
	audio_profileAndLevel = 0x01; /* main profile L1 */
	visual_profileAndLevel = 0xff; /* none required */
	graphics_profileAndLevel = 0xff; /* none required */
	timeScale = 8000; /* sampling frequency */
	err = MP4NewMovie( &moov,
		initialObjectDescriptorID, 
		OD_profileAndLevel,
		scene_profileAndLevel,
		audio_profileAndLevel,
		visual_profileAndLevel,
		graphics_profileAndLevel ); if (err) goto bail;
	err = MP4NewMovieTrack( moov, MP4NewTrackIsAudio, &trak ); if (err) goto bail;
	err = MP4NewTrackMedia( trak, &media, MP4AudioHandlerType, timeScale, NULL ); if (err) goto bail;
	err = MP4BeginMediaEdits( media ); if (err) goto bail;
	err = addMySamples( infile, trak, media ); if (err) goto bail;
	err = MP4EndMediaEdits( media ); if (err) goto bail;
	err = MP4GetMediaDuration( media, &mediaDuration ); if (err) goto bail;
	err = MP4InsertMediaIntoTrack( trak, 0, 0, mediaDuration, 1 ); if (err) goto bail;
	err = MP4WriteMovieToFile( moov, outfile ); if (err) goto bail;
	err = MP4DisposeMovie( moov ); if (err) goto bail;
bail:
	return err;
}

MP4Err addMySamples( char *infile, MP4Track trak, MP4Media media )
{
	MP4Err err;
	MP4Handle sampleEntryH;
	MP4Handle sampleDataH;
	MP4Handle sampleDurationH;
	MP4Handle sampleSizeH;
	MP4Handle decoderSpecificInfoH;
//	FILE *fp;
	u32 objectTypeIndication;
	u32 streamType;
	u32 decoderBufferSize;
	u32 maxBitrate;
	u32 avgBitrate;
	
	err = MP4NoErr;
	if ((fp = fopen(infile, "r")) == NULL) {
		err = MP4FileNotFoundErr;
		goto bail;
	}
	err = MP4SetMediaLanguage( media, "und" ); if (err) goto bail;

	objectTypeIndication = 0x40; /* mpeg-4 audio */
	streamType = 0x05; /* audio stream */
	decoderBufferSize = 8196; /* mono */
	maxBitrate = 1000;
	avgBitrate = 1000;
	err = MP4NewHandle( 0, &decoderSpecificInfoH ); if (err) goto bail;
	err = getDecoderSpecificInfo( decoderSpecificInfoH ); if (err) goto bail;

	err = MP4NewHandle( 0, &sampleEntryH ); if (err) goto bail;
	err = MP4NewSampleDescription( trak, sampleEntryH,
		1,
		objectTypeIndication,
		streamType,
		decoderBufferSize,
		maxBitrate,
		avgBitrate,
		decoderSpecificInfoH ); if (err) goto bail;

	err = MP4NewHandle( sizeof(u32), &sampleDurationH ); if (err) goto bail;
	*((u32*) *sampleDurationH) = 40000; /* one 5-second sentence */
	err = MP4NewHandle( 0, &sampleDataH ); if (err) goto bail;
	err = MP4NewHandle( sizeof(u32), &sampleSizeH ); if (err) goto bail;
	for ( ;; )
	{
		err = getNextAudioFrame( fp, sampleDataH );
		if ( err )
		{
			if ( err == MP4EOF )
				err = 0;
			break;
		}
		err = MP4GetHandleSize( sampleDataH, (u32*) *sampleSizeH ); if (err) goto bail;
		err = MP4AddMediaSamples( media, sampleDataH, 1, sampleDurationH, sampleSizeH, sampleEntryH, NULL, NULL ); if (err) goto bail;
		if ( sampleEntryH )
		{
			err = MP4DisposeHandle( sampleEntryH ); if (err) goto bail;
			sampleEntryH = NULL;
		}	
	}
bail:
	return err;
}

MP4Err getDecoderSpecificInfo( MP4Handle decoderSpecificInfoH )
{
	MP4Err err;
	int i, j;
	u8 *p, c, array[8];
	u32 frameLength;
	
	err = MP4NoErr;
	frameLength = sizeof(ttsiDecoderSpecificInfo[TTS-1]);

	/* encode sizeOfInstance 
	 * from 14496-1, section 12.3.3
	 *
	 * int sizeOfInstance = 0;
	 * bit(1) nextByte;
	 * bit(7) sizeOfInstance;
	 * while(nextByte) {
	 * 	bit(1) nextByte;
	 * 	bit(7) sizeByte;
	 * 	sizeOfInstance = sizeOfInstance<<7 | sizeByte;
	 * }
	 */
	p = array;
	for (i=4; i>=0; i--) {
		c = (frameLength >> 7*i) & 0x7f;
		if (c > 0) {
			if (i > 0)
				c |= 0x80;
			*p++ = c;
		}
	}
	j = p - array;

	/* set handle size */
	err = MP4SetHandleSize( decoderSpecificInfoH, 1+j+frameLength ); if (err) goto bail;
	fprintf( stderr,  "decoderSpecificInfo size is %d\n", 1+j+frameLength );

	p = (u8*) *decoderSpecificInfoH;
	*p++ = 0x05; /* DecSpecificInfoTag, 14496-1, Sec 8.2.2.2, Table 1*/
	for (i=0; i<j; i++)
		*p++ = array[i];
	for (i=0; i<frameLength; i++)
		*p++ = ttsiDecoderSpecificInfo[TTS-1][i];

bail:
	return err;
}

#define	MAX_SENTENCE_LEN 		1024
#define	MAX_PREAMBLE_LEN		16

static u32 bword = 0;
static u32 bbit = 0x80;

void putbyte ( u8 **pp )
{
	u8 *p = *pp;
	*p++ = bword;
/* fprintf(stderr, "0x%02x\n", bword); */
	bword = 0;
	bbit = 0x80;
	*pp = p;
}

void putbits( u8 **pp, u32 value, u32 n)
{
	while (n-- > 0) {
		if ( value & (1<<n) )
			bword |= bbit;
		bbit >>= 1;
		if (bbit == 0)
			putbyte( pp );
	}
}

void bytealign( u8 **pp )
{
	while ( bbit != 0x80 )
		putbits( pp, 0, 1 );
}

u32 createAccessUnit( u8 *accessUnit, u32 sentenceLength, u8 *sentence)
{
	int i;
	u8 *p = accessUnit;
	int j, tmp, ntmp;
	u32 Dur_Enable;
	u32 F0_Contour_Enable;
	u32 Energy_Contour_Enable;
	u32 Number_of_Phonemes;
	u32 Phoneme_Symbols_Length;

	// TTS Specific Config
	putbits( &p, 1, TTS_Sentence_ID_Len );
	putbits( &p, 0, Silence_Len );
	if(TTS==1 || TTS==7 || TTS==8) // Gender enable
		putbits( &p, 0, Gender_Len );
	if(TTS==1) // Age, Speech rate enable
	{
		putbits( &p, 3, Age_Len );
		putbits( &p, 8, Speech_Rate_Len );
	}
	putbits( &p, sentenceLength, Length_of_Text_Len );
	for (i=0; i<sentenceLength; i++) 
		putbits( &p, sentence[i], 8);

	// Add by ETRI
	if(TTS==2 || TTS==5 || TTS==6) // Prosody enable 
	{
		Dur_Enable=1;
		F0_Contour_Enable=1;
		Energy_Contour_Enable=0;
		if(TTS==2) Energy_Contour_Enable=1;	 
		Number_of_Phonemes=39;
		Phoneme_Symbols_Length=78;

		putbits( &p, Dur_Enable, Dur_Enable_Len );
		putbits( &p, F0_Contour_Enable, F0_Contour_Enable_Len );
		putbits( &p, Energy_Contour_Enable, Energy_Contour_Enable_Len );
		putbits( &p, Number_of_Phonemes, Number_of_Phonemes_Len );
		putbits( &p, Phoneme_Symbols_Length, Phoneme_Symbols_Len );

		fgets( sentence, MAX_SENTENCE_LEN, fp );
		for (i=0; i<strlen( sentence )-1; i++) 
			putbits( &p, sentence[i], 8);
	
		for(i=0;i<Number_of_Phonemes;i++)
		{
			fscanf(fp,"%d", &tmp); 
			putbits( &p, tmp, Dur_each_Phoneme_Len);
			fscanf(fp,"%d", &ntmp);
			putbits( &p, ntmp, Num_F0_Len);
			for(j=0; j<ntmp; j++) 
			{
				fscanf(fp,"%d", &tmp); 
				putbits( &p, tmp/2, F0_Contour_each_Phoneme_Len);
				fscanf(fp,"%d", &tmp); 
				putbits( &p, tmp, F0_Contour_each_Phoneme_Time_Len);
			}
			if(Energy_Contour_Enable)
			{
				for(j=0; j<3; j++)  
				{
					fscanf(fp,"%d", &tmp); 
					putbits( &p, tmp, Energy_Contour_each_Phoneme_Len);
				}
			}
		}
	}
	if(TTS==3 || TTS==5) // Video enable
	{
		u32 Sentence_Duration=101;
		u32 Position_in_Sentence=0;
		u32 Offset=1;
		putbits( &p, Sentence_Duration, Sentence_Duration_Len);
		putbits( &p, Position_in_Sentence, Position_in_Sentence_Len);
		putbits( &p, Offset, Offset_Len);
	}
	if(TTS==4 || TTS==6) // Lip Shape enable
	{
		u32 Number_of_Lip_Shape=39;
		putbits( &p, Number_of_Lip_Shape, Number_of_Lip_Shape_Len);
		for(j=0;j<Number_of_Lip_Shape;j++)
		{
			fscanf(fp,"%d", &tmp); 
			putbits( &p, tmp, Lip_Shape_in_Sentence_Len);
			fscanf(fp,"%d", &tmp); 
			putbits( &p, tmp, Lip_Shape_Len);
		}
	}
//////////////////////////////////////////////////////

	bytealign( &p );
	
	return( p - accessUnit );
}

MP4Err getNextAudioFrame( FILE *fp, MP4Handle sampleH )
{
	MP4Err err;
	u8 *p, sentence[MAX_SENTENCE_LEN], accessUnit[MAX_PREAMBLE_LEN+MAX_SENTENCE_LEN];
	u32 i, frameLength;
	
	err = MP4NoErr;
	if ( fgets( sentence, MAX_SENTENCE_LEN, fp ) == NULL ) /* one sentence */
		return MP4EOF;
	frameLength = createAccessUnit( accessUnit, strlen( sentence ), sentence);
	fprintf( stderr, "accessUnitSize is %d\n", frameLength );

	err = MP4SetHandleSize( sampleH, frameLength ); if (err) goto bail;
	p = (u8*) *sampleH;
	for (i=0; i<frameLength ; i++)
		*p++ = accessUnit[i];
bail:
	return err;
}
