/*
        This software module was originally developed by Jung-Chul Lee
        (ETRI: Electronics and Telecommunications Research Institute)
        and edited by Youngjik Lee (ETRI),  in
        the course of development of the MPEG-4 Audio (ISO/IEC 14496-3).
        This software module is an implementation of a part of one or more
        MPEG-4 Audio (ISO/IEC 14496-3) tools as
        specified by the MPEG-4 Audio (ISO/IEC 14496-3). ISO/IEC gives
        users of the MPEG-4 Audio (ISO/IEC 14496-3) free license to this
        software module or modifications thereof for use in hardware or
        software products claiming conformance to the MPEG-4 Audio
        (ISO/IEC 14496-3). Those intending to use this software module in
        hardware or software products are advised that its use may infringe
        existing patents. The original developer of this software module
        and his/her company, the subsequent
        editors and their companies, and ISO/IEC have no liability for use
        of this software module or modifications thereof in an implementation.
        Copyright is not released for non MPEG-4 Audio (ISO/IEC 14496-3)
        conforming products. ETRI retains full right to use the code for
        his/her own purpose, assign or donate the code to a
        third party and to inhibit third parties from using the code for
        non MPEG-4 Audio (ISO/IEC 14496-3) conforming products. This
        copyright notice must be included in all
        copies or derivative works. Copyright 1997, 1998, 1999.
*/ 

#include "MP4Movies.h"
#include "ttsi.h"

MP4Err playMyMovie( char *filename, u32 trackNumber );

// Add by ETR
FILE *fpw;

int main( int argc, char **argv )
{
	MP4Err err;
	
	fpw=fopen(argv[2],"w");
	if (argc < 3) {
		fprintf( fpw, 
			"usage: %s <ttsi_bitstream_filename> <output_filename>\n", 
			argv[0]);
		exit(1);
	}
	
	err = MP4NoErr;
	err = playMyMovie( argv[1], 1 );
	fprintf( fpw, "%s returns %d\n", argv[0], err );
	fclose(fpw);
	return err;
}

static u32 bword = 0;
static u32 bbit = 0;

void getbyte( u8 **pp )
{
	u8 *p = *pp;
	bword = *p++;
	bbit = 0x80;
	*pp = p;
}

void resetbits( void )
{
	bword = 0;
	bbit = 0;
}

u32 getbits( u8 **pp, u32 n)
{
	u32 bits = 0;
	while (n-- > 0) {
		if (bbit == 0)
			getbyte( pp );
		bits <<= 1;
		if (bword & bbit)
			bits |= 1;
		bbit >>= 1;
	}
	return bits;
}

void
print_bits(u32 *value, u8 **pp, char *s, u32 len ) {
	u32 *pv, dummy;
	if (value)
		pv = value;
	else
		pv = &dummy;

	if ( len <= 8 )	
	{
		fprintf( fpw, "%-20s 0x%02x\n", s, (*pv = getbits( pp, len )) );
	}
	else
	{
		fprintf( fpw, "%-20s 0x%0x\n", s, (*pv = getbits( pp, len )) );
	}
}

MP4Err playMyMovie( char *filename, u32 trackNumber )
{
	int i,j,ntmp;
	u8 *p;
	MP4Err err;
	MP4Movie moov;
	MP4Track trak;
	MP4Media media;
	MP4TrackReader reader;
	MP4Handle initialObjectDescriptorH;
	MP4Handle decoderConfigH;
	MP4Handle decoderSpecificInfoH;
	MP4Handle sampleH;
	u32 handlerType;
	u32 handleSize;
 	u32 objectTypeIndication;
	u32 streamType;
	u32 decoderBufferSize;
	
	u32 TTS_Seq_ID, Lang_Code, Gender_Enable, Age_Enable, Speech_Rate_Enable;
	u32 Prosody_Enable, Video_Enable, Lip_Shape_Enable, Trick_Mode_Enable;
	u32 Silence, Length_of_Text;

	// Add by ETRI
	u32 AudioObjectType, samplingFrequencyIndex, channelConfiguration;
	u32 Dur_Enable;
	u32 F0_Contour_Enable;
	u32 Energy_Contour_Enable;
	u32 Number_of_Phonemes;
	u32 Phoneme_Symbols_Length;
	u32 Num_F0, Number_of_Lip_Shape;

	err = MP4NoErr;
	err = MP4OpenMovieFile( &moov, filename, MP4OpenMovieNormal ); if (err) goto bail;
 	err = MP4NewHandle( 0, &initialObjectDescriptorH ); if (err) goto bail;
	err = MP4GetMovieInitialObjectDescriptor( moov, initialObjectDescriptorH ); if (err) goto bail;
	err = MP4GetMovieIndTrack( moov, trackNumber, &trak ); if (err) goto bail;
	err = MP4GetTrackMedia( trak, &media ); if (err) goto bail;
	err = MP4GetMediaHandlerDescription( media, &handlerType, NULL ); if (err) goto bail;
	err = MP4CreateTrackReader( trak, &reader ); if (err) goto bail;
	err = MP4NewHandle( 0, &decoderConfigH ); if (err) goto bail;
	err = MP4TrackReaderGetCurrentDecoderConfig( reader, decoderConfigH ); if (err) goto bail;
 	err = MP4NewHandle( 0, &decoderSpecificInfoH ); if (err) goto bail;
	err = MP4GetMediaDecoderType( media, 1, 
		&objectTypeIndication, 
		&streamType, 
		&decoderBufferSize, 
		decoderSpecificInfoH ); if (err) goto bail;
	/*
	 * check parameters in handlerType, objectTypeIndication, 
	 * streamType, profile and level (in initialObjectDescriptor) and
	 * decoderConfig and instantiate appropriate decoder
	 * (not shown)
	 */
	err = MP4GetHandleSize( decoderSpecificInfoH, &handleSize ); if (err) goto bail;
	fprintf( fpw, "send decoderSpecificInfo to decoder: size = %d\n", handleSize );
	p = *decoderSpecificInfoH;
	resetbits();

	// Audio Specific Config
	print_bits( &AudioObjectType, &p, "AudioObjectType", AudioObjectType_len );
	print_bits( &samplingFrequencyIndex, &p, "samplingFrequencyIndex", samplingFrequencyIndex_Len );
	print_bits( &channelConfiguration, &p, "channelConfiguration", channelConfiguration_Len );

	// TTS Specific Config
	print_bits( &TTS_Seq_ID, &p, "TTS_Seq_ID", TTS_Seq_ID_Len );
	print_bits( &Lang_Code, &p, "Lang_Code", Lang_Code_Len );
	print_bits( &Gender_Enable, &p, "Gender_Enable", Gender_Enable_Len );
	print_bits( &Age_Enable, &p, "Age_Enable", Age_Enable_Len );
	print_bits( &Speech_Rate_Enable, &p, "Speech_Rate_Enable", Speech_Rate_Enable_Len );
	print_bits( &Prosody_Enable, &p, "Prosody_Enable", Prosody_Enable_Len );
	print_bits( &Video_Enable, &p, "Video_Enable", Video_Enable_Len );
	print_bits( &Lip_Shape_Enable, &p, "Lip_Shape_Enable", Lip_Shape_Enable_Len );
	print_bits( &Trick_Mode_Enable, &p, "Trick_Mode_Enable", Trick_Mode_Enable_Len );

	err = MP4NewHandle( 0, &sampleH ); if (err) goto bail;	
	/* play every frame */
	for (;;)
	{
		u32 unitSize;
		u32 sampleFlags;
		u32 cts;
		u32 dts;
		
		err = MP4TrackReaderGetNextAccessUnit( reader, sampleH, &unitSize, &sampleFlags, &cts, &dts );
		if ( err )
		{
			if ( err == MP4EOF )
				err = MP4NoErr;
			break;
		}

		/* send AU to decoder */
		fprintf( fpw, "send AU to decoder: size = %d, cts = %d, dts = %d\n", unitSize, cts, dts );
		p = *sampleH;
		resetbits();
		print_bits( 0, &p, "TTS_Sentence_ID", TTS_Sentence_ID_Len );
		print_bits( &Silence, &p, "Silence", Silence_Len );
		if ( Silence ) {
			print_bits( 0, &p, "Silence_Dur", Silence_Dur_Len );
		}
		else {
			if ( Gender_Enable )
				print_bits( 0, &p, "Gender", Gender_Len );
			if ( Age_Enable )
				print_bits( 0, &p, "Age", Age_Len );
			if ( !Video_Enable && Speech_Rate_Enable ) 
				print_bits( 0, &p, "Speech_Rate", Speech_Rate_Len );
			fprintf( fpw, "%-20s %d\n", "Length_of_Text", (Length_of_Text = getbits( &p, Length_of_Text_Len )) );
			for (i=0; i<Length_of_Text; i++)
				fprintf( fpw, "%c", getbits( &p, 8 ) );
			fprintf( fpw, "\n");

			// Add by ETRI
			if ( Prosody_Enable )
			{
				fprintf( fpw, "<Prosody enable>\n");
				fprintf( fpw, "%-20s %d\n", "Dur_Enable", (Dur_Enable = getbits( &p, Dur_Enable_Len )) );
				fprintf( fpw, "%-20s %d\n", "F0_Contour_Enable", (F0_Contour_Enable = getbits( &p, F0_Contour_Enable_Len )) );
				fprintf( fpw, "%-20s %d\n", "Energy_Contour_Enable", (Energy_Contour_Enable = getbits( &p, Energy_Contour_Enable_Len )) );
				fprintf( fpw, "%-20s %d\n", "Number_of_Phonemes", (Number_of_Phonemes = getbits( &p, Number_of_Phonemes_Len )) );
				fprintf( fpw, "%-20s %d\n", "Phoneme_Symbols_Length", (Phoneme_Symbols_Length = getbits( &p, Phoneme_Symbols_Len )) );
				for(i=0;i<Phoneme_Symbols_Length*2;i++)
					fprintf( fpw, "%c", getbits( &p, 8 ) );
				fprintf( fpw, "\n");

				for(i=0;i<Number_of_Phonemes;i++)
				{
					if(Dur_Enable) 
						fprintf( fpw, "%d ", getbits( &p, Dur_each_Phoneme_Len ) );
					if(F0_Contour_Enable)
					{
						fprintf( fpw, "%d ", (Num_F0=getbits( &p, Num_F0_Len )) );
						for(j=0; j<Num_F0; j++) 
						{
							fprintf( fpw, "{%d, ", getbits( &p, F0_Contour_each_Phoneme_Len ) );
							fprintf( fpw, "%d} ", getbits( &p, F0_Contour_each_Phoneme_Time_Len ) );
						}
						fprintf( fpw,"\n");
					}
					if(Energy_Contour_Enable)
					{
						for(j=0; j<3; j++)  
							fprintf( fpw, "%d ", getbits( &p, Energy_Contour_each_Phoneme_Len ) );
						fprintf( fpw, "\n");
					}
				}
			}

			if ( Video_Enable )
			{
				fprintf( fpw, "<Video enable>\n");
				fprintf( fpw, "%-20s %d\n", "Sentence_Duration", getbits( &p, Sentence_Duration_Len ) );
				fprintf( fpw, "%-20s %d\n", "Position_in_Sentence", getbits( &p, Position_in_Sentence_Len ) );
				fprintf( fpw, "%-20s %d\n", "Offset", getbits( &p, Offset_Len ) );
			}
			if ( Lip_Shape_Enable )
			{
				fprintf( fpw, "<Lip_Shape enable>\n");
				fprintf( fpw, "%-20s %d\n", "Number_of_Lip_Shape", (Number_of_Lip_Shape=getbits( &p, Number_of_Lip_Shape_Len )) );
				for(j=0;j<Number_of_Lip_Shape;j++)
				{
					fprintf( fpw, "{%d, ", getbits( &p, Lip_Shape_in_Sentence_Len ) );
					fprintf( fpw, "%d}\n", getbits( &p, Lip_Shape_Len ) );
				}
				fprintf( fpw, "\n");
			}
		}
	}
bail:
	return err;
}
