/*****************************************************************************
 *                                                                           *
SC 29 Software Copyright Licencing Disclaimer:

This software module was originally developed by
  AT&T, Dolby Laboratories, Fraunhofer IIS-A

and edited by
  Yoshiaki Oikawa (Sony Corporation),
  Mitsuyuki Hatanaka (Sony Corporation)

in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3 
standards for reference purposes and its performance may not have been 
optimized. This software module is an implementation of one or more tools as 
specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
ISO/IEC gives users free license to this software module or modifications 
thereof for use in products claiming conformance to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International 
Standards. ISO/IEC gives users the same free license to this software module or 
modifications thereof for research purposes and further ISO/IEC standardisation.
Those intending to use this software module in products are advised that its 
use may infringe existing patents. ISO/IEC have no liability for use of this 
software module or modifications thereof. Copyright is not released for 
products that do not conform to audiovisual and image-coding related ITU 
Recommendations and/or ISO/IEC International Standards.
The original developer retains full right to modify and use the code for its 
own purpose, assign or donate the code to a third party and to inhibit third 
parties from using the code for products that do not conform to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International Standards.
This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 1996.
 *                                                                           *
 ****************************************************************************/

#include "adifdec.h"
#include "diagnose.h"
#include "audioout.h"
#include "bitinput.h"
#include "config.h"
#include "common.h"

#ifdef DRC
#include "drc.h" /* to control drc settings */
#endif
#include "gc.h"  /* to contol ssr settings */

static void             usage(char *s);
static int              parse_cmdline(int argc, char *argv[]);
static void             print_banner(char *cmd);

/* parameters */
static int param_outMultifiles = 0;
static int param_outWordsize = 16;
static char *param_outFormat = "wav";
#ifdef ENABLE_RTPLAY
static int param_outLevel = -1;
#endif

static char *bitstreaminputfile;

long            bno;
int             debug[256];
int             delay;
char            result[256];
char            resultQmf[256];
char            *audioinput_path;

static void
print_banner(char *cmd)
{
  PRINT (SE, "\n");
  PRINT (SE, "ISO/IEC 14496-4 MPEG-4 Conformance Verification\n");
  PRINT (SE, "%s  version %s\n", cmd, SBRCONF_VERSION);


#ifdef LOW_POWER_SBR
  PRINT (SE, "LOW POWER SBR\n");
#else
  PRINT (SE, "HIGH QUALITY SBR\n");
#endif
  PRINT (SE, "\n");
}

static void usage(char *s)
{
    PRINT(SE, " options are:\n\t");
    PRINT(SE, " -h show this help\n\t");
    PRINT(SE, " -d delay (delay introduced by decoder under test; default: 0)\n\t");
#ifdef CT_SBR
    PRINT(SE, " -l level  (Level of the HE-AAC Profile decoder (2,3,4,5), 5 is default)\n\t");
    PRINT(SE, " -r (Forces the decoder to work in down sampled SBR mode, default is off)\n\t");
#ifdef PARAMETRICSTEREO
    PRINT(SE, " -c (Enables parametric stereo decoding for HE-AAC)\n");
#endif
#endif
    myexit(0);
}

static int
parse_cmdline(int argc, char *argv[])
{
  int i, j, *p;
  char *cmd, *s;

  MC_Info *mip = &mc_info;

#ifdef	DRC
  Float param_drcHigh=0;
  Float param_drcLow=0;
  int param_drcTargetRefLevel=-1;
#endif
  int param_selectedProgram = -1;
  int param_ssrDecoderBand = -1;
  extern int delay;
  extern int fileIndex;

  /* save cmd 
   * ARGBEGIN is so clever that it throws away argv[0] and 
   * increments the argv pointer
   */
  cmd = argv[0];
    
  print_banner(cmd);

  ARGBEGIN {

  default:
    PRINT(SE, "unknown option\n");
  case 'h':
    usage(cmd);
  case 'd':
    s = ARGF();
    delay = atoi(s);
    break;
#ifdef CT_SBR
  case 'l':
    s = ARGF();
    mip->HE_AAC_level = atoi(s);
    break;
  case 'r':
    mip->bDownSampledSbr = 1;
    break;
#ifdef PARAMETRICSTEREO
  case 'c':
#ifdef LOW_POWER_SBR
    PRINT(SE, "PS not available for LP-SBR\n");
    myexit("sbrEnablePS");
#endif
    mip->sbrEnablePS = 1;
    break;
#endif
#endif
  case 'D':
    for (s = ARGF(); *s; s++) {
      debug[(int)(*s)] = 1;
      PRINT(SE, "debug option %c enabled\n", *s);
    }
  }
  ARGEND
      
    if(argc != 2) {
      usage(cmd);
    }

  printf("%s %s %s\n", GREP_STRING, argv[0], argv[1]);

  /* set decoder-global variables */
  decoder_selectProgram(param_selectedProgram);
  if (param_ssrDecoderBand>0) ssr_decoder_band = param_ssrDecoderBand;
  decoder_setupParameters(mc_info.sampling_rate_idx);
#ifdef DRC
  decoder_setDrc(param_drcHigh, param_drcLow, param_drcTargetRefLevel);
#endif


  bitstreaminputfile = argv[0];

  audioinput_path = argv[1];

  if(strstr(argv[0],"sbr_qmf_")){
    fileIndex = 0;
  }
  else if(strstr(argv[0],"sbr_e_")){
    fileIndex = 1;
  }
  else if(strstr(argv[0],"sbr_gh_")){
    fileIndex = 2;
  }
  else if(strstr(argv[0],"sbr_i_")){
    fileIndex = 3;
  }
  else if(strstr(argv[0],"sbr_s_")){
    fileIndex = 4;
  }
  else if(strstr(argv[0],"sbr_twi_")){
    fileIndex = 5;
  }
#ifdef PARAMETRICSTEREO
  else if(strstr(argv[0],"sbr_ps_")){
    fileIndex = 6;
  }
#endif

  return argc;
}

/* global, so that it can be closed in myexit() */
static HANDLE_BITINPUT bitstreaminput;
int
main(int argc, char *argv[])
{
  long avgBitrate = 0;
  long maxBitrate = 0;
  int bufferFullness = -1;
  int channelConfig;
  int remainingArgs;
  BUFFER_SIMULATION sim;

  decoder_init();
  remainingArgs = parse_cmdline(argc, argv);	    /* parse command line */

  bitstreaminput = bitinput_openFile(bitstreaminputfile,
                                     &prog_config,
                                     &mc_info,
                                     &channelConfig,
                                     &maxBitrate,
                                     &avgBitrate,
                                     &bufferFullness,
                                     (remainingArgs==1));

  decoder_setupParameters(mc_info.sampling_rate_idx);
  if (bitstreaminput==NULL) myexit("error opening bitstream");
  if (remainingArgs==1) myexit(NULL);

  /* prepare buffer simulation */
  sim = create_bitBuffer_sim();
  {
    long bufferSize;
    double avgFrameLen = .0;

    getDecoderBufferSize(&prog_config, channelConfig, &bufferSize);
    if (NULL != bitinput_getCurrentADTSHeader(bitstreaminput)) avgBitrate = 0; /* get more accurate estimate from ADTS-Headers */

    if (avgBitrate>0) avgFrameLen = BITRATE2FRAMELEN(avgBitrate);
    else if (avgBitrate==0) avgFrameLen = 0; /* unknown, calculate on the fly */
    else if (avgBitrate<0) avgFrameLen = -1; /* variable bitrate */

    if (bufferFullness == 0x7fffffff) { /* also indicates variable bitrate */
      bufferFullness = 0;
      avgFrameLen = -1;
    }

    bitBuffer_sim_setFrameLen(sim, avgFrameLen, bufferSize);
    if (maxBitrate>0) bitBuffer_sim_testPeakBitrate(sim, maxBitrate);
  }

  /* main loop */
  for(bno = 0;; bno++) {
    Float **output_data;
    int output_blockLen;
    int err = bitinput_getNextBlock(bitstreaminput);
    if (err == -2) break; /* eof */
    if (err) myexit("error reading next block");

    dec(&output_data, &output_blockLen);
    /* do buffer simulation */
    {
      const ADTS_Header *tmp_head = bitinput_getCurrentADTSHeader(bitstreaminput);
      update_bitBuffer_sim(sim, bitinput_resetCounter(bitstreaminput)<<3, tmp_head);
    }
    /* initial buffer fullness */
    if ((bno==0)&&(bufferFullness!=-1)) {
      bitBuffer_sim_adjustFullness(sim, bufferFullness);
    }
    check_bitBuffer_sim(sim, "", debug['b'], debug['B'], debug['R']);
  }

  myexit(NULL);
  return (EXIT_SUCCESS);
}


void
myexit(char *s)
{
  bitinput_close(bitstreaminput);
  decoder_free();
  printf("%s\n", resultQmf);
  printf("%s\n", result);
  if (s) PRINT(SE,"\n%s\n", s);
  exit(s ? 1 : 0);
}
