/*****************************************************************************
 *                                                                           *
"This software module was originally developed by NTT Mobile
Communications Network, Inc. in the course of development of the
MPEG-2 AAC/MPEG-4 Audio standard ISO/IEC 13818-7, 14496-1,2 and
3. This software module is an implementation of a part of one or more
MPEG-2 AAC/MPEG-4 Audio tools as specified by the MPEG-2 AAC/MPEG-4
Audio standard. ISO/IEC gives users of the MPEG-2 AAC/MPEG-4 Audio
standards free license to this software module or modifications
thereof for use in hardware or software products claiming conformance
to the MPEG-2 AAC/MPEG-4 Audio standards. Those intending to use this
software module in hardware or software products are advised that this
use may infringe existing patents. The original developer of this
software module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software
module or modifications thereof in an implementation. Copyright is not
released for non MPEG-2 AAC/MPEG-4 Audio conforming products.The
original developer retains full right to use the code for his/her own
purpose, assign or donate the code to a third party and to inhibit
third party from using the code for non MPEG-2 AAC/MPEG-4 Audio
conforming products. This copyright notice must be included in all
copies or derivative works." Copyright(c)1998.
 *                                                                           *
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "eptool.h"
#include "dcm_io.h"


/* Class Data Concatenation Procedure */

#if 0
static void
readBitsConcat(int *odata, EPConfig *epcfg, int dlen, FILE *ininfo, FILE *instream, FILE *outstream)
{
  char buf[255];
  int i,j,k;
  int idata[MAXCONCAT][MAXDATA];
  int PrevPred = 0;
  int ocnt, icnt[CLASSMAX];
  int pred;

  /* Read Input data for the concatenated frames */
  for(i=0; i<epcfg->NConcat; i++){
    if(i != 0)
      if(NULL == fgets(buf,sizeof(buf), ininfo)){
	fprintf(stderr, "No input data for concatenation!!\n");
	writebits(outstream, -1, NULL); /* Flush */
	exit(1);
      }
    if(1 != sscanf(buf, "%d", &dlen)) fprintf(stderr, "Data length Error\n");
    readbits(instream, dlen, idata[i]);
    SideInfoParse(ininfo, epcfg, &pred, NULL, 0);
    /* All the frames concatenated should have same pred value */
    if(i==0)
      PrevPred = pred;
    else
      if(PrevPred != pred){ /* Illeagal!! */
	fprintf(stderr, "Pred Mismatch on Concatenation!!\n");
	exit(1);
      }
  }
  
  ocnt = 0;
  for(i=0; i<epcfg->NConcat; i++) icnt[i] = 0;
  /* Concatenation */
  for(i=0; i<epcfg->fattr[pred].ClassCount; i++)
    for(j=0; j<epcfg->NConcat; j++)
      for(k=0; (unsigned int)k<epcfg->fattr[pred].cattr[i].ClassBitCount; k++)
	odata[ocnt++] = idata[j][icnt[j]++];
}
#endif

/* Generate New Class Attribution */

void ConcatAttrib(EPConfig *epcfg, EPConfig *nepcfg){
  int i,j;
  int nclass;
  int onclass;
  int pred;
  int npred;
  int TotalIninf = 0; /* init value was missing */
  EPFrameAttrib *fattr;
  ClassAttribContent *cattresc;
  ClassAttribContent *cattr;
  int *lenbit;

  /* Number of predefined set should be 1 */
  if(epcfg->NPred != 1){
    fprintf(stderr, "NPred should be 1 for Concatenation\n");
    exit(1);
  }
  /* EPConfig */
  npred = nepcfg->NPred = epcfg->NPred;
  nepcfg->Interleave = epcfg->Interleave;
  nepcfg->BitStuffing = epcfg->BitStuffing;
  if(nepcfg->BitStuffing) TotalIninf += 3;
  nepcfg->NConcat = 1;		/* Not used */
  nepcfg->HeaderProtectExtend = epcfg->HeaderProtectExtend;
  nepcfg->HeaderRate = epcfg->HeaderRate;
  nepcfg->HeaderCRC = epcfg->HeaderCRC;
#if RS_FEC_CAPABILITY
  nepcfg->RSFecCapability = epcfg->RSFecCapability;
#endif 

  /* EPFrameAttrib */
  /* Number of predefined set is 1, thus only one EPFrameAttrib !!*/
  if(NULL == (fattr = (EPFrameAttrib *)calloc(npred, sizeof(EPFrameAttrib)))){
    perror("nepcfg->fattr");
    exit(1);
  }
  nepcfg->fattr = fattr;
  
  pred = 0;

  /* Total Number of Class */
  nclass = 0;
  for(i=0; i<epcfg->fattr[pred].ClassCount; i++){
    if(epcfg->fattr[pred].cattr[i].Concatenate)
      nclass ++;
    else
      nclass += epcfg->NConcat;
  }
  nepcfg->fattr[pred].ClassCount = nclass;

  if(NULL == (cattr = (ClassAttribContent *)calloc(nclass, sizeof(ClassAttribContent)))){
    fprintf(stderr, "Unable to calloc for Cattr\n");
    exit(1);
  }
  if(NULL == (cattresc = (ClassAttribContent *)calloc(nclass, sizeof(ClassAttribContent)))){
    fprintf(stderr, "Unable to calloc for CattrESC\n");
    exit(1);
  }
  if(NULL == (lenbit = (int *)calloc(nclass, sizeof(int)))){
    fprintf(stderr, "Unable to calloc for CattrESC\n");
    exit(1);
  }
  fattr[pred].TotalIninf = TotalIninf;
  fattr[pred].cattresc = cattresc;
  fattr[pred].cattr = cattr;
    fattr[pred].lenbit = lenbit;


  onclass = 0;
  for(i=0; i<epcfg->fattr[pred].ClassCount; i++){
    
    if(epcfg->fattr[pred].cattr[i].Concatenate){ /* Concatenated */
      int lene, ratee, crce;
      /* New One Class */
      lene = cattresc[onclass].ClassBitCount =
	epcfg->fattr[pred].cattresc[i].ClassBitCount;
      ratee = cattresc[onclass].ClassCodeRate =
	epcfg->fattr[pred].cattresc[i].ClassCodeRate;
      crce = cattresc[onclass].ClassCRCCount =
	epcfg->fattr[pred].cattresc[i].ClassCRCCount;
      cattr[onclass].ClassInterleave = 
	epcfg->fattr[pred].cattr[i].ClassInterleave;
      /* No ESC allowed!! */
      if(lene || ratee || crce){
	fprintf(stderr, "No ESC allowed for Concatenation!\n");
	exit(1);
      }

      cattr[onclass].ClassBitCount =
	epcfg->fattr[pred].cattr[i].ClassBitCount * epcfg->NConcat;
      cattr[onclass].ClassCodeRate = epcfg->fattr[pred].cattr[i].ClassCodeRate;
      cattr[onclass].ClassCRCCount = epcfg->fattr[pred].cattr[i].ClassCRCCount;
      onclass ++;
    }else{			/* Not Cocatenated */
      for(j=0; j<epcfg->NConcat; j++, onclass++){
	int lene, ratee, crce;
	/* New One Class */
	lene = cattresc[onclass].ClassBitCount =
	  epcfg->fattr[pred].cattresc[i].ClassBitCount;
	ratee = cattresc[onclass].ClassCodeRate =
	  epcfg->fattr[pred].cattresc[i].ClassCodeRate;
	crce = cattresc[onclass].ClassCRCCount =
	  epcfg->fattr[pred].cattresc[i].ClassCRCCount;
	cattr[onclass].ClassInterleave = 
	  epcfg->fattr[pred].cattr[i].ClassInterleave;
	/* No ESC allowed!! */
	if(lene || ratee || crce){
	  fprintf(stderr, "No ESC allowed for Concatenation!\n");
	  exit(1);
	}
      
	cattr[onclass].ClassBitCount =
	  epcfg->fattr[pred].cattr[i].ClassBitCount;
	cattr[onclass].ClassCodeRate =
	  epcfg->fattr[pred].cattr[i].ClassCodeRate;
	cattr[onclass].ClassCRCCount =
	  epcfg->fattr[pred].cattr[i].ClassCRCCount;

      }	
    }
  }
  if(onclass != nclass){
    fprintf(stderr, "onclass:%d <> nclass:%d\n!!\n", onclass, nclass);
  }

}
