#include <stdio.h>
#include <stdlib.h>

#include "libeptool.h"
#include "calcoverhead.h"
#include "eptool.h"

#define D(a) {const void *dummyfilepointer; dummyfilepointer = &a;}


typedef struct tagEpConfig {

  EPConfig epConfig;            /* modified perdefinition set */
  EPConfig origEpConfig;        /* original predefinition set */
  int classAvailableTable[MAXPREDwCA][CLASSMAX];
  int classAvailableBits [MAXPREDwCA];
  int predCA[MAXPREDwCA][255];
  /* pointer to arrays managed by HANDLE_CLASS_FRAME */
  unsigned char    **classBuffer;            /* to avoid memory holes ;-) */
  long              *classLength;            /* to avoid memory holes ;-) */
  long              *classCodeRate;          /* SRCPC code rate */
  long              *classCRCLength;         /* CRC length for each class */
  short             *crcError;               /* to avoid memory holes ;-) */
} T_EP_CONFIG;

static void EPTool_BytesToBits ( unsigned char *inbuf, int *trans, int translen );
static int  EPTool_BitsToBytes ( unsigned char *outbuf, int *trans, int translen );
static int  EPTool_BytesToBits_IntToChar( unsigned int *inbuf, unsigned char *outbuf, int bitlen );
static int  EPTool_BitsToBytes_CharToInt( unsigned char *inbuf, int *outbuf, int bitlen );
static void EPTool_BytesToBits_CharToChar( unsigned char *inbuf, unsigned char *outbuf, int bitlen );
static int  FrameHandle_Create ( HANDLE_EP_CONFIG inFrame, int bitFlag );
static int  FrameHandle_Destroy( HANDLE_EP_CONFIG inFrame );
static int  EpTool_GetNumberOfClasses( HANDLE_EP_CONFIG  hEpConfig );
static void EPTool_SideInfoWrite( HANDLE_CLASS_FRAME outBuffer, EPConfig *modepcfg, int modpred, int modcheck[], int reorderFlag, EPConfig *origepcfg, int CAtbl[MAXPREDwCA][CLASSMAX], int CAbits[MAXPREDwCA], int integrated );
static void EPTool_NullSideInfoWrite( HANDLE_CLASS_FRAME outBuffer, EPConfig *modepcfg, int modpred, int modcheck[], EPConfig *origepcfg, int CAtbl[MAXPREDwCA][CLASSMAX], int CAbits[MAXPREDwCA], int integrated );
static int EPTool_OrigClassOutputIndex( EPConfig *modepcfg, int modpred, EPConfig *origepcfg, int CAtbl[MAXPREDwCA][CLASSMAX], int CAbits[MAXPREDwCA], int origOutputIndex[CLASSMAX] );
static int EPTool_ParseEPSpecificConfig( unsigned char *inbuffer, long int configLength, HANDLE_EP_CONFIG hEpConfig );
static long int EPTool_WriteEPSpecificConfig( unsigned char *outbuffer, HANDLE_EP_CONFIG hEpConfig );


int EPTool_CreateExpl (HANDLE_EP_CONFIG  *hEpConfig,
                       const char        *predFileName,
                       unsigned char    **configBuffer, /* output */
                       long              *configLength,  /* in bit */
                       int                reorderFlag,
                       int                bitFlag
                       )
{
  FILE *fp=NULL;
  EPConfig nepcfg;
  int i,origpred;
  int fileParse = 0;

  if (NULL != predFileName ){
    if (NULL == (fp = fopen(predFileName, "r"))) {
      fprintf(stderr,"EPTool: EPTool_CreateExpl(): Error: predefinition file not found\n");
      return 1;
    }
    fileParse = 1;
  } else {
    if (NULL == configBuffer || NULL == *configBuffer){
      fprintf(stderr, "EPTool: EPTool_CreateExpl(): Error: configBuffer not available\n");
      return 1;
    }
  }

  /* memory allocation */
  *hEpConfig = (HANDLE_EP_CONFIG) calloc(1, sizeof(T_EP_CONFIG));
  
  def_GF();                     /* generate GF(2^8) tables*/

  if ( fileParse ){
    OutInfoParse(fp, &((*hEpConfig)->origEpConfig), 0, (*hEpConfig)->classAvailableTable, (*hEpConfig)->classAvailableBits, 1);
  } else {
    if(EPTool_ParseEPSpecificConfig( *configBuffer, *configLength, *hEpConfig )){
      free(hEpConfig);
      return -1;
    }
  }
  PredSetConvert(&((*hEpConfig)->origEpConfig), &((*hEpConfig)->epConfig), (*hEpConfig)->classAvailableTable , (*hEpConfig)->classAvailableBits, (*hEpConfig)->predCA, reorderFlag);

  FrameHandle_Create ( *hEpConfig, bitFlag );

  origpred = (*hEpConfig)->origEpConfig.NPred - 1;

  for (i=0; i< (*hEpConfig)->epConfig.fattr[origpred].ClassCount; i++) {
    (*hEpConfig)->classCodeRate[i]  = (*hEpConfig)->epConfig.fattr[origpred].cattr[i].ClassCodeRate ;
    (*hEpConfig)->classCRCLength[i] = (*hEpConfig)->epConfig.fattr[origpred].cattr[i].ClassCRCCount ;
  }

  if((*hEpConfig)->epConfig.NConcat > 1) {        /* Concatenation! */
    ConcatAttrib(&((*hEpConfig)->epConfig), &nepcfg);
  }

  if(fileParse){
    if((configBuffer != NULL) && (*configBuffer !=NULL) && (configLength != NULL)){
      *configLength = EPTool_WriteEPSpecificConfig(*configBuffer, *hEpConfig);
      fclose(fp);
    } else {
      fprintf( stderr, "EP_Tool: EPTool_CreateExpl(): libeptool.c: Failed writing to configBuffer\n");
      return 0;
    }
  }
  return 0;
}

int EPTool_CalculateOverhead( HANDLE_EP_CONFIG hEpConfig ,  HANDLE_CLASS_FRAME inBuffer, int *outLength)
{
  int pred = 0, hlen = 0, h1len = 0, h2len = 0, bitstuffing = 0, dlen = 0;
  int nclass, nstuff, interleave, total;
  int i;
  int codedlen [CLASSMAX] = {0};
  int origpred, skipped = 0, CA;

  origpred = hEpConfig->origEpConfig.NPred -1;
  CA = 0;
  for (i=0; i< hEpConfig->epConfig.fattr[origpred].ClassCount; i++){
    if (hEpConfig->classAvailableTable[origpred][i] == 1 && inBuffer->classLength[i] == 0)
      CA |= 1 << i;
  }  

  pred = hEpConfig->predCA[origpred][CA];

  bitstuffing = hEpConfig->epConfig.BitStuffing;

  for  (i=0; i < inBuffer->nrValidBuffer; i++) {
    if (hEpConfig->classAvailableTable[origpred][i] == 1 && inBuffer->classLength[i] == 0)
      skipped++;
    else {
      hEpConfig->epConfig.fattr[pred].cattr[i-skipped].ClassBitCount = inBuffer->classLength[i];
      hEpConfig->epConfig.fattr[pred].cattr[i-skipped].ClassCodeRate = inBuffer->classCodeRate[i];
      hEpConfig->epConfig.fattr[pred].cattr[i-skipped].ClassCRCCount = inBuffer->classCRCLength[i];
    }
  }

  hlen = CalculateHeaderOverhead(&hEpConfig->epConfig, pred, &h1len, &h2len, bitstuffing);

  nclass = hEpConfig->epConfig.fattr[pred].ClassCount;
  interleave = hEpConfig->epConfig.Interleave; 

  dlen = 0;
  for (i=0; i< nclass; i++) {
    dlen += hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount ;
  }

  total =  CalculatePayloadOverhead(dlen, &hEpConfig->epConfig.fattr[pred], codedlen);

  if(bitstuffing) {
    nstuff = ((total + hlen + 7) /8)*8 - (total + hlen);
    hlen = CalculateHeaderOverhead(&hEpConfig->epConfig, pred, &h1len, &h2len, bitstuffing);
    total += nstuff;
  } else {
    nstuff = 0;
  }
  
  *outLength = h1len + h2len + total;
  
  return 0;
}

int EPTool_EncodeFrame( HANDLE_EP_CONFIG hEpConfig, HANDLE_CLASS_FRAME inBuffer, unsigned char *outBuffer, int *outLength , int bitFlag)
{
  int pred, hlen, /* h1len, h2len, */ bitstuffing;
  int nphdr[MAXHEADER], bdyhdr[MAXHEADER] ;
  int nclass, nstuff, interleave, total;
  int i, j, k;
  static int coded [MAXCODE] = {0};
  int codedlen [CLASSMAX] = {0}, codedstart[CLASSMAX] = {0};
  int clen, nclen, cptr;
  int trans[MAXDATA*2], translen = 0;
  int rs_index[256], rs_codelen;
  unsigned int rs_code[256];
  int rs_code_bin[256*8];
  int rs_parity[127*(MAXDATA/255)];
  int origpred ,skipped, CA;
  int offset;
  headerLength h1len, h2len; /* header lengths for header part choiceOfPred + choiceOfPredParity, classAttrib + classAttrobParity */

  origpred = hEpConfig->origEpConfig.NPred -1;
  
  CA = 0;
  for (i=0; i< hEpConfig->origEpConfig.fattr[origpred].ClassCount; i++){
    if (hEpConfig->classAvailableTable[origpred][i] == 1 && inBuffer->classLength[i]  == 0)
      CA |= 1 << i;
  }  

  pred = hEpConfig->predCA[origpred][CA]; 
  bitstuffing = hEpConfig->epConfig.BitStuffing;
  if (!bitstuffing && !bitFlag){
    fprintf(stderr, "EP_Tool: EPTool_EncodeFrame(): Error: Your in- and output is in bytes and *not* in real bits. Therefore you have to use the bitstuffing-functionality. If you change your in- and output, please change the bitFlag, too!\n");
    exit(1);
  }

  skipped = 0;
  for  (i=0; i < inBuffer->nrValidBuffer; i++) {
    if (hEpConfig->classAvailableTable[origpred][i] == 1 && inBuffer->classLength[i] == 0)
      skipped++;
    else {
      hEpConfig->epConfig.fattr[pred].cattr[i-skipped].ClassBitCount = inBuffer->classLength[i];
      hEpConfig->epConfig.fattr[pred].cattr[i-skipped].ClassCodeRate = inBuffer->classCodeRate[i];
      hEpConfig->epConfig.fattr[pred].cattr[i-skipped].ClassCRCCount = inBuffer->classCRCLength[i];
    }
  }

  hlen = GenerateHeader (&hEpConfig->epConfig, pred, nphdr, &h1len, bdyhdr, &h2len, bitstuffing, 0 );

  nclass = hEpConfig->epConfig.fattr[pred].ClassCount;
  interleave = hEpConfig->epConfig.Interleave; 

  for (i=0; i< nclass; i++) {
    translen += hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount ;
  }

  /* check, if at least inputbuffers are sufficent */
  if(translen > MAXDATA){
    fprintf(stderr, "EPTool: EPTool_EncodeFrame: Inputlength exceeds max. Buffersize. Please increase MAXDATA in eptool.h.");
    exit(-1);
  }

  offset = 0;
  /* write bytes to bits */
  for (i = 0, k = 0; i < nclass; k++) { 
    if (hEpConfig->classAvailableTable[origpred][k] == 1 && inBuffer->classLength[k] == 0) 
      continue;
    if (!bitFlag){
      EPTool_BytesToBits(inBuffer->classBuffer[k], &trans[offset], hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount);
    }else{
      unsigned int j;
      for (j=0; j<hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount;j++)
        trans[offset + j] = inBuffer->classBuffer[k][j];
    }
      
    offset += hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount;
    i++;
  } 

  total =  PayloadEncode(trans, translen, coded, &hEpConfig->epConfig.fattr[pred], codedlen);

  if(bitstuffing) {
    nstuff = ((total + hlen + 7) /8)*8 - (total + hlen); 
    hlen = GenerateHeader(&hEpConfig->epConfig, pred, nphdr, &h1len, bdyhdr, &h2len, bitstuffing, nstuff);
    for(i=0; i<nstuff;i++){
      coded[total+i] = 0;
    }
    total += nstuff;
  } else {
    nstuff = 0;
  }

  if(interleave) {
    int *tmp3;
    if(NULL == (tmp3 = (int *)calloc(sizeof(int), total))){
      perror("EP_Tool: EPTool_EncodeFrame(): tmp3"); exit(1);
    }
    clen = nclen = cptr = 0;

    translen = 0;
    codedstart[0] = 0;
    for(i=0; i<nclass-1; i++)
      codedstart[i+1] = codedstart[i] + codedlen[i];

    /* Concat Non-Interleaved classes */
    for(i=0; i<nclass; i++) {
      if (hEpConfig->epConfig.fattr[pred].ClassReorderedOutput)
        k = hEpConfig->epConfig.fattr[pred].OutputIndex[i] ;
      else
        k = i;
      if(!hEpConfig->epConfig.fattr[pred].cattr[k].ClassInterleave) {
        for(j=0; j<codedlen[i]; j++)
          tmp3[nclen++] = coded[codedstart[i]+j];
      }
    }
    
    /* Concat "Concatenation:3" classes */
    for(i=0; i<nclass; i++) {
      if (hEpConfig->epConfig.fattr[pred].ClassReorderedOutput)
        k = hEpConfig->epConfig.fattr[pred].OutputIndex[i] ;
      else
        k = i;
      if(hEpConfig->epConfig.fattr[pred].cattr[k].ClassInterleave == 3) {
        for(j=0; j<codedlen[i]; j++)
          trans[translen++] = coded[codedstart[i]+j];
        /*      printf("translen=%d\n",translen); */
      }
    }

    /* Interleave classes */
    for(i=nclass-1; i>=0; i--){
      int Interleave;
      if (hEpConfig->epConfig.fattr[pred].ClassReorderedOutput)
        k = hEpConfig->epConfig.fattr[pred].OutputIndex[i] ;
      else
        k = i;
      Interleave = hEpConfig->epConfig.fattr[pred].cattr[k].ClassInterleave;
      if(hEpConfig->epConfig.fattr[pred].cattr[k].FECType==0) {
        if(Interleave == 1){ /* Intra w/o Inter */
          translen = recinter(&coded[codedstart[i]], codedlen[i], codedlen[i], trans, translen, trans);
          cptr += codedlen[i];
        }else if(Interleave == 2){ /* Intra w/ Inter */
          translen = recinter(&coded[codedstart[i]], codedlen[i], PINTDEPTH, trans, translen, trans);
          cptr += codedlen[i];
        }
      }
      else { 
        if(Interleave == 1 || Interleave == 2){
          /* bytewise interleave for RS  */
          translen = recinterBytewise(&coded[codedstart[i]], codedlen[i], hEpConfig->epConfig.fattr[pred].cattr[k].ClassCodeRate,trans, translen, trans);
          cptr += codedlen[i];
        }
      }
    }

    for(i=0; i<nstuff; i++)  
      tmp3[nclen++] = 0; 
      
    for(i=0; i<nclen; i++) {
      trans[translen++] = tmp3[i];
    }
    translen = recinter(bdyhdr, h2len.headerLen, h2len.interleavingWidth, trans, translen, trans);
    translen = recinter(nphdr, h1len.headerLen, h1len.interleavingWidth, trans, translen, trans);
    
    free(tmp3);
  } else {
    for(i=j=0; i<h1len.headerLen; i++)
      trans[j++] = nphdr[i];
    for(i=0; i<h2len.headerLen; i++)
      trans[j++] = bdyhdr[i];
    for(i=0; i<total; i++)
      trans[j++] = coded[i]; /* there are the stuffin bits also in ... in case there is _no_ interleaving 020410 multrums@iis.fhg.de*/
    translen = j;
  }

  if (bitFlag){
    for(i=0;i<translen;i++)
      outBuffer[i] = trans[i];
  } else {
  EPTool_BitsToBytes (outBuffer, trans, translen);
  }

#if RS_FEC_CAPABILITY
  /* RS Encoding */
  if(hEpConfig->epConfig.RSFecCapability){
    int e_target, rs_inflenmax;
    int nparity;
    int pcnt, icnt;
    int inflen;

    /* Initialize tables for RS encoder */
    make_genpoly(hEpConfig->epConfig.RSFecCapability); /* generate generator poly */

    e_target = hEpConfig->epConfig.RSFecCapability;
    rs_inflenmax = 255 - 2 * e_target;
    nparity = (translen/8 + rs_inflenmax - 1) / rs_inflenmax;
    pcnt = icnt = 0;

    /*if(nparity != 1) fprintf(stderr, "EP_Tool: EPTool_EncodeFrame(): Parity %d\n", nparity);*//* what's that??? */
    for(i=0; i< nparity; i++){
      if(i == nparity - 1) /*last*/
        inflen = (translen - icnt)/8;
      else
        inflen = rs_inflenmax;
      /*      printf("[inflen = %d/%d]", inflen, translen/8); */

      rs_codelen = RS_input(&trans[icnt], rs_index, inflen*8);
      rs_codelen = RS_enc(rs_index, rs_codelen, rs_code, e_target);
      rs_codelen = RS_output(rs_code, rs_code_bin, rs_codelen, e_target, 0);

#if 0
      {
        /* Local Decode for Debugging. */
        unsigned int test[255];
        int testbin[255*8], tmp, flg, err, rs_dindex[256];

        tmp = RS_input(rs_code_bin, rs_dindex, rs_codelen);
        err = RS_dec(rs_dindex, test, tmp, e_target);
        tmp = RS_output(test, testbin, tmp, e_target, 1);

        flg = 0;
        for(j=0; j<tmp; j++)
          if(testbin[j] != rs_code_bin[j]){
            fprintf(stderr, "![%d]",j);
            flg = 1;
          }
        if(flg){
          fprintf(stderr, "\nEP_Tool: EPTool_EncodeFrame(): BUG!!(%d) tmp=%d, inflen =%d, translen/8 = %d rs_codelen = %d\n",
                 i, tmp, inflen, translen/8, rs_codelen);
        }
      }
#endif

      for(j=0; j<inflen*8; j++){
        if(rs_code_bin[j] != trans[icnt+j])
          fprintf(stderr, "EP_Tool: EPTool_EncodeFrame(): !BUG!\n");
      }
      for(j=0; j<2*e_target*8; j++){
        rs_parity[pcnt++] = rs_code_bin[inflen*8+j];
      }

      icnt += inflen*8;
    }
    if(icnt != translen) fprintf(stderr, "EP_Tool: EPTool_EncodeFrame(): <%d,%d>\n", icnt,translen);

    if(bitFlag){
      for(i=0;i<pcnt;i++)
        outBuffer[translen+i] = rs_parity[i];
    } else {
    EPTool_BitsToBytes (&outBuffer[(translen+7)/8], rs_parity, pcnt);
    }
    
    translen += pcnt;
  }
#endif
  
  *outLength = translen ; 
  
  return 0;
}

int EPTool_DecodeFrame( HANDLE_EP_CONFIG hEpConfig, HANDLE_CLASS_FRAME outBuffer, unsigned char *inBuffer, int inLength, int bitFlag)
{
  int i, j;
  int trans[MAXDATA*2];
  int translen = inLength;
  int rs_index[256], rs_codelen, rs_detect;
  unsigned int rs_out[256];
  int rs_data[256*8];
  int hlen, pred, nclass, nstuff = 0, clen = 0;
  static int ibuf[MAXDATA*2];
  static int ncbuf[MAXDATA];
  static int coded[MAXCODE];
  static int data[MAXDATA];
  int payloadlen;
  int check[CLASSMAX];
  int codedlen[CLASSMAX];
  int bound_punc[CLASSMAX];


  /* check at least buffersize for inputlength */
  if(inLength > MAXCODE){
    fprintf(stderr, "EPTool: EPTool_DecodeFrame: Inputlength exceeds max. buffersize. Please increase MAXCODE in eptool.h.");
    exit(-1);
  }

  if (bitFlag){
    for (i=0; i<translen; i++)
      trans[i] = inBuffer[i];
  } else {
  EPTool_BytesToBits(inBuffer, trans, translen);
  }

  if (!hEpConfig->epConfig.BitStuffing && !bitFlag){ 
    fprintf(stderr, "EP_Tool: EPTool_DecodeFrame(): Error: Your in- and output is in bytes and *not* in real bits. Therefore you have to use the bitstuffing-functionality. If you change your in- and output, please change the bitFlag, too!\n"); 
    exit(1); 
   } 

#if RS_FEC_CAPABILITY
  if (hEpConfig->epConfig.RSFecCapability) {
    int e_target, rs_inflenmax;
    int nparity, lastinflen;
    int inflen;
    int icnt, dcnt, pcnt;

    /* Initialize tables for RS encoder */
    make_genpoly(hEpConfig->epConfig.RSFecCapability); /* generate generator poly */

    e_target = hEpConfig->epConfig.RSFecCapability;
    rs_inflenmax = 255 - 2 * e_target;

    nparity = (translen/8 + 254) / 255;
    lastinflen = translen/8 - 255*(nparity-1) - 2*e_target;
    if(nparity != 1) fprintf(stderr, "EPTool: EPTool_DecodeFrame(): Parity %d\n",nparity);

    pcnt = translen - 2*e_target*8*nparity;
    icnt = 0;
    for(i=0; i<nparity; i++){
      dcnt = 0;
      inflen = (i == nparity-1 ? lastinflen : rs_inflenmax);
      /*      printf("[%d/%d]", inflen, translen/8);*/
      
      for(j=0; j< inflen*8; j++)
        rs_data[dcnt++]= trans[icnt + j]; 
      for(j=0; j<2*e_target*8; j++)
        rs_data[dcnt++] = trans[pcnt++];
      
      
      rs_codelen = RS_input(rs_data, rs_index, dcnt);
      rs_detect = RS_dec(rs_index, rs_out, rs_codelen, e_target);
      rs_codelen = RS_output(rs_out, rs_data, rs_codelen, e_target, 1);
      
      for(j=0; j< rs_codelen; j++){
        trans[icnt + j] = rs_data[j];
      }
      
      icnt += j;
    }
    translen -=  2*e_target*8 * nparity;
  }
#endif

  if (hEpConfig->epConfig.Interleave) {
    int *tmp1,*tmp2;
    int nclen, sptr, nptr, jcnt;

    int SRCPCStart  = 0;
    int rsconcat    = 0;
    int rsconcatlen = 0;

    hlen = RetrieveHeaderDeinter(&hEpConfig->epConfig, &pred, trans, translen, ibuf, &nstuff);
    /* ibuf safes the remaining interleaved data of the body */

    if (hlen == -1) {
      EPTool_NullSideInfoWrite(outBuffer, &hEpConfig->epConfig, 0, check, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, 1);
      return -1;
    }
    
    if(NULL == (tmp1 = (int *)calloc(sizeof(int), translen))){
      perror("EPTool: EPTool_DecodeFrame(): tmp1"); exit(1);
    }
    if(NULL == (tmp2 = (int *)calloc(sizeof(int), translen))){
      perror("EPTool: EPTool_DecodeFrame(): tmp2"); exit(1);
    }

    nclass = hEpConfig->epConfig.fattr[pred].ClassCount;
    clen = nclen = sptr = 0;
    /* Length other than last class */
    sptr = 0;
    for(i=0; i< CLASSMAX; i++) codedlen[i] = 0;  /* Initialization */
    bound_punc[0] = 0;
  
    for(i=0; i<nclass-1; i++){
      int srclen, rate;

      srclen = hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount + (hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount?hEpConfig->epConfig.fattr[pred].cattr[i].ClassCRCCount:0);

      if(hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 0){
        bound_punc[i+1] = srclen + bound_punc[i];
        if(hEpConfig->epConfig.fattr[pred].cattr[i].TermSW){
          if(bound_punc[i+1] - bound_punc[SRCPCStart]){
            bound_punc[i+1] += 4;                       /* add tailbits */
          }
          SRCPCStart = i+1;
        }
      }else{
        bound_punc[i+1] = bound_punc[i];
      }
      rate = hEpConfig->epConfig.fattr[pred].cattr[i].ClassCodeRate;

      if(hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 0){ /* SRCPC */
        {
          int tmp, m, tmpClassRate;
        
          tmp = 8 - bound_punc[i]%8;
          if(tmp == 8){
            tmp = 0;
          }
          
          if(tmp){
            if(tmp > bound_punc[i+1]-bound_punc[i]){
              tmp = bound_punc[i+1]-bound_punc[i];
            }
            
            tmpClassRate = 0;                   /* to prevent reading of not initialized mem */
            for(m=i-1; m>=0; m--){
              if(hEpConfig->epConfig.fattr[pred].cattr[m].FECType == 0){
                tmpClassRate = hEpConfig->epConfig.fattr[pred].cattr[m].ClassCodeRate;
                if(bound_punc[i]-bound_punc[m] >= bound_punc[i]%8){
                  break;
                }
              }
            }
            codedlen[i] = SRCPCDepuncCount(bound_punc[i], bound_punc[i]+tmp, 0, tmpClassRate);
          }
          
          codedlen[i] += SRCPCDepuncCount(bound_punc[i]+tmp, bound_punc[i+1], 0, hEpConfig->epConfig.fattr[pred].cattr[i].ClassCodeRate); /* Calculate length of SRCPC encode class */
        }
        
        /*  codedlen[i] = SRCPCDepuncCount(sptr, sptr+srclen, 0, rate); */
      } else { /* RS */
        /* FEC sequence */
        /* ... 2 2 ... */
        if ( hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 2 ) {
          rsconcat++;
          codedlen[i]   = 0;
          rsconcatlen += srclen; 
        }
        if ( hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 1) {
          /* ... 2 1 ... */
          if ( rsconcat ) {
            rsconcat=0;
            rsconcatlen += srclen;
            codedlen[i] = RSEncodeCount(0, rsconcatlen, rate);
          }
          else 
            /*  ... 1 1 ... */
            codedlen[i] = RSEncodeCount(0, srclen, rate);
        }
      }
      if(hEpConfig->epConfig.fattr[pred].cattr[i].ClassInterleave) {
        clen += codedlen[i];
      }
      else {
        nclen += codedlen[i];
      }
    }

    if(translen < clen + hlen){ /* Check the length other than last */
      EPTool_NullSideInfoWrite(outBuffer, &hEpConfig->epConfig, 0, check, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, 1);
      return -1;
    }

    if (translen < hlen+clen+nclen+nstuff) { /* check all length */
      EPTool_NullSideInfoWrite(outBuffer, &hEpConfig->epConfig, 0, check, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, 1);
      return -1;
    }

    /* Last class length */
    codedlen[nclass-1] = translen-hlen-clen-nclen-nstuff;

    if(hEpConfig->epConfig.fattr[pred].cattr[nclass-1].ClassInterleave)
      clen += codedlen[nclass-1];
    else
      nclen += codedlen[nclass-1];
    
    /* The number of Stuffing Bits */
    nclen += nstuff;
    
    /* De-interleaving !!*/
    sptr = 0; nptr = 0;
    for(i=0; i<nclen; i++) ncbuf[i] = ibuf[i+clen];
    for(i=0; i<nclass; i++){
      int Interleave;
      Interleave = hEpConfig->epConfig.fattr[pred].cattr[i].ClassInterleave;
      if(Interleave == 0){ /* No interleave */
        for(j=0; j<codedlen[i]; j++)
          coded[sptr++] = ncbuf[nptr++];
      }else if(Interleave == 1 && hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 0){ /*Intra w/o Inter */
        recdeinter(&coded[sptr], codedlen[i], codedlen[i], ibuf, clen-codedlen[i], ibuf);
        sptr += codedlen[i];
        clen -= codedlen[i];
      }else if(Interleave == 2 && hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 0){ /*Intra w/o Inter */
        recdeinter(&coded[sptr], codedlen[i], PINTDEPTH, ibuf, clen-codedlen[i], ibuf);
        sptr += codedlen[i];
        clen -= codedlen[i];
        
      } else if((Interleave == 1 || Interleave == 2) && (hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 1 || hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 2)){
        recdeinterBytewise(&coded[sptr], codedlen[i], hEpConfig->epConfig.fattr[pred].cattr[i].ClassCodeRate, ibuf, clen-codedlen[i], ibuf);
        sptr += codedlen[i];
        clen -= codedlen[i];
      }else if(Interleave == 3){ 
        sptr += codedlen[i];
      }  
    }
    
    /* insert data for ClassInterleave == 3 */
    sptr=0;
    for(i=jcnt=0;i<nclass;i++){
      if(hEpConfig->epConfig.fattr[pred].cattr[i].ClassInterleave == 3){
        for(j=0;j<codedlen[i];j++){
          coded[sptr++]=ibuf[j+jcnt];
        }
        jcnt+= codedlen[i];
      }else{
        sptr+=codedlen[i];
      }
    }
    
    clen = sptr;
    free(tmp1); free(tmp2);
   
  } else { /* Not Interleaved */
    hlen = RetrieveHeader(&hEpConfig->epConfig, &pred, trans, &nstuff);
    translen -= nstuff;
    nclass = hEpConfig->epConfig.fattr[pred].ClassCount;
    
    {
      int i, from, to, rate;
      int SRCPCStart = 0;
      clen = from = 0;
      bound_punc[0] = 0;
      for(i=0; i<nclass-1; i++){

        to = from + hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount +
            (hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount?hEpConfig->epConfig.fattr[pred].cattr[i].ClassCRCCount:0);

        if(hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 0 && hEpConfig->epConfig.fattr[pred].cattr[i].TermSW){
          bound_punc[i+1] = bound_punc[i] + to - from;
          if((bound_punc[i+1] - bound_punc[SRCPCStart]) && hEpConfig->epConfig.fattr[pred].cattr[i].ClassCodeRate){
            bound_punc[i+1] += 4;
          }
          SRCPCStart = i+1;
        }else{
          bound_punc[i+1] = bound_punc[i];
        }

        rate = hEpConfig->epConfig.fattr[pred].cattr[i].ClassCodeRate;
        if(hEpConfig->epConfig.fattr[pred].cattr[i].FECType == 0){ /*SRCPC*/
          {
            int tmp, m, tmpClassRate;
            
            tmp = 8 - bound_punc[i]%8;
            if(tmp == 8){
              tmp = 0;
            }
            
            if(tmp){
              if(tmp > bound_punc[i+1]-bound_punc[i]){
                tmp = bound_punc[i+1]-bound_punc[i];
              }
              
              tmpClassRate = 0;                   /* to prevent reading of not initialized mem */
              for(m=i-1; m>=0; m--){
                if(hEpConfig->epConfig.fattr[pred].cattr[m].FECType == 0){
                  tmpClassRate = hEpConfig->epConfig.fattr[pred].cattr[m].ClassCodeRate;
                  if(bound_punc[i]-bound_punc[m] >= bound_punc[i]%8){
                    break;
                  }
                }
              }
              clen += SRCPCDepuncCount(from, from+tmp, 0, tmpClassRate);
            }
            
            clen -= SRCPCDepuncCount(from + tmp, to, 0, hEpConfig->epConfig.fattr[pred].cattr[i].ClassCodeRate); /* Calculate length of SRCPC encode class */
          }
          
        /*  clen += SRCPCDepuncCount(from, to, 0, rate); */
        } else{                  /* RS */
          int rsconcat;
          rsconcat = 0;
          for(j=i; hEpConfig->epConfig.fattr[pred].cattr[j].FECType == 2; j++, rsconcat ++) ;
          for(j=0; j<rsconcat; j++){
            to += hEpConfig->epConfig.fattr[pred].cattr[i+j+1].ClassBitCount +  
              (hEpConfig->epConfig.fattr[pred].cattr[i+j+1].ClassBitCount?hEpConfig->epConfig.fattr[pred].cattr[i+j+1].ClassCRCCount:0);
          }
          clen += RSEncodeCount(from, to, rate);
          i += rsconcat;
        }
        from = to;
      }
    }
    
    if ( hlen == -1) { /*check header-length*/
      EPTool_NullSideInfoWrite(outBuffer, &hEpConfig->epConfig, 0, check, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, 1);
      return -1;
    }

     if ( translen < clen + hlen ) {
      EPTool_NullSideInfoWrite(outBuffer, &hEpConfig->epConfig, 0, check, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, 1);
      return -1;
    }

    clen =translen - hlen;
    for(i=0; i<clen; i++)
      coded[i] = (int)(unsigned)trans[i+hlen];
  }

  payloadlen = PayloadDecode(data, coded, clen, &hEpConfig->epConfig.fattr[pred], check);

  if(hEpConfig->epConfig.fattr[pred].ClassReorderedOutput){
    int bound[CLASSMAX+1];
    int origOutputIndex[CLASSMAX];
    int origNClass;
    bound[0] = 0;
    for(i=0; i<nclass; i++){
      bound[i+1] = bound[i] + hEpConfig->epConfig.fattr[pred].cattr[i].ClassBitCount;
    }

    origNClass = EPTool_OrigClassOutputIndex(&hEpConfig->epConfig, pred, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, origOutputIndex); 
    
    for(i=nclass;i<origNClass+1;i++){
      bound[i] = bound[nclass];
    }

    for(i=0; i<origNClass; i++){
      j = origOutputIndex[i];
      if (j == -1)
        continue;
      if (bitFlag){
        int k; 
        for (k=0; k<bound[j+1]-bound[j]; k++)
          outBuffer->classBuffer[i][k] = data[bound[j]+k];
      } else {
      EPTool_BitsToBytes(outBuffer->classBuffer[i], &data[bound[j]], ((bound[j+1]-bound[j] + 7)/8) * 8); 
      }
    }
    
    EPTool_SideInfoWrite(outBuffer, &hEpConfig->epConfig, pred, check, 1, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, 1);

  }else{
    int offset = 0;
    
    EPTool_SideInfoWrite(outBuffer, &hEpConfig->epConfig, pred, check, 1, &hEpConfig->origEpConfig, hEpConfig->classAvailableTable, hEpConfig->classAvailableBits, 1);
    
    for(i=0; i<outBuffer->nrValidBuffer; i++){
      if (bitFlag){
        int k; 
        for (k=0; k<outBuffer->classLength[i]; k++)
          outBuffer->classBuffer[i][k] = data[offset+k];
      } else {
      EPTool_BitsToBytes(outBuffer->classBuffer[i], &data[offset], ((outBuffer->classLength[i] + 7)/8) * 8);
      }
      offset += outBuffer->classLength[i];
    }
  }

  return 0;
}


HANDLE_CLASS_FRAME EPTool_GetClassFrame ( HANDLE_EP_CONFIG hEpConfig, int bitFlag)
{
  HANDLE_CLASS_FRAME tmpHandle;
  int scaling = 1;
  
  if(bitFlag)
    scaling = 8;

  tmpHandle = (HANDLE_CLASS_FRAME) calloc (1, sizeof ( T_CLASS_FRAME));
  
  tmpHandle->nrValidBuffer  = EpTool_GetNumberOfClasses(hEpConfig);
  tmpHandle->maxClasses     = CLASSMAX;
  tmpHandle->maxBufferSize  = (MAXBITS * scaling / 8);

  tmpHandle->classLength    = hEpConfig->classLength;
  tmpHandle->classCRCLength = hEpConfig->classCRCLength;
  tmpHandle->classCodeRate  = hEpConfig->classCodeRate;
  tmpHandle->crcError       = hEpConfig->crcError;

  tmpHandle->classBuffer    = hEpConfig->classBuffer;

  return tmpHandle;
}

int EPTool_Destroy( HANDLE_EP_CONFIG hEpConfig )
{
  int pred;

  FrameHandle_Destroy( hEpConfig );
  
  for (pred = 0; pred < hEpConfig->epConfig.NPred; pred++) {
    if (hEpConfig->epConfig.fattr[pred].cattr)
      free(hEpConfig->epConfig.fattr[pred].cattr);
    
    if (hEpConfig->epConfig.fattr[pred].cattresc)
      free(hEpConfig->epConfig.fattr[pred].cattresc);

    if (hEpConfig->epConfig.fattr[pred].lenbit)
      free(hEpConfig->epConfig.fattr[pred].lenbit);

  }

  for (pred = 0; pred < hEpConfig->origEpConfig.NPred; pred++) {

    if (hEpConfig->origEpConfig.fattr[pred].cattr) 
      free(hEpConfig->origEpConfig.fattr[pred].cattr); 
    
    if (hEpConfig->origEpConfig.fattr[pred].cattresc) 
      free(hEpConfig->origEpConfig.fattr[pred].cattresc);

    if (hEpConfig->origEpConfig.fattr[pred].lenbit) 
      free(hEpConfig->origEpConfig.fattr[pred].lenbit); 

  }

  if (hEpConfig->epConfig.fattr)
    free(hEpConfig->epConfig.fattr);

  if (hEpConfig->origEpConfig.fattr)
    free(hEpConfig->origEpConfig.fattr);

  if (hEpConfig)
    free(hEpConfig);

  return 0;
}

static void EPTool_BytesToBits(unsigned char *inbuf, int *trans, int translen)
{
  int i, bitrem = 0;
  unsigned char tmp = 0;

  for(i=0; i< translen; i++){
    if(bitrem == 0){
      tmp = inbuf[i/8] ;
      bitrem = 8;
    }
    trans[i] = ((tmp & 0x80) == 0x80);
    tmp <<= 1;
    bitrem --;
  }
}

static int EPTool_BitsToBytes(unsigned char *outbuf, int *trans, int translen )
{
  int bitrem = 0;
  unsigned char tmp = 0;
  int i;

  if(translen == -1){		/* Flash */
    if(bitrem == 0) return bitrem;
    tmp <<= (8 - bitrem);
    outbuf[0] = tmp; /* ?????? dh */ 
    /* fwrite(&tmp, sizeof(char), 1, fpout); */
    return bitrem;
  }
  
  for(i=0; i<translen; i++){
    tmp <<= 1;
    tmp |= *(trans++);
    bitrem ++;
    if(bitrem == 8){
      outbuf[i/8] = tmp;
      /* fwrite(&tmp, sizeof(char), 1, fpout); */
      bitrem = 0;
    }
  }

  return bitrem;
}

static int EpTool_GetNumberOfClasses(HANDLE_EP_CONFIG  hEpConfig)
{
  int classmax,  origpred = hEpConfig->origEpConfig.NPred - 1; 
  classmax = hEpConfig->epConfig.fattr[origpred].ClassCount ;
  return classmax;
}

int FrameHandle_Create( HANDLE_EP_CONFIG inFrame, int bitFlag)
{
  int i, classmax, origpred = inFrame->origEpConfig.NPred -1;
  int scaling = 1;
  
  if (bitFlag)
    scaling = 8;

  classmax =  inFrame->epConfig.fattr[origpred].ClassCount;

  inFrame->classLength    = (long*)calloc (classmax, sizeof(long));
  inFrame->classCRCLength = (long*)calloc (classmax, sizeof(long));
  inFrame->classCodeRate  = (long*)calloc (classmax, sizeof(long));

  inFrame->classBuffer    = (unsigned char**)calloc (classmax * (MAXBITS*scaling/8), sizeof(char));
  
  for  (i = 0; i < classmax; i++) {
    inFrame->classBuffer[i] = (unsigned char*)calloc (MAXBITS*scaling/8, sizeof(char));
  }

  inFrame->crcError       = (short*)calloc (classmax, sizeof(short));

  return 0;
}

int FrameHandle_Destroy( HANDLE_EP_CONFIG inFrame )
{
  int i, classmax;
  int origpred = 0 ; /* BUG BUG */

  classmax = inFrame->epConfig.fattr[origpred].ClassCount;

  if (inFrame->classLength )
    free (inFrame->classLength );

  if (inFrame->classCRCLength)
    free(inFrame->classCRCLength);

  if (inFrame->classCodeRate)
    free(inFrame->classCodeRate);

  for (i= 0; i < classmax; i++) {
    if (inFrame->classBuffer[i])
      free(inFrame->classBuffer[i]);
  }

  if (inFrame->classBuffer)
    free (inFrame->classBuffer); 

  if (inFrame->crcError)
    free (inFrame->crcError); 

  return 0;
}


static void EPTool_SideInfoWrite(HANDLE_CLASS_FRAME outBuffer, EPConfig *modepcfg, int modpred, int modcheck[], int reorderFlag, EPConfig *origepcfg, int CAtbl[MAXPREDwCA][CLASSMAX], int CAbits[MAXPREDwCA], int integrated)
{
  int i,j;
  EPFrameAttrib *fattr;
  int nclass;
  int order[CLASSMAX] = {0} ;
  

  if ( integrated ){
    int k;
    int revPred[MAXPREDwCA];
    int revCA[MAXPREDwCA];
    int origcheck[CLASSMAX];
    int CA, skipped;
    int modnpred, origpred;
    int CAptr[MAXPREDwCA][CLASSMAX];
    
    
    reorderFlag = 0; 
    /* Count the number of Pred-set after conversion */
    modnpred = 0;
    for(i=0; i<origepcfg->NPred; i++){
      int ptr;
      ptr = 0;
      for(j=0; j< origepcfg->fattr[i].ClassCount; j++){
        if(CAtbl[i][j] == 1){ /* CA switch */
          CAptr[i][ptr] = j; /* CAptr[A][B] :Class for  A-th pred, B-th switch */
          ptr++;
        }
      }
      
      for(j=0; j< (1<<CAbits[i]); j++){  /* #modepred for this origpred *//* 1 << CAbits[i] corresponds to 2^NCO[i] in standard */ 
        CA = 0;
        for(k=0; k<CAbits[i]; k++) /* CAbits[i] corresponds to NCO[i] in standard */
          if(j & (0x01 << k))
            CA |= (0x01 << CAptr[i][k]);
        revPred[modnpred] = i;
        revCA[modnpred] = CA;
        modnpred ++;
        
        if(modnpred > MAXPREDwCA){
          /* fprintf(stderr, "Error: number of predifined sets is to high, please increase MAXPREDwCA in eptool.h\n") ; */
          exit(1);
        }
      }
    }
    
    origpred = revPred[modpred];
    CA = revCA[modpred];
    skipped = 0;
    
    if (modepcfg->fattr[modpred].ClassReorderedOutput) {
      for (i=0; i<modepcfg->fattr[modpred].ClassCount; i++) {
        for (j = 0; j <modepcfg->fattr[modpred].ClassCount; j++) { 
          if (modepcfg->fattr[modpred].ClassOutputOrder[j] == modepcfg->fattr[modpred].SortedClassOutputOrder[i])
            order[i] = j;
        }
        /*fprintf(stderr,"wrote class %d, length %d\n",order[i],fattr->cattr[order[i]].ClassBitCount); */
      }
    }
    
    for(i=0; i<origepcfg->fattr[origpred].ClassCount; i++){
      /*      printf("pred %d -> %d,  revCA[%d] = %d\n", origpred, modpred, modpred, revCA[modpred]);*/ 
      if ( origepcfg->fattr[origpred].ClassReorderedOutput ){
        k = order[i-skipped];
      }
      else{ 
        k = i-skipped;
      }
      
      if(CAtbl[origpred][i] == 1 && (revCA[modpred] & (1<<i))){  
        origepcfg->fattr[origpred].cattr[i].ClassBitCount = 0;  
        origcheck[i] = 0;  
        skipped += 1;  
      }else{ 
        origcheck[i] = modcheck[k]; 
        origepcfg->fattr[origpred].cattr[i].ClassBitCount = 
          modepcfg->fattr[modpred].cattr[k].ClassBitCount; 
        origepcfg->fattr[origpred].cattr[i].ClassCodeRate = 
          modepcfg->fattr[modpred].cattr[k].ClassCodeRate; 
        origepcfg->fattr[origpred].cattr[i].ClassCRCCount = 
          modepcfg->fattr[modpred].cattr[k].ClassCRCCount;
      }
      
    }
    modpred = origpred;
    for (i=0; i<modepcfg->fattr[modpred].ClassCount; i++){
      modcheck[i] = origcheck[i];
    }
    if (origepcfg->NPred > 1){
      /* fprintf(outfp, "%d\n", modpred); */
    }
    fattr = &origepcfg->fattr[origpred];
  }
  else{
    /* Pred Signaling */
    if(modepcfg->NPred > 1){
      /* fprintf(outfp, "%d\n", modpred); */
    }
    fattr = &modepcfg->fattr[modpred];
  }
  
  nclass = fattr->ClassCount;
  
  if (fattr->ClassReorderedOutput && reorderFlag) {
    for (i=0; i<nclass; i++) {
      for (j = 0; j <nclass; j++) { 
        if (fattr->ClassOutputOrder[j] == fattr->SortedClassOutputOrder[i])
          order[i] = j;
      }
      /*fprintf(stderr,"wrote class %d, length %d\n",order[i],fattr->cattr[order[i]].ClassBitCount); */
    }
  }
  for(i=0; i<nclass; i++) {
    if (fattr->ClassReorderedOutput && reorderFlag)
      j = order[i];
    else
      j = i;
    
    outBuffer->crcError[i] = modcheck[j];

  }

  for(i=0; i<nclass; i++) {
    if (fattr->ClassReorderedOutput && reorderFlag)
      j = order[i];
    else
      j = i;

    /* set class length */
    outBuffer->classLength[i]    = fattr->cattr[j].ClassBitCount;
 
    if(fattr->cattresc[j].ClassBitCount){ /* Class Length ESC */
      outBuffer->classLength[i] = fattr->cattr[j].ClassBitCount;
    }
    if(fattr->cattresc[j].ClassCodeRate){ /* SRCPC rate ESC */
      outBuffer->classCodeRate[i] = fattr->cattr[j].ClassCodeRate;
    }
    if(fattr->cattresc[j].ClassCRCCount){ /* Class Length ESC */
      outBuffer->classCRCLength[i] = fattr->cattr[j].ClassCRCCount;
    }
  }
}

static void EPTool_NullSideInfoWrite(HANDLE_CLASS_FRAME outBuffer, EPConfig *modepcfg, int modpred, int modcheck[], EPConfig *origepcfg, int CAtbl[MAXPREDwCA][CLASSMAX], int CAbits[MAXPREDwCA], int integrated)
{
  int i;
  EPFrameAttrib *fattr;
  int nclass;

#if 1 /* why do we need all this stuff ??? d.h 001205 */ /* don't set the #if-#endif brackets this way - we need at least an init fattr. I'll check the rest later 010307 multrums@iis.fhg.de */
  if ( integrated ){
    int j,k;
    int revPred[MAXPREDwCA];
    int revCA[MAXPREDwCA];
    int origcheck[CLASSMAX];
    int CA, skipped;
    int modnpred, origpred;
    int CAptr[MAXPREDwCA][CLASSMAX];
    int order[CLASSMAX] = {0} ;

    
    /* Count the number of Pred-set after conversion */
    modnpred = 0;
    for(i=0; i<origepcfg->NPred; i++){
      int ptr;
      ptr = 0;
      for(j=0; j< origepcfg->fattr[i].ClassCount; j++){
        if(CAtbl[i][j] == 1){ /* CA switch */
          CAptr[i][ptr] = j; /* CAptr[A][B] :Class for  A-th pred, B-th switch */
          ptr++;
        }
      }
      
      for(j=0; j< (1<<CAbits[i]); j++){  /* #modepred for this origpred */ 
        CA = 0;
        for(k=0; k<CAbits[i]; k++)
          if(j & (0x01 << k))
            CA |= (0x01 << CAptr[i][k]);
        revPred[modnpred] = i;
        revCA[modnpred] = CA;
        modnpred ++;
        
        if(modnpred > MAXPREDwCA){
          fprintf(stderr, "EPTool: EPTool_NullSideInfoWrite(): Error: number of predifined sets is to high, please increase MAXPREDwCA in eptool.h\n") ;
          exit(1);
        }
      }
    }
    
    origpred = revPred[modpred];
    CA = revCA[modpred];
    skipped = 0;
    
    if (modepcfg->fattr[modpred].ClassReorderedOutput) {
      for (i=0; i<modepcfg->fattr[modpred].ClassCount; i++) {
        for (j = 0; j <modepcfg->fattr[modpred].ClassCount; j++) { 
          if (modepcfg->fattr[modpred].ClassOutputOrder[j] == modepcfg->fattr[modpred].SortedClassOutputOrder[i])
            order[i] = j;
        }
        /*fprintf(stderr,"wrote class %d, length %d\n",order[i],fattr->cattr[order[i]].ClassBitCount); */
      }
    }
    
    for(i=0; i<origepcfg->fattr[origpred].ClassCount; i++){
      /*      printf("pred %d -> %d,  revCA[%d] = %d\n", origpred, modpred, modpred, revCA[modpred]);*/ 
      if ( origepcfg->fattr[origpred].ClassReorderedOutput ){
        k = order[i-skipped];
      }
      else { 
        k = i-skipped;
      }
      
      if(CAtbl[origpred][i] == 1 && (revCA[modpred] & (1<<i))){  
        origepcfg->fattr[origpred].cattr[i].ClassBitCount = 0;  
        origcheck[i] = 0;  
        skipped += 1;  
      }else{ 
        origcheck[i] = modcheck[k]; 
        origepcfg->fattr[origpred].cattr[i].ClassBitCount = 
          modepcfg->fattr[modpred].cattr[k].ClassBitCount; 
        origepcfg->fattr[origpred].cattr[i].ClassCodeRate = 
          modepcfg->fattr[modpred].cattr[k].ClassCodeRate; 
        origepcfg->fattr[origpred].cattr[i].ClassCRCCount = 
          modepcfg->fattr[modpred].cattr[k].ClassCRCCount;
      }
      
    }
    modpred = origpred;
    for (i=0; i<modepcfg->fattr[modpred].ClassCount; i++){
      modcheck[i] = origcheck[i];
    }
    if (origepcfg->NPred > 1){
      /* fprintf(outfp, "%d\n", modpred); */
    }
    fattr = &origepcfg->fattr[origpred];
  } else {
    /* Pred Signaling */
    if(modepcfg->NPred > 1){
      /* fprintf(outfp, "%d\n", modpred); */
    }
    fattr = &modepcfg->fattr[modpred];
  }
#endif
  
  nclass = origepcfg->fattr[origepcfg->NPred-1].ClassCount;

  for(i=0; i<nclass; i++)
    outBuffer->crcError[i] = 1;

  for(i=0; i<nclass; i++){
    if(fattr->cattresc[i].ClassBitCount){ /* Class Length ESC */
      outBuffer->classLength[i] = 0;
    }
    if(fattr->cattresc[i].ClassCodeRate){ /* SRCPC rate ESC */
      outBuffer->classCodeRate[i] = fattr->cattr[i].ClassCodeRate;
    }
    if(fattr->cattresc[i].ClassCRCCount){ /* Class CRC ESC */
      outBuffer->classCRCLength[i] = fattr->cattr[i].ClassCRCCount;
    }
  }
}


static int EPTool_OrigClassOutputIndex(EPConfig *modepcfg, int modpred, EPConfig *origepcfg, int CAtbl[MAXPREDwCA][CLASSMAX], int CAbits[MAXPREDwCA], int origOutputIndex[CLASSMAX])
{
  int i,j;
  int order[CLASSMAX] = {0} ;
  
  int k;
  int revPred[MAXPREDwCA];
  int revCA[MAXPREDwCA];
  int CA, skipped;
  int modnpred, origpred;
  int CAptr[MAXPREDwCA][CLASSMAX];
  
  
  
  /* Count the number of Pred-set after conversion */
  modnpred = 0;
  for(i=0; i<origepcfg->NPred; i++){
    int ptr;
    ptr = 0;
    for(j=0; j< origepcfg->fattr[i].ClassCount; j++){
      if(CAtbl[i][j] == 1){ /* CA switch */
        CAptr[i][ptr] = j; /* CAptr[A][B] :Class for  A-th pred, B-th switch */
        ptr++;
      }
    }
    
    for(j=0; j< (1<<CAbits[i]); j++){  /* #modepred for this origpred */ 
      CA = 0;
      for(k=0; k<CAbits[i]; k++)
        if(j & (0x01 << k))
          CA |= (0x01 << CAptr[i][k]);
      revPred[modnpred] = i;
      revCA[modnpred] = CA;
      modnpred ++;
      
      if(modnpred > MAXPREDwCA){
        /* fprintf(stderr, "Error: number of predifined sets is to high, please increase MAXPREDwCA in eptool.h\n") ; */
        exit(1);
      }
    }
  }
  
  origpred = revPred[modpred];
  CA = revCA[modpred];
    skipped = 0;
    
    if (modepcfg->fattr[modpred].ClassReorderedOutput) {
      for (i=0; i<modepcfg->fattr[modpred].ClassCount; i++) {
        for (j = 0; j <modepcfg->fattr[modpred].ClassCount; j++) { 
          if (modepcfg->fattr[modpred].ClassOutputOrder[j] == modepcfg->fattr[modpred].SortedClassOutputOrder[i])
            order[i] = j;
        }
      }
    }
    
    for(i=0; i<origepcfg->fattr[origpred].ClassCount; i++){
      if ( origepcfg->fattr[origpred].ClassReorderedOutput ){
        k = order[i-skipped];
      }
      else{ 
        k = i-skipped;
      }
      
      if(CAtbl[origpred][i] == 1 && (revCA[modpred] & (1<<i))){  
        skipped += 1;  
        origOutputIndex[i] = -1;
      }else{ 
        origOutputIndex[i] = k;
      }
    }
  return origepcfg->fattr[origpred].ClassCount;
}


int EPTool_CreateBufferEPSpecificConfig(unsigned char **outBuffer, const char *predFileName){

  int npred;
  char buffer[1024];
  FILE *fp;

  if (NULL == (fp = fopen(predFileName, "r"))) {
    fprintf(stderr,"EPTool: EPTool_CreateExpl(): Error: predefinition file not found\n");
    exit(-1);
  }

  if(NULL == fgets(buffer, sizeof(buffer), fp)) {
    fprintf(stderr, "EP_Tool: EPTool_CreateBufferEPSpecificConfig(): libeptool.c: #pred NULL");
  }

  if(1 != sscanf(buffer, "%d", &npred)) {
    fprintf(stderr, "EP_Tool: EPTool_CreateBufferEPSpecificConfig(): libeptool.c: Error: Please check PredefinitionFile!");
  }

  if (!npred){
    fprintf (stderr, "EP_Tool: EPTool_CreateBufferEPSpecificConfig(): libeptool.c: Error: Please set number_of_predefined_set unequal 0!");
    exit(-1);
  }

  if (NULL == ((*outBuffer) = (unsigned char *)malloc((32+npred*429)*sizeof(char)))){
    fprintf(stderr, "EPTool: EPTool_CreateBufferEPSpecificConfig(): libeptool.c: Failed malloc for outBuffer");
    exit(-1);
  }

  fclose(fp);
  return 0;

}


int EPTool_DestroyBufferEPSpecificConfig(unsigned char **outBuffer){
  
  free(*outBuffer);
  return 0;

}


static long int EPTool_WriteEPSpecificConfig(unsigned char *outBuffer, HANDLE_EP_CONFIG hEpConfig){

  int i, j, k, offset;
  unsigned char tmp[16];
  int *tmpBuffer;
  unsigned int aux;

  offset = 0;


  if (NULL == (tmpBuffer = (int *)calloc(1, (32+hEpConfig->origEpConfig.NPred*429)*sizeof(int)))){
    fprintf(stderr, "EPTool: EPTool_WriteEPSpecificConfig(): libeptool.c: Failed malloc for tmpBuffer");
    exit(-1);
  }

  EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.NPred, tmp, 8);
  for (k=0; k<8; k++){
    tmpBuffer[k] = tmp[k];
  }
  offset += k;

  EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.Interleave, tmp, 2);
  for (k=0; k<2; k++){
    tmpBuffer[offset+k] = tmp[k];
  }
  offset += k;

  EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.BitStuffing, tmp, 3);
  for (k=0; k<3; k++){
    tmpBuffer[offset+k] = tmp[k];
  }
  offset += k;
  
  EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.NConcat, tmp, 3);
  for (k=0; k<3; k++){
    tmpBuffer[offset+k] = tmp[k];
  }
  offset += k;


  for(i=0; i<hEpConfig->origEpConfig.NPred; i++){
    
    EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.fattr[i].ClassCount, tmp, 6);
    for (k=0; k<6; k++){
      tmpBuffer[offset+k] = tmp[k];
    }
    offset += k;

    for(j=0; j<hEpConfig->origEpConfig.fattr[i].ClassCount; j++){
      
      switch(hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassBitCount){
      case 0:
        tmpBuffer[offset++] = 0;
        break;
      case 1:
        tmpBuffer[offset++] = 1;
          break;
      }

      switch(hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCodeRate){
      case 0:
        tmpBuffer[offset++] = 0;
        break;
      case 1:
        tmpBuffer[offset++] = 1;
        break;
      }

      switch(hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCRCCount){
      case 0:
        tmpBuffer[offset++] = 0;
        break;
      case 1:
        tmpBuffer[offset++] = 1;
        break;
      }

      if(hEpConfig->origEpConfig.NConcat != 1){
        switch(hEpConfig->origEpConfig.fattr[i].cattr[j].Concatenate){
        case 0:
          tmpBuffer[offset++] = 0;
          break;
        case 1:
          tmpBuffer[offset++] = 1;
          break;
        }
      }

      EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.fattr[i].cattr[j].FECType, tmp, 2);
      for(k=0; k<2; k++){
        tmpBuffer[offset+k] = tmp[k];
      }
      offset += k;
   
      if(!hEpConfig->origEpConfig.fattr[i].cattr[j].FECType){
        switch(hEpConfig->origEpConfig.fattr[i].cattr[j].TermSW){
        case 0:
          tmpBuffer[offset++] = 0;
          break;
        case 1:
          tmpBuffer[offset++] = 1;
          break;
        }
      }

      if(hEpConfig->origEpConfig.Interleave == 2){
        EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.fattr[i].cattr[j].ClassInterleave, tmp, 2);
        for(k=0; k<2; k++){
          tmpBuffer[offset+k] = tmp[k];
        }
        offset += k;
      }

      switch(hEpConfig->classAvailableTable[i][j]){
      case 0:
        tmpBuffer[offset++] = 0;
        break;
      case 1:
        tmpBuffer[offset++] = 1;
        break;
      }

      if(hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassBitCount){
        EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.fattr[i].lenbit[j], tmp, 4);
        for(k=0; k<4; k++){
          tmpBuffer[offset+k] = tmp[k];
        }
        offset += k;
      } else {
        EPTool_BytesToBits_IntToChar(&hEpConfig->origEpConfig.fattr[i].cattr[j].ClassBitCount, tmp, 16);
        for(k=0; k<16; k++){
          tmpBuffer[offset+k] = tmp[k];
        }
        offset += k;
      }
      
      if(!hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCodeRate){
        if(hEpConfig->origEpConfig.fattr[i].cattr[j].FECType){
          EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCodeRate, tmp, 7);
          for(k=0; k<7; k++){
            tmpBuffer[offset+k] = tmp[k];
          }
        } else {
          EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCodeRate, tmp, 5);
          for(k=0; k<5; k++){
            tmpBuffer[offset+k] = tmp[k];
          }
        }
        offset += k;
      }

      if(!hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCRCCount){
        switch(hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCRCCount){
        case 24:
          aux = 17;
          break;
        case 32:
          aux = 18;
          break;
        default: 
          aux = hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCRCCount;
          break;
        }
        EPTool_BytesToBits_IntToChar(&aux, tmp, 5);
        for(k=0; k<5; k++){
          tmpBuffer[offset+k] = tmp[k];
        }
        offset += k;
      }
    }

    switch(hEpConfig->origEpConfig.fattr[i].ClassReorderedOutput){
    case 0:
      tmpBuffer[offset++] = 0;
      break;
    case 1:
      tmpBuffer[offset++] = 1;
    }
    
    if(hEpConfig->origEpConfig.fattr->ClassReorderedOutput){
      for(j=0; j<hEpConfig->origEpConfig.fattr[i].ClassCount; j++){
        EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.fattr[i].ClassOutputOrder[j], tmp, 6);
        for(k=0; k<6; k++){
          tmpBuffer[offset+k] = tmp[k];
        }
        offset += k;
      }
    }
  }

  switch(hEpConfig->origEpConfig.HeaderProtectExtend){
  case 0:
    tmpBuffer[offset++] = 0;
    break;
  case 1:
    tmpBuffer[offset++] = 1;
    break;
  }

  if(hEpConfig->origEpConfig.HeaderProtectExtend){
    EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.HeaderRate, tmp, 5);
    for(k=0; k<5; k++){
      tmpBuffer[offset+k] = tmp[k];
    }
    offset += k;

    switch(hEpConfig->origEpConfig.HeaderCRC){
    case 24:
      aux = 17;
      break;
    case 32:
      aux = 18;
      break;
    default:
      aux = hEpConfig->origEpConfig.HeaderCRC;
      break;
    }
    EPTool_BytesToBits_IntToChar(&aux, tmp, 5);
    for(k=0; k<5; k++){
      tmpBuffer[offset+k] = tmp[k];
    }
    offset += k;
  }

#if RS_FEC_CAPABILITY
  EPTool_BytesToBits_IntToChar((unsigned int *)&hEpConfig->origEpConfig.RSFecCapability, tmp, 7);
  for(k=0; k<7; k++){
    tmpBuffer[offset+k] = tmp[k];
  }
  offset += k;
#endif

  EPTool_BitsToBytes(outBuffer, tmpBuffer, ((offset+7)/8)*8);
 
  free(tmpBuffer);

  return offset;
}


static int EPTool_ParseEPSpecificConfig(unsigned char *inBuffer, long int configLength, HANDLE_EP_CONFIG hEpConfig){

  int i, j;
  int offset = 0;
  int npred = 0;
  int nclass = 0;
  int tmp[1];
  unsigned char *tmpBuffer = NULL;

  hEpConfig->origEpConfig.NPred =  npred = inBuffer[0];
  offset += 8;

  if(NULL == (hEpConfig->origEpConfig.fattr = (EPFrameAttrib *)calloc(npred, sizeof(EPFrameAttrib)))){
    fprintf(stderr, "EP_Tool: EPTool_ParseEPSpecificConfig(): libeptool.c: Unable to calloc for EPFrameAttrib\n");
    exit(1);
  }

  if (NULL == (tmpBuffer = (unsigned char *)malloc(configLength*sizeof(char)))){
    fprintf(stderr, "EPTool: EPTool_ParseEPSpecificConfig: libeptool.c: Failed malloc tmpBuffer");
    exit(-1);
  }

  EPTool_BytesToBits_CharToChar(inBuffer, tmpBuffer, configLength);
      
  EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 2);
  hEpConfig->origEpConfig.Interleave = *tmp;
  offset += 2;

  EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 3);
  hEpConfig->origEpConfig.BitStuffing = *tmp;
  offset += 3;

  EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 3);
  hEpConfig->origEpConfig.NConcat = *tmp;
  offset += 3;

  for (i=0; i<hEpConfig->origEpConfig.NPred; i++){
    
    EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 6);
    hEpConfig->origEpConfig.fattr[i].ClassCount = nclass = *tmp;
    offset += 6;

    if(NULL == (hEpConfig->origEpConfig.fattr[i].cattr = (ClassAttribContent *)calloc(nclass, sizeof(ClassAttribContent)))){
      fprintf(stderr, "EP_Tool: EPTool_ParseEPSpecificConfig(): libeptool.c: Unable to calloc for Cattr\n");
      exit(1);
    }
    if(NULL == (hEpConfig->origEpConfig.fattr[i].cattresc = (ClassAttribContent *)calloc(nclass, sizeof(ClassAttribContent)))){
      fprintf(stderr, "EP_Tool: EPTool_ParseEPSpecificConfig(): libeptool.c: Unable to calloc for CattrESC\n");
      exit(1);
    }
    if(NULL == (hEpConfig->origEpConfig.fattr[i].lenbit = (int *)calloc(nclass, sizeof(int)))){
      fprintf(stderr, "EP_Tool: EPTool_ParseEPSpecificConfig(): libeptool.c: Unable to calloc for CattrESC\n");
      exit(1);
    }
    

    for (j=0; j<hEpConfig->origEpConfig.fattr[i].ClassCount; j++){

      EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
      hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassBitCount = *tmp;
      offset += 1;

      EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
      hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCodeRate = *tmp;
      offset += 1;

      EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
      hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCRCCount = *tmp;
      offset += 1;

      if (hEpConfig->origEpConfig.NConcat !=1){
        EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
        hEpConfig->origEpConfig.fattr[i].cattr[j].Concatenate = *tmp;
        offset += 1;
      }

      EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 2);
      hEpConfig->origEpConfig.fattr[i].cattr[j].FECType = *tmp;
      offset += 2;

      if (!hEpConfig->origEpConfig.fattr[i].cattr[j].FECType){
        EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
        hEpConfig->origEpConfig.fattr[i].cattr[j].TermSW = *tmp;
        offset += 1;
      }

      if (hEpConfig->origEpConfig.Interleave == 2){
        EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 2);
        hEpConfig->origEpConfig.fattr[i].cattr[j].ClassInterleave = *tmp;
        offset += 2;
      }

      EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
      hEpConfig->classAvailableTable[i][j] = *tmp;
      hEpConfig->classAvailableBits[i] += *tmp;
      offset += 1;
      

      if (hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassBitCount){
        EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 4);
        hEpConfig->origEpConfig.fattr[i].lenbit[j] = *tmp;
        offset += 4; 
      } else {
        EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 16);
        hEpConfig->origEpConfig.fattr[i].cattr[j].ClassBitCount = *tmp;
        offset += 16;
      }

      if (!hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCodeRate){
        if(hEpConfig->origEpConfig.fattr[i].cattr[j].FECType){
          EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 7);
          hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCodeRate = *tmp;
          offset += 7;
        } else {
          EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 5);
          hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCodeRate = *tmp;
          offset += 5;
        }
      }
        
       
      if (!hEpConfig->origEpConfig.fattr[i].cattresc[j].ClassCRCCount){
        EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 5);
        switch(*tmp){ /* handle right crc-len */
        case 17:
          hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCRCCount = 24;
          break;
        case 18:
          hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCRCCount = 32;
          break;
        default: 
          hEpConfig->origEpConfig.fattr[i].cattr[j].ClassCRCCount = *tmp;
          break;
        }
        offset += 5;
      }
    }
      
    EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
    hEpConfig->origEpConfig.fattr[i].ClassReorderedOutput = *tmp;
    offset += 1;

    if (hEpConfig->origEpConfig.fattr[i].ClassReorderedOutput){
      for (j=0; j<hEpConfig->origEpConfig.fattr[i].ClassCount; j++){
        EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 6);
        hEpConfig->origEpConfig.fattr[i].ClassOutputOrder[j] = *tmp;
        offset += 6;
      }
    }
  }

  EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 1);
  hEpConfig->origEpConfig.HeaderProtectExtend = *tmp;
  offset += 1;

  if (hEpConfig->origEpConfig.HeaderProtectExtend){
    EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 5);
    hEpConfig->origEpConfig.HeaderRate = *tmp;
    offset += 5;

    EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 5);
    switch(*tmp){
    case 17:
      hEpConfig->origEpConfig.HeaderCRC = 24;
      break;
    case 18:
      hEpConfig->origEpConfig.HeaderCRC = 32;
      break;
    default:
      hEpConfig->origEpConfig.HeaderCRC = *tmp;
      break;
    }
    offset += 5;
  }

#if RS_FEC_CAPABILITY
  EPTool_BitsToBytes_CharToInt(&tmpBuffer[offset], tmp, 7);
  hEpConfig->origEpConfig.RSFecCapability = *tmp;
  offset += 7;
#endif  

  if(offset != configLength){
    fprintf(stderr, "EPTool_ParseEPSpecificConfig(): libeptool.c: Error Parsing configBuffer");
    free(tmpBuffer);
    return -1;
  }
      
  free(tmpBuffer);

  return 0;
  
}


static int EPTool_BytesToBits_IntToChar(unsigned int *inbuf, unsigned char *outbuf, int bitlen)
{
  int i, j, k;
  unsigned char tmpBuffer[16]={0};
  int tmp;

  k = bitlen;

  for(i=0; i<(bitlen+15)/16; i++){
    tmp = inbuf[i];
    for (j=16; j>0; ){
      tmpBuffer[--j] = tmp%2;
      tmp >>= 1;
    }
    
    if (k<16){
      for (j=0; j<k; j++){
        outbuf[j] = tmpBuffer[16-k+j];
      }
      return 0;
    } else {
      for (j=0; j<16; j++){
        outbuf[j] = tmpBuffer[j];
      }
      k -= 16;
    }
  }
  return 0;
}


static int EPTool_BitsToBytes_CharToInt(unsigned char *inbuf, int *outbuf, int bitlen)
{
  int i, j, k;
  int factor;
  int tmp = 0;

  k = bitlen;

  for (i=0; i<(bitlen+15)/16; i++){
    if (k > 16){
      for (j=0, factor = 32768; j< 16; i++){
        tmp += inbuf[j]*factor;
        factor >>= 1;
      }
      outbuf[i] = tmp;
      k -= 16;
    } else {
      for (j=0, factor = (32768 >> (16-k)); j<k; j++){
        tmp += inbuf[j]*factor;
        factor >>= 1;
      }
      outbuf[i] = tmp;
      return 0;
    }
  }
  return 0;
}


static void EPTool_BytesToBits_CharToChar(unsigned char *inbuf, unsigned char *outbuf, int bitlen)
{
  int i, bitrem = 0;
  unsigned char tmp = 0;

  for(i=0; i< bitlen; i++){
    if(bitrem == 0){
      tmp = inbuf[i/8] ;
      bitrem = 8;
    }
    outbuf[i] = ((tmp & 0x80) == 0x80);
    tmp <<= 1;
    bitrem --;
  }
}

void EPTool_GetClassLength(HANDLE_EP_CONFIG self, int choiceOfPred, unsigned int* class_length)
{
  int i;

  for(i=0; i<(int)self->epConfig.fattr[choiceOfPred].ClassCount; i++) {
    class_length[i] = self->epConfig.fattr[choiceOfPred].cattr[i].ClassBitCount;
  }
}

